#include <iostream>
#include <cassert>
#include <map>
#include "Game/Item.h"
#include "NPCBushBomber.h"
#include "Tile/Map.h"
#include "Game/NPC.h"
#include "Game/Player.h"
#include "General/graphics.h"
#include "Game/Scoreboard.h"
#include "Game/ItemEffect.h"

using namespace Salp::General;
using namespace std;

namespace Salp {
namespace Game {

Game::State *Item::_gamestate = 0;

Item::Item(const Tile::Sequence *tseq,
           const Uint32 id) {
    //    assert(tseq != 0); // well, unless you don't draw it, i guess.

    _hp = 0;
    _sequence = 0;
    _dirty = 0;
    _xpos = 0;
    _ypos = 0;
    _xvel = 0;
    _yvel = 0;
    _realxpos = 0;
    _realypos = 0;
    _xvelthresh = 10;
    _yvelthresh = 100;
    _frameid = 0;
    _framestarttime = 0;
    _seqstarttime = 0;
    _type = (~0) & ~GI_CAN_RIDE_ITEMS
            & ~GI_FACING_RIGHT; // no item-riding for you!
    _teamid = 0;
    _status = _old_status = 0;
    _map = 0;

    if(tseq) {
        _width = tseq->getWidth();
        _height = tseq->getHeight();
        setSequence(tseq); // cuz we do more stuff here
        assert(_seqstarttime != 0);
        assert(_sequence != 0);
    }


    setId(id);
    
//assert(_seqstarttime != 0);
}

Item::~Item() {
}

void Item::animate() {

}

void Item::animatePrimitive() {
    // Check seq info - call onDrewLastFrame if necessary    
    if(_seqstarttime && _sequence) {
        unsigned int curr_time = SDL_GetTicks();
        unsigned int seqdiff = curr_time - _seqstarttime;
        unsigned int cycle_time = _sequence->getCycleTime();
        if (seqdiff >= cycle_time && _sequence->_framecount > 1) {
            onDrewLastFrame();
            // update start time
            _seqstarttime = cycle_time ? 
                            curr_time - (seqdiff % cycle_time) :
                            curr_time;
        }
    }
}

void Item::animatePos(int frame_count) {
    // Uses basic item properties such as _xvel and _yvel to
    // change _xpos and _ypos accordingly. Does collision detection
    // if the item is allowed to collide with other items or the map.
    // see _type properties for these settings; they're on by default.
    //
    assert(frame_count != 0);
    // save old position
    int oldxpos=_xpos, oldypos=_ypos;

    int inherited_xvel=0, inherited_yvel=0;

    // add cheap gravity -- remember, gravity is upside down
    if(_type & GI_IS_AFFECTED_BY_GRAVITY)
        _yvel += 1 * frame_count;

    // clip velocity values
    if(_type & GI_USE_VEL_THRESH) {
        if(_yvel < -_yvelthresh)
            _yvel = -_yvelthresh;
        else if(_yvel > _yvelthresh)
            _yvel = _yvelthresh;
        if(_xvel < -_xvelthresh)
            _xvel = -_xvelthresh;
        else if(_xvel > _xvelthresh)
            _xvel = _xvelthresh;
    }

    if (!_yvel && !_xvel) return; // skip out from the fun if we're not moving at all
    
    //
    // ###### collision Detection Below ########
    //

    _old_status = _status; // track old status so we can see what's different
    _status &= ~GI_FORCED_ANYWHERE; // clear collision list

    // animate ypos
    _ypos += _yvel*frame_count + inherited_yvel;

    // check y-map collisions
    if( _type & GI_CAN_COLLIDE_WITH_MAP   &&
            oldypos != _ypos )
        detectCollisionMapY(oldxpos, oldypos);

    // animate xpos
    _xpos += _xvel*frame_count + inherited_xvel;

    // check x-map collisions
    if( (_type & GI_CAN_COLLIDE_WITH_MAP)   &&
            oldxpos != _xpos )
        detectCollisionMapX(oldxpos, oldypos);

    int collided_with_map = _status & GI_FORCED_ANYWHERE;

    // check for any item-item collisions
    Item * collided_with_item = 0;

    if( _type & GI_CAN_COLLIDE_WITH_ITEM )
        collided_with_item = detectCollisionItem(oldxpos, oldypos);

    // Check whether I've gone OOB
    if (_type & GI_USE_LOOSE_OOB)
        detectOOB();
    else
        detectOOBStrict();
        
    // Make sure the new tile position is not out of map bounds.
    // ... or die!
    assert(_ypos > (int) -_height);
    assert(_xpos > (int) -_width);
    assert(_ypos < (int) _map->getMapHeight()*_map->getTileSet()->getHeight());
    assert(_xpos < (int) _map->getMapWidth()*_map->getTileSet()->getWidth());


    // Alert me of what's going on, if I care to know.
    if(!collided_with_item && collided_with_map) { // && _type & GI_TRACK_GROUNDEDNESS
        onCollisionMap();
    }
}

void Item::animatePosSoft() {
     unsigned int dx = abs(_xpos - _realxpos), dy = abs(_ypos - _realypos);
     
     if (dx < 4)         _xpos += sign(_realxpos - _xpos);
     else if (dx < 9)   _xpos += sign(_realxpos - _xpos) << 1;
     else if (dx < 200)   _xpos = (_xpos + _realxpos) >> 1;
     else                _xpos = _realxpos;
     
     if (dy < 4)         _ypos += sign(_realypos - _ypos);
     else if (dy < 9)   _ypos += sign(_realypos - _ypos) << 1;
     else if (dy < 200)   _ypos = (_ypos + _realypos) >> 1;
     else                _ypos = _realypos;
/*     cout << "XVEL: " << int (_xvel) << ", YVEL: " << int(_yvel) << endl;
     _xpos += _xvel;
     _ypos += _yvel;
     
     if (dx < 4)         _xpos += sign(_realxpos - _xpos);
     else if (dx > 100)               _xpos = _realxpos;
     
     if (dy < 4)         _ypos += sign(_realypos - _ypos);
     else if (dy > 100)               _ypos = _realypos;
//_xpos = _realxpos;
//_ypos = _realypos;
*/
}


bool Item::canMoveInXVel() {
    // Returns true if you can move in the direction your _xvel
    // faces. So if your xvel is positive, it returns true if
    // you can move right. False if _xvel is zero.
    if(!_xvel)
        return true;

    return _xvel>0?!(_status & GI_FORCED_LEFT)  :
           !(_status & GI_FORCED_RIGHT);
}

bool Item::canMoveInYVel() {
    // Returns true if you can move in the direction your _yvel
    // faces. So if your yvel is positive, it returns true if
    // you can move down. False if _yvel is zero.
    if(!_yvel)
        return true;

    return _yvel>0?!(_status & GI_FORCED_UP)  :
           !(_status & GI_FORCED_DOWN);
}


bool Item::collidesWithItemX(Item *target) const {
    // Returns true if this object collides with <target>
    // in the 1-dimensional X plane.
    return  (
                (_xpos >= target->_xpos &&
                 _xpos < target->_xpos + target->_width)

                ||

                (target->_xpos >= _xpos &&
                 target->_xpos < _xpos + _width)

            );

}

bool Item::collidesWithItemY(Item *target) const {
    // Returns true if this object collides with <target>
    // in the 1-dimensional Y plane.
    return  (
                (_ypos >= target->_ypos &&
                 _ypos < target->_ypos + target->_height)

                ||

                (target->_ypos >= _ypos &&
                 target->_ypos < _ypos + _height)

            );

}

Game::ItemEffect * Item::createEffect(Tile::Sequence *tseq, int direction)
{
    // Creates, adds and returns an effect where i am currently
    // located.
    assert(tseq != 0);

    Game::ItemEffect *neweffect = new ItemEffect(tseq);
    
    if (direction) {
        neweffect->setPos(this);
        if (direction & DIR_UP)          ((Game::Item *) neweffect)->_ypos -= this->_height;
        else if (direction & DIR_DOWN)   ((Game::Item *) neweffect)->_ypos += this->_height;
        if (direction & DIR_LEFT)        ((Game::Item *) neweffect)->_xpos -= this->_width;
        else if (direction & DIR_RIGHT)  ((Game::Item *) neweffect)->_xpos += this->_width;
    } else {
        neweffect->setPosC(this);
    }
    
    neweffect->setMap(_map);
    getGameState()->addItem(neweffect);

    return neweffect;
}

bool Item::decHP(const int hp, Item *killer, WeaponType wtype)
{
    bool fatal = setHP(_hp - hp, killer);    

    // Log the damage that occurred. This should account for most of the possible deaths and stuff.
    if (isPlayer()) {
        DamageRec *drec = new DamageRec(_id, wtype, hp, fatal, killer?killer->_id:0);
        assert(drec != 0);
        getGameState()->getScoreboard()->addEvent(drec);
    }

    return fatal;
}

Item * Item::detectCollisionItem(const Sint32& oldxpos,
                                  const Sint32& oldypos) {
    // Detects and handles X- and Y- collisions with other items
    //

    assert(_type & GI_CAN_COLLIDE_WITH_ITEM);
    
    // Get list of objects to test collision against
    map<Uint32, Game::Item *> *items;
    map<Uint32, Game::Item *>::iterator candidate;

    items = getGameState()->getItemsICollidable();

    Item *lastvictim=0;
    int oldxvel = 0, oldyvel = 0;

    unsigned int new_status = 0, new_vic_status = 0;

    // Check item collision against all items. Unfortunately, that means
    // this function is O(n), and that we could get to a worst-case scenerio
    // of O(n^2) where n is the number of MOVING, ITEM-COLLIDABLE items in the game.
    // So really it doesn't get nearly this bad if we choose _types properly. It
    // may make sense down the road to use a data structure (even if they're 
    // X- and Y- linked lists) to speed up collision detection for most cases.
    
    for(candidate = items->begin(); candidate != items->end(); ++candidate) {
        // don't test against self
        if(candidate->first == _id)
            continue;

        // Get their pointer
        Item *victim = candidate->second;
        
        new_vic_status = 0;
        
        //
        // CHECK Y
        //
        if(_ypos > oldypos &&
                victim->_ypos  >=  oldypos + _height && // they're where I'm heading
                victim->_ypos <= _ypos + _height &&       // they were in the way
                collidesWithItemX(victim)
          ) {
            // COLLIDED GOING DOWN
            if((_type & GI_ON_COLLIDE_REPOS) && (victim->_type & GI_ON_COLLIDE_REPOS)) {
                _ypos = victim->_ypos - _height;

                oldyvel = _yvel?_yvel:oldyvel;
                if (_type & GI_ON_COLLIDE_LOSE_VEL)
                    _yvel = 0;

            }

            new_status |= GI_FORCED_UP;
            new_vic_status |= GI_FORCED_DOWN;
 
            lastvictim = victim;

        } else if (_ypos < oldypos &&
                   victim->_ypos + victim->_height  <=  oldypos &&   // they're where i'm heading
                   victim->_ypos + victim->_height  >= _ypos &&      // they were in the way
                   collidesWithItemX(victim)
                  ) {

            // COLLIDED GOING UP
            if((_type & GI_ON_COLLIDE_REPOS) && (victim->_type & GI_ON_COLLIDE_REPOS)) {
                _ypos = victim->_ypos + victim->_height;

                oldyvel = _yvel?_yvel:oldyvel;
                if (_type & GI_ON_COLLIDE_LOSE_VEL)
                    _yvel = 0;
            }

            new_status |= GI_FORCED_DOWN;
            new_vic_status |= GI_FORCED_UP;

            lastvictim = victim;
        }


        //
        // CHECK X
        //
        if(_xpos > oldxpos &&
                victim->_xpos  >=  oldxpos + _width && // they're where I'm heading
                victim->_xpos <= _xpos + _width &&     // they were in the way
                collidesWithItemY(victim)
          ) {

            // COLLIDED GOING RIGHT
            if((_type & GI_ON_COLLIDE_REPOS) && (victim->_type & GI_ON_COLLIDE_REPOS)) {
                _xpos = victim->_xpos - _width;

                oldxvel = _xvel?_xvel:oldxvel;

                if (_type & GI_ON_COLLIDE_LOSE_VEL)
                    _xvel = 0;
            }

            new_status |= GI_FORCED_LEFT;
            new_vic_status |= GI_FORCED_RIGHT;
            
            lastvictim = victim;

        } else if (_xpos < oldxpos &&
                   victim->_xpos + victim->_width  <=  oldxpos && // they're where i'm heading
                   victim->_xpos + victim->_width  >= _xpos &&    // they were in the way
                   collidesWithItemY(victim)
                  ) {

            if((_type & GI_ON_COLLIDE_REPOS) && (victim->_type & GI_ON_COLLIDE_REPOS)) {
                _xpos = victim->_xpos + victim->_width;

                oldxvel = _xvel?_xvel:oldxvel;
                if (_type & GI_ON_COLLIDE_LOSE_VEL)
                    _xvel = 0;
            }

            new_status |= GI_FORCED_RIGHT;
            new_vic_status |= GI_FORCED_LEFT;

            lastvictim = victim;
        }


    }

    // If we hit someone, let'em know
    if(lastvictim) {

        if((_type & GI_ON_COLLIDE_REPOS) && (lastvictim->_type & GI_ON_COLLIDE_REPOS)) {
        
            // So they collided right? Well, let's do some transfer of velocity.
            
            if ((lastvictim->_type & GI_ON_COLLIDE_TAKE_VEL) &&
                    (_type & GI_ON_COLLIDE_GIVE_VEL)
               ) {
                lastvictim->_xvel += oldxvel;
                lastvictim->_yvel += oldyvel;
            }

            // ??? Seems like something's missing in here now.
        }

        // update statuses
        _status |= new_status;
        lastvictim->_status |= new_vic_status;
        
        lastvictim->onCollisionItem(this, new_status);
        onCollisionItem(lastvictim, new_vic_status);

    }

    // And return the victim for other use
    return lastvictim;

}

Item * Item::detectCollisionItemAll() {
    // Detects and handles X- and Y- collisions with other items
    //

    assert(_type & GI_CAN_COLLIDE_WITH_ITEM);
    
    // Get list of objects to test collision against
    map<Uint32, Game::Item *> *items;
    map<Uint32, Game::Item *>::iterator candidate;

    items = getGameState()->getItemsICollidable();

    Item *lastvictim=0;
    for(candidate = items->begin(); candidate != items->end(); ++candidate) {
        // don't test against self
        if(candidate->first == _id)
            continue;

        // Get their pointer
        Item *victim = candidate->second;

        if (collidesWithItemY(victim) && collidesWithItemX(victim)) {
            // status goes to 0? or unchanged?
            //_status = GI_FORCED_RIGHT;
            //victim->_status |= GI_FORCED_LEFT;
            
            // should this keep looping, or quit here?
            lastvictim = victim;
            break;
        }
    }

    // If we hit someone, let'em know
    if(lastvictim) {
        lastvictim->onCollisionItem(this, false);
        onCollisionItem(lastvictim, true);
    }

    // And return the victim for other use
    return lastvictim;

}


void Item::detectCollisionMapX(const Sint32& oldxpos,
                                const Sint32& oldypos) {
    //
    // Checks for map collisions in the X direction only.
    //

    int scanx, scany;

    int yfrom = _map->getMapCellIndexY(_ypos),
                yto   = _map->getMapCellIndexY(_ypos + _height - 1);

    if(_xpos > oldxpos) {
        // If object MOVED RIGHT, we want to check
        // MapCells' **LEFT** bounds

        int frightx = _map->getMapCellIndexX(oldxpos + _width - 1);
        int trightx = _map->getMapCellIndexX(_xpos + _width - 1);

        for(scanx=frightx+1; scanx <= trightx; scanx++) {
            // we have a collision if the bottom of the object
            // (marked by scanx) >= any upper bounds

            for(scany=yfrom; scany <= yto; scany++) {
                Tile::MapCell *scancell = _map->getMapCell(scanx, scany);
                if(scancell->bounds & MC_LEFT_BOUND) {
                    _xpos = scanx * _map->getTileSet()->getWidth() -
                            _width;

                    // if our bounding efforts have been counter-productive
                    // leave it as-is
                    if(_xpos < oldxpos)
                        _xpos = oldxpos;

                    // we were blocked when heading right. so absorb some nrg
                    if (_type & GI_ON_COLLIDE_LOSE_VEL) {
                        if(_xvel != 0)
                            --_xvel;
                        else
                            _xvel = 0;
                    }

                    _status |= GI_FORCED_LEFT;

                    goto end_x_bounds_scan;
                }
            }

        }
    } else if(_xpos < oldxpos) {
        // If object MOVED LEFT, we want to check
        // MapCells' **RIGHT** bounds

        int fleftx = _map->getMapCellIndexX(oldxpos);
        int tleftx = _map->getMapCellIndexX(_xpos);

        // remember, we're using raster coords
        for(scanx=fleftx; scanx >= tleftx; scanx--) {
            // we have a collision if the top of the object
            // (marked by scanx) >= any upper bounds

            for(scany=yfrom; scany <= yto; scany++) {
                Tile::MapCell *scancell = _map->getMapCell(scanx, scany);
                if(scancell->bounds & MC_RIGHT_BOUND) {
                    _xpos = (scanx+1) * _map->getTileSet()->getWidth();

                    // if our bounding efforts have been counter-productive
                    // leave it as-is
                    if(_xpos > oldxpos)
                        _xpos = oldxpos;

                    // we were blocked when heading left. so absorb some nrg
                    if (_type & GI_ON_COLLIDE_LOSE_VEL) {
                        if(_xvel != 0)
                            ++_xvel;
                        else
                            _xvel = 0;
                    }

                    _status |= GI_FORCED_RIGHT;

                    goto end_x_bounds_scan;
                }
            }

        }
    }

end_x_bounds_scan:

    return;
}

void Item::detectCollisionMapY(const Sint32& oldxpos,
                                const Sint32& oldypos) {
    //
    // Checks for map collisions in the X direction only.
    //

    int scanx, scany;

    int xfrom = _map->getMapCellIndexX(_xpos),
                xto   = _map->getMapCellIndexX(_xpos + _width - 1);

    if(_ypos > oldypos) {
        // If object MOVED DOWNward, we want to check
        // MapCells' **TOP** bounds
        int fboty = _map->getMapCellIndexY(oldypos + _height - 1);
        int tboty = _map->getMapCellIndexY(_ypos + _height - 1);

        for(scany=fboty+1; scany <= tboty; scany++) {
            // we have a collision if the bottom of the object
            // (marked by scany) >= any upper bounds

            for(scanx=xfrom; scanx <= xto; scanx++) {
                Tile::MapCell *scancell = _map->getMapCell(scanx, scany);
                if(scancell->bounds & MC_TOP_BOUND) {
                    _ypos = scany * _map->getTileSet()->getHeight() -
                            _height;

                    // if our bounding efforts have been counter-productive
                    // leave it as-is

                    // we were blocked when heading down. so set vel to 0
                    if (_type & GI_ON_COLLIDE_LOSE_VEL)
                        _yvel=0;
                    _status |= GI_FORCED_UP;

                    goto end_y_bounds_scan;
                }
            }

        }
    } else if(_ypos < oldypos) {
        // If object MOVED UPward, we want to check
        // MapCells' **BOTTOM** bounds
        int ftopy = _map->getMapCellIndexY(oldypos);
        int ttopy = _map->getMapCellIndexY(_ypos);

        // remember, we're using raster coords
        for(scany=ftopy; scany >= ttopy; scany--) {
            // we have a collision if the top of the object
            // (marked by scany) >= any upper bounds

            for(scanx=xfrom; scanx <= xto; scanx++) {
                Tile::MapCell *scancell = _map->getMapCell(scanx, scany);
                if(scancell->bounds & MC_BOTTOM_BOUND) {
                    _ypos = (scany+1) * _map->getTileSet()->getHeight();

                    // if our bounding efforts have been counter-productive
                    // leave it as-is

                    // we were blocked when heading down. so absorb some nrg
                    if (_type & GI_ON_COLLIDE_LOSE_VEL)
                        _yvel=0;
                    _status |= GI_FORCED_DOWN;

                    goto end_y_bounds_scan;
                }
            }

        }
    }

end_y_bounds_scan:

    return;
}

void Item::detectOOB() {
    //
    // Temp oob checks:
    //     If this object goes completely out of bounds, throw it on the
    //     edge so that it *just* doesn't show.
    //
    //

    int inbounds;
    bool x_oob = false, y_oob = false;

    if(_ypos <= (inbounds = (int) -_height)) {
        _ypos = inbounds+1;
        y_oob = true;
    } else if(_ypos >= (inbounds = (int)
                                   _map->getMapHeight()*_map->getTileSet()->getHeight())) {
        _ypos = inbounds-1;
        y_oob = true;
    }

    if(_xpos <= (inbounds = (int) -_width)) {
        _xpos = inbounds+1;
        x_oob = true;
    } else if(_xpos >= (inbounds = (int)
                                   _map->getMapWidth()*_map->getTileSet()->getWidth())) {
        _xpos = inbounds-1;
        x_oob = true;
    }

    if(x_oob || y_oob) {
        onOOB(x_oob, y_oob);
    }
}

void Item::detectOOBStrict() {
    //
    // Temp oob checks:
    //     If this object goes completely out of bounds, throw it on the
    //     edge so that it *just* doesn't show.
    //
    //

    bool x_oob = false, y_oob = false;

    if(_ypos < 0 || _xpos + _height > (int) _map->getPixHeight()-1)
        y_oob = true;
        
    if(_xpos < 0 || _xpos + _width > (int) _map->getPixWidth()-1)
        x_oob = true;

    if(x_oob || y_oob) 
        onOOB(x_oob, y_oob);
    
}


void Item::drawImage(SDL_Surface *screen,
                     const Sint32 topx,
                     const Sint32 topy) {
    // Draws the Item's image on the <screen> surface.
    // We are to assume that the upper left coordinate of the screen
    // is at (topx, topy).
    
    assert(_sequence != 0);
    assert(_type & GI_IS_VISIBLE);

    // don't bother drawing it if it's off the screen
    assert(canDrawImage(screen, topx, topy));
    _sequence->drawImage(screen, _frameid, _xpos - topx, _ypos - topy,
                         _framestarttime, (_type & GI_FACING_RIGHT) != 0);

    // don't bother drawing it if it's off the screen
    if(General::debugGraphics && !(_realxpos >= topx +  screen->w ||
            _realypos >= topy + screen->h ||
            (Sint32) (_realxpos + _width) <= topx ||
            (Sint32) (_realypos + _height) <= topy))
    {        
        _sequence->drawRect(screen, _realxpos - topx, _realypos - topy);
    }
}

void Item::drawHealthBar(SDL_Surface *screen, const Sint32 topx, const Sint32 topy) {
    // also don't bother drawing it if no room to the left
    if(topx >= _xpos - 1)
        return;

    // otherwise let's draw the bar
    static SDL_Rect dst;
/* OLD VERTICAL BAR
    dst.h = _height * _hp / 100;
    dst.w = 3;
    dst.x = _xpos - 4 - topx;
    dst.y = _ypos - topy - dst.h + _height;
*/
    dst.h = 8;
//    dst.w = _width * _hp / 100;
    dst.w = _width;
    dst.x = _xpos - topx;
    dst.y = _ypos - topy - dst.h + _height;
    static unsigned char greenness=0;
    static bool greenup  = true; // green value is going up?
    if (greenup)  greenup = (greenness += 4) != 132;
    else          greenup = (greenness -= 4) == 0;

    fillRectAlpha(screen, &dst,
                 SDL_MapRGB(screen->format,
                            255 - 255 * _hp / 100,  // R 
                            greenness,              // G
                            255 * _hp / 100         // B
                ), 
                   128 // half there, faster
                );

}

void Item::faceDown() {
    // Has our Item face downward.
    setType((getType() & ~GI_FACING_UP) | GI_FACING_DOWN);
}

void Item::faceLeft() {
    // Has our Item face leftward.
    setType((getType() & ~GI_FACING_RIGHT) | GI_FACING_LEFT);
}

void Item::faceRight() {
    // Has our Item face right.
    setType((getType() & ~GI_FACING_LEFT) | GI_FACING_RIGHT);
}

void Item::faceUp() {
    // Has our Item face upward.
    setType((getType() & ~GI_FACING_DOWN) | GI_FACING_UP);
}

string Item::getDirtyString()
{
    return getDirtyString(_dirty);
}

unsigned int Item::getDirForcedIndex() const
{
    // Returns index signifying directions that it probably
    // was going in order to be forced the way it is.

    unsigned int i=4; // middle

    // returns direction i am forced, 0-8
    if (_status & GI_FORCED_LEFT)  ++i;
    if (_status & GI_FORCED_RIGHT) --i;
    if (_status & GI_FORCED_UP)    i += 3;
    if (_status & GI_FORCED_DOWN)  i -= 3;
    return i;
}

#define string_AppendIfDirty(DIRTY_VAR, STR, DIRTY_TYPE) \
            if (DIRTY_VAR & DIRTY_TYPE)  { STR += " "; STR += #DIRTY_TYPE; } 
string Item::getDirtyString(unsigned int dirty)
{
    string tmp = "The following are dirty:";
    string_AppendIfDirty(dirty, tmp, DIRTY_SEQUENCE);
    string_AppendIfDirty(dirty, tmp, DIRTY_XPOS);
    string_AppendIfDirty(dirty, tmp, DIRTY_YPOS);
    string_AppendIfDirty(dirty, tmp, DIRTY_MAP);
    string_AppendIfDirty(dirty, tmp, DIRTY_TYPE);
    string_AppendIfDirty(dirty, tmp, DIRTY_HP);
    string_AppendIfDirty(dirty, tmp, DIRTY_XVEL);
    string_AppendIfDirty(dirty, tmp, DIRTY_YVEL);
    string_AppendIfDirty(dirty, tmp, DIRTY_NAME);
    if (!tmp.length()) return "Nothing was dirty.";
    return tmp;
}

void Item::getPosBL(Sint32 &xpos, Sint32 &ypos) {
    // Sets xpos and ypos to the coordinate of the
    // [B]ottom [L]eft corner of this item.
    xpos = _xpos;
    ypos = _ypos + _height - 1;
}

void Item::getPosBR(Sint32 &xpos, Sint32 &ypos) {
    // Sets xpos and ypos to the coordinate of the
    // [B]ottom [R]ight corner of this item.
    xpos = _xpos + _width - 1;
    ypos = _ypos + _height - 1;
}

void Item::getPosC(Sint32 &xpos, Sint32 &ypos) {
    // Sets xpos and ypos to the coordinate of the
    // [C]enter of this item. The coordinate is rounded
    // to the upper-left when no true center exists.
    xpos = ((_xpos << 1) + _width - 1) >> 1;
    ypos = ((_ypos << 1) + _height - 1) >> 1;
}

void Item::getPosTL(Sint32 &xpos, Sint32 &ypos) {
    // Sets xpos and ypos to the coordinate of the
    // [T]op [L]eft corner of this item.
    xpos = _xpos;
    ypos = _ypos;
}

void Item::getPosTR(Sint32 &xpos, Sint32 &ypos) {
    // Sets xpos and ypos to the coordinate of the
    // [T]op [R]ight corner of this item.
    xpos = _xpos + _width - 1;
    ypos = _ypos;
}

unsigned int Item::getTeam() const {
    // Returns the ID of the team which this Item beSint32s to.
    return _teamid;
}

bool Item::incHP(const int hp, Item *killer, WeaponType wtype)
{    
    bool fatal = setHP(_hp + hp, killer);

    // Log the damage that occurred. This should account for most of the possible deaths and stuff.
    if (isPlayer()) {
        DamageRec *drec = new DamageRec(_id, wtype, hp, fatal, killer?killer->_id:0);
        assert(drec != 0);
        getGameState()->getScoreboard()->addEvent(drec);
    }
    
    if (isNPC()) {
        Game::NPC *npc = (Game::NPC *) this;
        if (_hp > npc->getMaxHP())
            _hp = npc->getMaxHP();
    }

    return fatal;
}

bool Item::isOnSameTeamAs(Item * gi) const {
    // Returns true if I am on the same team as <gi>
    return gi->_teamid == _teamid && _teamid != 0; // Team id of 0 means no team!
}

void Item::onCollisionMap() {
    // Called when I have collided with the map.
}

void Item::onCollisionItem(Item *victim, unsigned int vic_collide_status) {
    // Called when I have collided with another item, <victim>
}

void Item::onDrewLastFrame() {
    // Called when my last frame of animation (sequence) has been drawn
}

void Item::onOOB(bool x_oob, bool y_oob) {
    // Called when my character has gone out of map bounds
}

void Item::onDeath(Item *killer) {
    // Called when my character has died. If <killer> is nonzero,
    // it points to the Item that killed me.
}

void Item::onKill(Item *victim) {
    // Called when I have killed <victim>. <victim> should never be zero.
}

void Item::setGameState(const Game::State *gs) {
    // Sets the gamestate that I beSint32 to
    _gamestate = (Game::State *) gs;
}

Tile::Sequence * Item::getSequence() const {
    // Returns a pointer to the sequence associated with this Item
    return _sequence;
}

unsigned int Item::getStatus() const {
    // Returns my set of _status flags
    return _status;
}

void Item::setTeam(const unsigned int teamid) {
    // Make me beSint32 to team <teamid>
    _teamid = teamid;
}

unsigned int Item::getType() const {
    // Returns my _type flags
    return _type;
}

void Item::setHeight(const int height) {
    // Returns my _height
    _height = height;
}

bool Item::setHP(const int hp, Item *killer)
{
    if (_hp == hp) return false;
    
    // Returns true if this setHP caused the victim to die
    if (hp <= 0 && _hp > 0) {
        _hp = hp;
        if(killer) killer->onKill(this);
        onDeath(killer);
        return true;
    } else {
        _hp = hp;
    }
    _dirty |= DIRTY_HP;

    return false;
}

Uint32 Item::setId(const Uint32 id,
                          const Uint32 newCounterValue) {
    static Uint32 idCounter = 0;

    // If newCounterValue is set, then just set the counter to
    // the new value and continue onward
    if(newCounterValue != 0) {
        idCounter = newCounterValue;
    } else if (id != 0) {
        _id = id;
    } else {
        assert(!_gamestate->isLevLocked() || _gamestate->getServer() != 0);
        _id = ++idCounter;
    }

    return newCounterValue;
}

void Item::setMap(const Tile::Map *tm) {
    // Sets my _map to <tm>
    assert(tm != 0);
    _map = (Tile::Map *) tm;
    _dirty |= DIRTY_MAP;
}

void Item::setPos(Game::Item *gi) {
    // Sets my position to be the same as <gi>'s (that is, the xpos and ypos)
    assert(gi != 0);

    _xpos = gi->_xpos;
    _ypos = gi->_ypos;
}

void Item::setPosByMapCode(const unsigned int mapcode) {
    // Sets my position to be the upper-left coordinate of the
    // tile with mapcode <mapcode>. Program exits with error
    // if that mapcode does not exist.
    assert(_map != 0);

    Sint32 xpos, ypos;
    bool success = _map->getMapCellCoordsByMapCode(mapcode, xpos, ypos);

    if(!success) { // Let's hope that mapcode actually existed :-).
        cout << "Could not find mapcode " << mapcode << " in map " << _map->getName() << endl;
        exit(1);
    }
    setPos(xpos, ypos); // Let setPos handle dirtiness
}

void Item::setPosC(Game::Item *gi) {
    // Sets center of me to be same as center of item gi
    assert(gi != 0);

    gi->getPosC(_xpos, _ypos);
    _xpos -= _width  >> 1; // move to be at center of item
    _ypos -= _height >> 1;
}

void Item::setPosClipToMap() {
    if(_ypos < 0)
        _ypos = 0;
    else if (_ypos + _height > (int) _map->getPixHeight()-1)
        _ypos = _map->getPixHeight() - _height;
        
    if(_xpos < 0)
        _xpos = 0;
    else if (_xpos + _width > (int) _map->getPixWidth()-1)
        _xpos = _map->getPixWidth() - _width;
}

void Item::setPosStart() {
    // Smartly go to my starting position.
    // What this means is dependent on how many people are in my team,
    // the starting positions available, etc.
    
    setPosByMapCode(Tile::MC_START_TEAM_A); // We're guaranteed that this exists.
}

void Item::setSequence(const Tile::Sequence *sequence)
{
    assert(sequence != 0);
    if(_sequence == sequence) return;
    setSequenceHard(sequence);
}


void Item::setStatus(const unsigned int status) {
    // Sets my _status to <status>
    _status = status;
}

void Item::setType(const unsigned int type) {
    // Sets my _type to <type>
    _type = type;
    _dirty |= DIRTY_TYPE;
}

void Item::setWidth(const int width) {
    // Sets my _width to <width>
    _width = width;
}

int Item::writeDataCreate(char * data) {
    // Writes to <data> the data necessary
    // to [send over a network so the client can] create said
    // item. Clearly, we don't include big details about the item
    // here for bandwidth (and triviality) reasons; these reflect
    // the visual (and perhaps dead reckoning-related) aspects of
    // the item.
    //
    // It is assumed that <data> is large enough to hold this info.
    // since this info length is O(1), there should be few worries of
    // buffer overflow here.
    //    
    char * olddata = data;
    DATA_COPY_AND_INC(data, _id);
    unsigned char itype = getItemType();
    DATA_COPY_AND_INC(data, itype);
    DATA_COPY_AND_INC(data, _sequence->_seqid); // how crude
    DATA_COPY_AND_INC(data, _map->_id);
    DATA_COPY_AND_INC(data, _xpos);
    DATA_COPY_AND_INC(data, _ypos);
//    DATA_COPY_AND_INC(data, _type);
//    DATA_COPY_AND_INC(data, _xvel);
//    DATA_COPY_AND_INC(data, _yvel);

    return data - olddata;
}

Item * Item::createFromData(char * data) {
    Uint32  id    = 0;
    unsigned short seqid = 0;
    unsigned char mapid = 0;
    unsigned char itype;

    DATA_READ_AND_INC(data, id);
    if (_gamestate->hasItem(id))
        return 0;
    DATA_IREAD_8(data, itype);
    DATA_READ_AND_INC(data, seqid);
    DATA_READ_AND_INC(data, mapid);

    Item *new_item;

    switch (itype) {
    case ITYPE_PLAYER:
        new_item = new Player(_gamestate->getTileSequence(seqid), id);
        break;

    case ITYPE_NPC:
        new_item = new NPC(_gamestate->getTileSequence(seqid), id);
        break;

    default:
        new_item = new Item(_gamestate->getTileSequence(seqid), id);
    }
    new_item->setMap(_gamestate->getTileMap(mapid));

    DATA_READ_AND_INC(data, new_item->_xpos); // ***
    DATA_READ_AND_INC(data, new_item->_ypos); // ***
//    DATA_READ_AND_INC(data, new_item->_type);

    new_item->_realxpos = new_item->_xpos;
    new_item->_realypos = new_item->_ypos;

//    DATA_READ_AND_INC(data, new_item->_xvel);
//    DATA_READ_AND_INC(data, new_item->_yvel);

    assert(new_item != 0);
    return new_item;
}

int Item::writeDataUpdate(char * data, bool full) {
    assert(_sequence != 0);

    char * olddata = data;
    DATA_COPY_AND_INC(data, _id);

    Uint16 dirty = full ? ~0 : _dirty;
    DATA_COPY_AND_INC(data, dirty);

    if(dirty & DIRTY_SEQUENCE)    DATA_COPY_AND_INC(data, _sequence->_seqid);
    if(dirty & DIRTY_MAP)         DATA_COPY_AND_INC(data, _map->_id);
    if(isPlayer() && (dirty & DIRTY_HP))
                                           DATA_COPY_AND_INC(data, _hp);
    if(dirty & DIRTY_XPOS)        DATA_COPY_AND_INC(data, _xpos);
    if(dirty & DIRTY_YPOS)        DATA_COPY_AND_INC(data, _ypos);
    if(dirty & DIRTY_TYPE)        DATA_COPY_AND_INC(data, _type);
    
    if (isNPC()) {
        Game::NPC *thisnpc = (Game::NPC *) this;
        if(dirty & DIRTY_NAME) { DATA_ICOPY_CSTR(data, thisnpc->_name.c_str(), thisnpc->_name.length()); }
    }
    
//    cout << "SXVE: " << int(_xvel) << ", SYVE: " << int(_yvel) << endl;
//    if(_dirty & DIRTY_XVEL)        DATA_COPY_AND_INC(data, _xvel);
//    if(_dirty & DIRTY_YVEL)        DATA_COPY_AND_INC(data, _yvel);
    //    DATA_COPY_AND_INC(data, _status);
    //    DATA_COPY_AND_INC(data, _teamid);     // could compress
    //    DATA_COPY_AND_INC(data, _xvelthresh); // this probably isn't true
    //    DATA_COPY_AND_INC(data, _yvelthresh); // .. might not even need sending
    //    DATA_COPY_AND_INC(data, _height);
    //    DATA_COPY_AND_INC(data, _width);

    return data - olddata;
}

Game::Item * Item::updateFromData(char * data) {


    Uint32  id    = 0;
    unsigned short seqid = 0;

    DATA_READ_AND_INC(data, id);

    Uint16 dirty;
    DATA_READ_AND_INC(data, dirty);
    Item *gi = _gamestate->getItemMaybe(id);
    if (!gi) return 0;
    assert(gi != 0);

    if (debugProtocol) cout << "\\--> " << getDirtyString(dirty) << "\n";


    if (dirty & DIRTY_SEQUENCE) {
        DATA_READ_AND_INC(data, seqid);
        gi->setSequenceHard(_gamestate->getTileSequence(seqid));
    }

    if (dirty & DIRTY_MAP) {
        unsigned char  mapid = 0;
        DATA_READ_AND_INC(data, mapid);
        gi->setMap(_gamestate->getTileMap(mapid));
    }

    if(gi->isPlayer() && (dirty & DIRTY_HP))
                                   DATA_READ_AND_INC(data, gi->_hp);
    if (dirty & DIRTY_XPOS)        DATA_READ_AND_INC(data, gi->_realxpos);
    if (dirty & DIRTY_YPOS)        DATA_READ_AND_INC(data, gi->_realypos);
    if (dirty & DIRTY_TYPE)        DATA_READ_AND_INC(data, gi->_type);
    if (gi->isNPC()) {
        Game::NPC *thisnpc = (Game::NPC *) gi;
        if(dirty & DIRTY_NAME)     DATA_IREAD_STR(data, thisnpc->_name); 
    }


//    if (dirty & DIRTY_XVEL)        DATA_READ_AND_INC(data, gi->_xvel);
//    if (dirty & DIRTY_YVEL)        DATA_READ_AND_INC(data, gi->_yvel);

     //    DATA_READ_AND_INC(data, gi->_status);
    //    DATA_READ_AND_INC(data, gi->_teamid); // could compress
    //    DATA_READ_AND_INC(data, gi->_xvelthresh); // this probably isn't true
    //    DATA_READ_AND_INC(data, gi->_yvelthresh); // .. might not even need sending
    //    DATA_READ_AND_INC(data, gi->_height);
    //    DATA_READ_AND_INC(data, gi->_width);

    return gi;
}



} // namespace Game
} // namespace Salp

