#ifndef __GAMEITEM_H
#define __GAMEITEM_H

#include "SALPSTUB.h"
#include "Game/State.h"
#include "Tile/Sequence.h"
#include "Tile/Map.h"

#include <iostream>
#include <cstring>

namespace Salp {
namespace Game {


class Item {
public:
    explicit Item(const Tile::Sequence *tseq,
                  const Uint32 id=0);
    virtual ~Item();

    virtual void animate();
    void animatePrimitive();
    void animatePos(int frames);
    void animatePosSoft();
    bool canDrawImage(SDL_Surface *screen, const Sint32 topx, const Sint32 topy);
    bool canMoveInXVel(); // returns true if the item can move in the dir that xvel heads. true if xvel==0
    bool canMoveInYVel();
    void clearDirty() { _dirty = 0; }
    Game::ItemEffect * createEffect(Tile::Sequence *tseq, int direction=0);
    bool decHP(const int hp, Item *killer=0, WeaponType wtype=WEAPON_UNKNOWN);
    Item * detectCollisionItemAll(); // checks against ALL possible items, not just those it might have run into when moving.
    virtual void drawImage(SDL_Surface *screen,
                           const Sint32 topx,
                           const Sint32 topy);
    void drawHealthBar(SDL_Surface *screen, const Sint32 topx, const Sint32 topy);
    void faceDown();
    void faceLeft();
    void faceRight();
    void faceUp();
    unsigned int getDirForcedIndex() const; // returns direction i am forced, 0-8
    string getDirtyString();
    static string getDirtyString(unsigned int dirty);
    Game::State * getGameState() const;
    int getHeight() const;
    int getHP() const;
    Uint32 getId() const;
    Tile::Map *getMap() const;
    void getPosBL(Sint32 &xpos, Sint32 &ypos);
    void getPosBR(Sint32 &xpos, Sint32 &ypos);
    void getPosC(Sint32 &xpos, Sint32 &ypos);
    void getPosTL(Sint32 &xpos, Sint32 &ypos);
    void getPosTR(Sint32 &xpos, Sint32 &ypos);
    virtual unsigned int getItemType() const { return ITYPE_GENERIC; };
    unsigned int getStatus() const;
    unsigned int getStatusOld() const;
    unsigned int getStatusDiff() const;
    Tile::Sequence * getSequence() const;
    unsigned int getTeam() const;
    unsigned int getType() const;
    int getWidth() const;
    int getXPos() const;
    int getXPosC() const;
    int getYPos() const;
    int getYPosC() const;
    int getXVel() const;
    int getYVel() const;
    bool incHP(const int hp, Item *killer=0, WeaponType wtype=WEAPON_UNKNOWN);
    unsigned int isDirty() const { return _dirty ; } // Returns nonzero val if certain item values have changed since _dirty was last cleared
    bool isFacingUp()    const { return (_type & GI_FACING_UP) != 0; }
    bool isFacingDown()  const { return (_type & GI_FACING_DOWN) != 0; }
    bool isFacingLeft()  const { return (_type & GI_FACING_LEFT) != 0; }
    bool isFacingRight() const { return (_type & GI_FACING_RIGHT) != 0; }
    bool isGrounded() const;
    bool isOnSameTeamAs(Item * gi) const;
    bool isPlayer() const { return getItemType() == ITYPE_PLAYER; }
    bool isNPC() const { return getItemType() == ITYPE_NPC || getItemType()==ITYPE_PLAYER; }
    virtual void onCollisionMap();
    virtual void onCollisionItem(Item *victim, unsigned int vic_collide_status);
    virtual void onDrewLastFrame();
    virtual void onDeath(Item *killer=0); // killer may not always be given
    virtual void onKill(Item *victim); 
    virtual void onOOB(bool x_oob, bool y_oob);
    void setDirty (const unsigned int flags) { _dirty |= flags; }
    bool setHP(const int hp, Item *killer=0);
    void setHeight(const int height);
    static void setGameState(const Game::State *gs);
    Uint32 setId(const Uint32 id=0, const Uint32 newCounterValue=0);
    void setMap(const Tile::Map *tm);
    void setPos(const Sint32 xpos, const Sint32 ypos);
    void setPos(Game::Item *gi); // sets to be same as item
    void setPosC(Game::Item *gi); // sets to be same as center of item
    void setPosClipToMap(); // moves pos so i can be seen on the map
    void setPosByMapCode(const unsigned int mapcode);
    void setPosStart(); // goes to start position (smartly)
    void setSequence(const Tile::Sequence *sequence);
    void setSequenceHard(const Tile::Sequence *sequence);
    void setStatus(const unsigned int status);
    void setTeam(const unsigned int teamid);
    void setType(const unsigned int type);
    void setTypeOff(const unsigned int type);
    void setTypeOn(const unsigned int type);
    void setVel(const Uint32 xvel, const Uint32 yvel);
    bool setVelTowardCoord(const int& targx, const int& targy, unsigned int speed);
    void setWidth(const int width);
    void setXPos(const int xpos);
    void setXVel(const int xvel);
    void setXVelThresh(const int xvelthresh);
    void setYPos(const int ypos);
    void setYVel(const int yvel);
    void setYVelThresh(const int yvelthresh);
    void unsetDirty (const unsigned int flags) { _dirty &= ~flags; }
    int writeDataCreate(char * data);
    int writeDataUpdate(char * data, bool full=false);
    static Game::Item *createFromData(char * data); // returns item created from data
    static Game::Item * Item::updateFromData(char * data);
    
private:

    bool collidesWithItemX(Item *target) const;
    bool collidesWithItemY(Item *target) const;
    Item * detectCollisionItem(const Sint32& oldxpos,
                                const Sint32& oldypos);
    void detectCollisionMapX(const Sint32& oldxpos,
                              const Sint32& oldypos);
    void detectCollisionMapY(const Sint32& oldxpos,
                              const Sint32& oldypos);
    void detectOOB();
    void detectOOBStrict();


    friend class Salp::Game::Player;
    friend class Salp::Game::ConvBubble;
    
    // ----- properties -----

    // ----- necessary [for server] to network
    Uint32         _id;           // if you change this type, check my networking code for this object
    
    Sint32 _xpos,   _ypos;          // top-left x/y position
    unsigned int   _type;
    unsigned int   _status;
    Tile::Sequence *_sequence;    // (holds general properties for this image)
    Tile::Map      *_map;         // this is the map we're on


    // ----- unnecessary to network
    static Game::State    *_gamestate;
    unsigned int _frameid;        // id of frame
    unsigned int _framestarttime; // when did we start using this frame? - used by sequence
    unsigned int _seqstarttime;   // when did we start using this SEQUENCE?
    unsigned int _dirty;
    unsigned int _old_status;
    int _hp;
  
    // ----- likely unnecessary to network
    Sint8  _xvel,   _yvel;          // movement velocity, in pixels/screenframe
    unsigned int   _teamid;
    int _xvelthresh, _yvelthresh; // velocity thresholds (item-specific??)
    int _height,     _width;      // item's width and height (in pixels) on the
    
    // ----- client-only
    Sint32 _realxpos, _realypos;   // what server tells us
};


//
// Include inline functions here in order to be linked
// properly.
//


inline bool Item::canDrawImage(SDL_Surface *screen, const Sint32 topx, const Sint32 topy)
{
    assert(screen != 0);
    return (!(_xpos >= topx +  screen->w ||
            _ypos >= topy + screen->h ||
            (Sint32) (_xpos + _width) <= topx ||
            (Sint32) (_ypos + _height) <= topy));
}


inline Game::State *Item::getGameState() const
{
    return _gamestate;
}

inline int Item::getHeight() const
{
    return _height;
}

inline Uint32 Item::getId() const
{
    return _id;
}

inline Tile::Map *Item::getMap() const
{
    return _map;
}

inline unsigned int Item::getStatusOld() const
{
    return _old_status;
}

inline unsigned int Item::getStatusDiff() const
{
    return _old_status ^ _status;
}

inline int Item::getWidth() const
{
    return _width;
}

inline int Item::getXPos() const
{
    return _xpos;
}

inline int Item::getXPosC() const
{
    return ((_xpos << 1) + _width - 1) >> 1;
}

inline int Item::getYPos() const
{
    return _ypos;
}

inline int Item::getYPosC() const
{
    return ((_ypos << 1) + _height - 1) >> 1;
}


inline int Item::getXVel() const
{
    return _xvel;
}

inline int Item::getYVel() const
{
    return _yvel;
}

inline int Item::getHP() const
{
    return _hp;
}


inline bool Item::isGrounded() const
{
    return _status & GI_FORCED_UP ? true : false;
}

inline void Item::setPos(const Sint32 xpos, const Sint32 ypos)
{
    if(_xpos != xpos) { _xpos = xpos; _dirty |= DIRTY_XPOS; }
    if(_ypos != ypos) { _ypos = ypos; _dirty |= DIRTY_YPOS; }
}

inline void Item::setSequenceHard(const Tile::Sequence *sequence)
{
    assert(sequence != 0);

    _sequence = (Tile::Sequence *) sequence;
    _frameid = 0;
    _framestarttime = _seqstarttime = SDL_GetTicks();
    _dirty |= DIRTY_SEQUENCE;
}

inline void Item::setTypeOff(const unsigned int type)
{
    // Turns off bits of _type that match those of <type>
    _type &= ~type;
}

inline void Item::setTypeOn(const unsigned int type)
{
    // Turns on bits of _type that match those of <type>
    _type |= type;
}

inline void Item::setVel(const Uint32 xvel, const Uint32 yvel)
{
    _xvel = xvel;
    _yvel = yvel;
}

inline bool Item::setVelTowardCoord(const int& targx, const int& targy, unsigned int speed) {
    //
    // Attempts to set my _xvel and _yvel so that we are heading toward
    // coordinate (targx, targy). Moves as fast as <speed> where possible.
    //
    // Returns true if we've reached the target (within granularity of "speed" value),
    // else false.
    //

    bool reached_target = true;

    static int dx, dy;
    if((dx = targx - _xpos)) {
        // move closer to the target x if we're not there yet
        if(((unsigned int) abs(dx)) <= speed)
            _xvel = dx;
        else
            _xvel = targx>_xpos ? speed : -speed;

        reached_target = false;
    } else
        _xvel = 0;

    if((dy = targy - _ypos)) {
        // move closer to the target y if we're not there yet
        if(((unsigned int) abs(dy)) <= speed)
            _yvel = dy;
        else
            _yvel = targy>getYPos() ? speed : -speed;

        reached_target = false;
    } else
        _yvel = 0;

    return reached_target;
}

/*inline void Item::setXAcc(const int xacc)
{
    _xacc = xacc;
}*/

inline void Item::setXPos(const int xpos)
{
    if(xpos != _xpos) {
        _xpos = xpos;
        _dirty |= DIRTY_XPOS;
    }
}

inline void Item::setXVel(const int xvel)
{
    if(xvel != _xvel) {
        _xvel = xvel;
//        _dirty |= DIRTY_XVEL;
    }
}

inline void Item::setXVelThresh(const int xvelthresh)
{
    _xvelthresh = xvelthresh;
}

/*inline void Item::setYAcc(const int yacc)
{
    _yacc = yacc;
}*/

inline void Item::setYPos(const int ypos)
{
    if(ypos != _ypos) {
  
        _ypos = ypos;
        _dirty |= DIRTY_YPOS;
    }
}

inline void Item::setYVel(const int yvel)
{
    if(yvel != _yvel) {
        _yvel = yvel;
//        _dirty |= DIRTY_YVEL;
    }
}

inline void Item::setYVelThresh(const int yvelthresh)
{
    _yvelthresh = yvelthresh;
}


} // namespace Game
} // namespace Salp

#endif
