#include <assert.h>
#include "Game/ItemProjectile.h"
#include "Game/NPC.h"

namespace Salp {
namespace Game {

ItemProjectile::ItemProjectile(const Tile::Sequence *tseq,
               WeaponType wtype, const Uint32 id) : Item(tseq, id)
{
    setHP(1);
    _wtype = wtype;
    _collide_sound = 0;
    _collide_effect = 0;
    _collide_effect_dir = 0;
    _target = 0;
    _set_owner_target_on_hit = false;
    _speed = 1;   // one px at a time please
    _owner = 0; // no one owns me by default
    setType(getType()  & ~GI_IS_AFFECTED_BY_GRAVITY
	                   & ~GI_ON_COLLIDE_LOSE_VEL
					   & ~GI_ON_COLLIDE_TAKE_VEL
                       & ~GI_ON_COLLIDE_REPOS
					   & ~GI_IS_AFFECTED_BY_TRACTION
                       & ~GI_ON_COLLIDE_GIVE_VEL
					   & ~GI_CAN_RIDE_ITEMS
                       & ~GI_USE_VEL_THRESH
                      // & ~GI_CAN_COLLIDE_WITH_ITEM
	        );
}

ItemProjectile::~ItemProjectile()
{
}

void ItemProjectile::animate()
{
    if (_target) {
        Sint32 myx, myy, tx, ty;
        getPosC(myx, myy);
        _target->getPosC(tx, ty);
        int dx  =  tx - myx,
            dy  =  ty - myy;
        int sum = abs(dx) + abs(dy);
        if (sum) {  // avoid floating point exception
            setXVel(_speed * dx / sum);
            setYVel(_speed * dy / sum);
        }
    }
    
}

unsigned int ItemProjectile::getSpeed() const
{
    return _speed;
}

void ItemProjectile::createEffectDirCollided(Tile::Sequence **tseq)
{
    // Note: tseq is 9-element array
    unsigned int dfi = getDirForcedIndex(); // not really true, but good precaution for now.

    if (dfi != 4) // because we might not have element 4 filled.
                  // I guess it'd be keener to do an if (tseq[dfi])
        createEffect(tseq[dfi]);
}


void ItemProjectile::onCollisionItem(Item *victim, unsigned int vic_collide_status)
{
    if(!isOnSameTeamAs(victim)) {
        // It hit!
        if (_collide_sound)
            _collide_sound->playOnce();

        if (_collide_effect)
            createEffect(_collide_effect);

        if (_collide_effect_dir)
            createEffectDirCollided(_collide_effect_dir);

        
        victim->setVel(victim->getXVel() + getXVel()*2, // Do this BEFORE hp modification;
                       victim->getYVel() + getYVel()*2);// else player dies and has extra vel
        victim->decHP(getHP(), _owner, _wtype);
        
        // Set the owner's target on hit, if they asked us to.
        // This is kind of lame, and could be handled in the onHit function
        if (_set_owner_target_on_hit && victim->isNPC()) {
            assert(_owner != 0);
            // set npc's target to victim
            _owner->setTarget((Game::NPC *) victim);
        }
        
        if (_owner) {            
            // Call owner's "onHit" function -- let them handle the event
            // (for example, by noting the damage taken, or doing bonus damage)
            _owner->onHit(victim, this);
        }
        
        getGameState()->delItemDelayed(this);
    }
}

void ItemProjectile::onCollisionMap()
{
    if (_collide_sound)
        _collide_sound->playOnce();

    if (_collide_effect)
        createEffect(_collide_effect);

    if (_collide_effect_dir)
        createEffectDirCollided(_collide_effect_dir);

    // It hit!
    getGameState()->delItemDelayed(this);
}

void ItemProjectile::onOOB(bool x_oob, bool y_oob)
{
    getGameState()->delItemDelayed(this);
}


void ItemProjectile::setCollideEffectDirectional(Tile::Sequence **tseq)
{
    // Note: tseq is assumed to be a 9-item array!
    //       so uh.. don't just give me one sequence or
    //       you're asking for it, mister!

    assert(tseq != 0);
    _collide_effect_dir = tseq;
}

void ItemProjectile::setOwnerTargetOnHit(bool should_set)
{
    assert(_owner != 0);
    assert(_owner->isNPC());
    _set_owner_target_on_hit = should_set;
}

void ItemProjectile::setSpeed(unsigned int speed)
{
     assert(speed >= 0);
     _speed = speed;
}

} // namespace Game
} // namespace Salp
