#include "OSD.h"
#include <cctype>
#include "General/graphics.h"
#include "Game/State.h"
#include "Game/Sock.h"
#include "Game/Server.h"
#include <iomanip>

namespace Salp {
namespace Game {

Tile::Set *OSD::_score_tiles = 0, 
          *OSD::_health_tiles = 0,
          *OSD::_alpha_tiles = 0;


// Class for SW's On-Screen Display
//

    OSD::OSD(Tile::Set *score, Tile::Set *health, Tile::Set *alpha, Game::State *gs, SDL_Surface *screen)
{
    assert(score  != 0);
    assert(health != 0);
    assert(alpha != 0);
    assert(gs != 0);
//    assert(*player != 0);

    _score_tiles  = score;
    _health_tiles = health;
    _alpha_tiles  = alpha;
    _player       = 0;
    _prec = 0;
    _gs = gs;
    _echo_stdout = false;
    _screen = screen;
    _show_radar = false;
}

OSD::~OSD()
{
    // Clear list of console messages
    while (_conmsgs.size() > 0) {
        delete _conmsgs[_conmsgs.size() - 1];
        _conmsgs.pop_back();
    }
}

void OSD::animate()
{
    expireOldMsgs();
}

bool OSD::appendCmdLine(const unsigned int sym)
{
    // Returns true if command line mode is still active;
    // false if we've exited from it

    switch (sym) {
        // Carriage Return
        case 10: case 11: case 12: case 13: case 14: case 15:
            if (_cmdline.length() > 1) // remember, first char is space
                _gs->doCmd(_cmdline.substr(1).c_str());
            _cmdline = "";
            
            return false;
            break;

        // Delete/Backspace
        case 127: case 8:
            _cmdline = _cmdline.substr(0, _cmdline.length() - 1);
            break;

        // Any other key (print it if we like it)
        default:
            if ((sym >= 'A' && sym <= 'Z' ||
                sym >= 'a' && sym <= 'z' ||
                sym >= '0' && sym <= '9' ||
                sym >= ' ' && sym <= '$' ||
                sym >= '&' && sym <= '/' ||
                sym >= ':' && sym <= '@' ||
                sym >= '[' && sym <= '`') && _cmdline.length() < 150)
                _cmdline += sym;
    }
    
    return _cmdline.length() != 0;
} 

void OSD::appendMsg(string msg)
{
   
   ConsoleMsg *newmsg;
   
   // Appends a message to the in-game display

   // Print full message if we should
   if (_echo_stdout) {
       cout << "<" << setw(10) << SDL_GetTicks() << "> " << msg << endl;
   }
   
   // If we know our screen width, then chop the message up for fun
   // (or if not for fun, so that we can read the thing!)
   if (_screen) {
       unsigned int cpr = (_screen->w - 4)/6; // characters per row
       while (msg.length() > cpr) {
           newmsg = new ConsoleMsg;
           newmsg->msg = msg.substr(0, cpr);
           _conmsgs.push_back(newmsg);
           msg = "    " + msg.substr(cpr);
       }
   }
   
   // Print remainder, or full message
   newmsg = new ConsoleMsg;
   newmsg->msg = msg;
   _conmsgs.push_back(newmsg);
   
}

void OSD::drawConsoleMsgs(SDL_Surface *screen)
{
    static Uint32   msgbkgdcolor  = SDL_MapRGB(screen->format, 51, 51, 51*4);
    static Uint32   cmdbkgdcolor  = SDL_MapRGB(screen->format, 51*4, 0, 0);
    static SDL_Rect dstrect;
    
    unsigned int msgcount = _conmsgs.size();

    // Draws messages on top of console
    if(msgcount) {
        Uint32 end_of_world=10000 - (SDL_GetTicks() - _conmsgs[msgcount - 1]->time_sent);
        // Draw background color
        dstrect.x=2;
        dstrect.h=6 * msgcount + 1; 
        dstrect.y=7;
        dstrect.w=screen->w-4;
        Uint8 msg_alpha = end_of_world > 250 ? 64 : end_of_world*64/250;
        General::fillRectAlpha(screen, &dstrect, msgbkgdcolor, msg_alpha);
        dstrect.y += 2;
        dstrect.x += 2;
        General::fillRectAlpha(screen, &dstrect, msgbkgdcolor, msg_alpha);
    
        int y = 2;
        for (unsigned int i=0; i<msgcount; ++i) {
            drawText5x5(screen, 5, y += 6, (char *) _conmsgs[i]->msg.c_str(), _conmsgs[i]->msg.length());
        }
    }
    
    // Draws command being typed, if any
    if (_cmdline.length() != 0) {
        dstrect.x=2;
        dstrect.h=7;
        dstrect.y=0;
        dstrect.w=screen->w - 4;
        General::fillRectAlpha(screen, &dstrect, cmdbkgdcolor, 128);
        dstrect.y += 2;
        dstrect.x += 2;
        General::fillRectAlpha(screen, &dstrect, cmdbkgdcolor, 128);

        drawText5x5(screen, 5, 1, "#", 1);
        unsigned int cpr = (_screen->w - 4)/6 -2; // characters per row
        
        // Draw differently depending on length of command being typed
        // (That is, we want to scroll it so that the player can see the last part of the 
        // message typed if they typed a lot of it.. sure, it's not wrapping, but it's hella-easier :-))
        if (_cmdline.length() <= cpr) {
            drawText5x5(screen, 11, 1, (char *) _cmdline.c_str(), _cmdline.length());
            if (_cmdline.length() < 150)  drawText5x5(screen, 11 + 6*_cmdline.length(), 1, "_", 1);
        } else { 
            drawText5x5(screen, 11, 1, (char *) _cmdline.substr(_cmdline.length()-cpr).c_str(), cpr);
            if (_cmdline.length() < 150)  drawText5x5(screen, 11 + (cpr)*6, 1, "_", 1);
        }
    } 
}

void OSD::drawRadar(SDL_Surface *screen)
{
    assert(_gs != 0);
    assert(_player != 0);
    assert(screen != 0);
    static Uint32   bkgdcolor  = SDL_MapRGB(screen->format, 51, 51, 51*4);
    static Uint32   evildotbkgdcolor  = SDL_MapRGB(screen->format, 51*4, 51, 51),
                    gooddotbkgdcolor  = SDL_MapRGB(screen->format, 51, 51, 51*5);
    
    
    Tile::Map *tm = _player->getMap();
    static SDL_Rect dstrect, dotdstrect;
    int mh = tm->getPixHeight(),
        mw = tm->getPixWidth();
    int maxwh= mh>mw?mh:mw; // max width or height
    
    // map is max of 100 * 100
    dstrect.h = 50 * mh/maxwh + 20; // obscure it a little
    dstrect.w = 50 * mw/maxwh + 20;
    dstrect.x = screen->w - dstrect.w - 5;
    dstrect.y = 5;
    
    // Draw radar background
    General::fillRectAlpha(screen, &dstrect, bkgdcolor, 64);
    
    
    // Draw blips on radar
    dotdstrect.w = 2;
    dotdstrect.h = 2;
    map<Uint32, Game::NPC *>::iterator gi;
    for (gi=_gs->_items_npc.begin(); gi != _gs->_items_npc.end(); ++gi) {
        dotdstrect.y = dstrect.y + dstrect.h * gi->second->getYPosC()/mh;
        dotdstrect.x = dstrect.x + dstrect.w * gi->second->getXPosC()/mw;
        
        SDL_FillRect(screen, &dotdstrect, _player->isOnSameTeamAs(gi->second) ? gooddotbkgdcolor : evildotbkgdcolor);
    }
}

void OSD::drawText5x5(SDL_Surface *screen, int x, int y, char * text, int len)
{
    if (len == -1) 
        len = strlen(text);
    
    
    assert(len > 0);
    assert(len < 1000); // keep it reasonable :-)
    assert(_alpha_tiles != 0);
    int boty = y+4,      // inclusive
        botx = x+len*6-2;
        
    // if totally off screen, die
    if (x >= screen->w || botx < 0 || y >= screen->h || boty < 0)
        return; 
        
    // don't draw chars that wont appear on the leftmost x
//    if (x + _alpha_tiles->getWidth() <= 0) {
//        int cskip = -x/6;
//        text += cskip;
//        len += cskip;
//        x += 6 * cskip;
//        cout << "X changed to " << x << " (skip " << cskip << endl;
//    }
    
    while(len && x < screen->w) {
        assert(x < screen->w);
        if(x + _alpha_tiles->getWidth() > 0 &&
           y + _alpha_tiles->getHeight() > 0) {
            assert(x + _alpha_tiles->getWidth() > 0);
            assert(y + _alpha_tiles->getHeight() > 0);
            _alpha_tiles->drawImage(screen, toupper(*text) - '!', x, y);
//            _alpha_tiles->drawImage(screen, *text - '!', x, y);
        }
        
        ++text; --len;
        x += 6; // 5 for character, and 1 for spacing
    }
}

void OSD::drawImage(SDL_Surface *screen)
{
    if(!_player || !_prec)   return; 

    assert(_player != 0);
    assert(_prec != 0);

//    static int base_time = SDL_GetTicks();
    

//    drawScore(screen,(SDL_GetTicks() - base_time) / 1000, 0 + 10 + 3, screen->h - _score_tiles->getHeight() - 10 + 3, 128 );
//    drawScore(screen,(SDL_GetTicks() - base_time) / 1000, 0 + 10, screen->h - _score_tiles->getHeight() - 10 );
    if (_show_radar)  drawRadar(screen);
    drawScore(screen,(_prec->score), 0 + 10 + 3, screen->h - _score_tiles->getHeight() - 10 + 3, 128 );
    drawScore(screen,(_prec->score), 0 + 10, screen->h - _score_tiles->getHeight() - 10 );
    
    drawConsoleMsgs(screen);

    // health
/*    static int hpct = 0;
    int x=10, y=screen->h-35;
    for(int i=0, tid=0;
        i<=hpct;
        i+= 100/_health_tiles->getCount()/3, ++tid, y-=_health_tiles->getHeight()) {
        _health_tiles->drawImage(screen, tid/3, x+3, y+3, 128);
        _health_tiles->drawImage(screen, tid/3, x, y);
    }
    hpct = (hpct + 1) % 101;
*/

}

void OSD::drawScore(SDL_Surface *screen, int score, int x, int y, int alpha)
{
    int digits = !score?1:int(log((float) score) / log((float) 10)) + 1;
    //cout << "Score: " << score << ", digits " << digits << endl;
    if(score < 0) { ++digits; score = -score; }
    x += (digits-1) * _score_tiles->getWidth();

    while(digits--) {
        _score_tiles->drawImage(screen, score % 10, x, y, alpha);
        score /= 10;
        x -= _score_tiles->getWidth();
    }
}

void OSD::expireOldMsgs()
{
    // Removes from _conmsgs the old messages on the screen.

    int msg_count = _conmsgs.size();

    // Expire based on message count (We don't want to overflow
    // the screen with icky blits).
    while (msg_count > 10) {
        delete _conmsgs[0];
        _conmsgs.pop_front();
        --msg_count;
    }

    Uint32 min_survival_time = SDL_GetTicks(); 
    if (SDL_GetTicks() < 10000) return; // 10 sec
    min_survival_time -= 10000;
    
    // Expire based on time
    while (msg_count > 0 && _conmsgs[0]->time_sent < min_survival_time) {
        delete _conmsgs[0];
        _conmsgs.pop_front();
        --msg_count;
    }
    
}

void OSD::setCmdLine(const string& cmd)
{
    // Sets effective command line string to <cmd>
    _cmdline = ' ' + cmd;
}

void OSD::setPlayer(Game::Player *player)
{
    assert(player != 0);
    assert(player->isNPC());
    assert(_gs != 0);
    _player = player;
    _prec   = _gs->getScoreboard()->getPlayerRecGuaranteed(player->getId());
    
}


} // namespace Game
} // namespace Salp

