#include <cassert>
#include <iostream>

#include "Game/Player.h"
#include "Game/ItemPowerUp.h"
#include "General/media.h"

using namespace General;

namespace Salp {
namespace Game {

Player::Player(const Tile::Sequence *tseq,
               const Uint32 id) : NPC(tseq, id)
{
    // initialize key states (set all to 0)
    unsigned int i;
    for(i=0; i<MAX_KEYS; i++)
        _keys[i] = false;
        
    setTeam(Game::State::getUniqueTeamId());
    
    _hp = _maxhp = 0;
    
    _active_weapon = 0;
    
    _pdata = 0;
}

Player::~Player()
{
    while (_weapons.size() > 0) {
        assert(_weapons[_weapons.size()-1] != 0);
        delete _weapons[_weapons.size()-1];
        _weapons.pop_back();
    }
}

void Player::addAmmo(WeaponStatus *ws)
{
    addAmmo(ws->type, ws->ammo_count, ws->has_weapon);
}

void Player::addAmmo(WeaponType t, int ammo_count, bool has_weapon) {
    // Adds <ammo> to ammo count of weapon type t
    // Adds weapon to _weapons vector if it's not there already.
    // If has_weapon is true, then the player will also be given the weapon
    // (not just the record of ammo).
    
    WeaponStatus *ws = 0;
    
    // Find weapon if it already exists
    for (unsigned int i=0; i<_weapons.size(); ++i)
        if (_weapons[i]->type == t) {
            ws = _weapons[i];
            break;
        }
        
    // If it doesn't already exist, add it to our array
    if (!ws) {
        ws = new WeaponStatus(t);
        _weapons.push_back(ws);
    }
        
    assert(ws != 0);

    // Finally, add stats
    ws->ammo_count += ammo_count;
    if (has_weapon) { 
        ws->has_weapon=true;
        if (!canFireCurrentWeapon() && selectNextWeapon()) {
            appendConsoleMsg("Selected " + _weapons[_active_weapon]->getName() + ".");
        }
    }
}

bool Player::canFireCurrentWeapon()
{
    // returns true if user has a weapon (not just ammo) selected,
    // and it has ammo loaded
    
    assert(_active_weapon >= 0);
    if (_weapons.size() == 0) return false;
    
    assert(_active_weapon < _weapons.size());
    return (_weapons[_active_weapon]->has_weapon &&
            _weapons[_active_weapon]->ammo_count > 0);
}

void Player::doCmd(string cmd)
{
    if (cmd == "ammo") {
        addAmmo(Game::WEAPON_SHOTGUN, 100, true);
        addAmmo(Game::WEAPON_SPREAD, 100, true);
        addAmmo(Game::WEAPON_SALP, 100, true);
        addAmmo(Game::WEAPON_HOMING, 100, true);
                
        getGameState()->appendConsoleMsgAll("# CHEAT: " + getName() + " #  has all weapons/ammo.");
    } else if (cmd == "reset scores") {
        getGameState()->getScoreboard()->reset(false);
        getGameState()->appendConsoleMsgAll("# CHEAT: " + getName() + " #  reset the score.");
    } else {
        appendConsoleMsg("Invalid command.");
    }
}

void Player::drawImage(SDL_Surface *screen,
                           const Sint32 topx,
                           const Sint32 topy)
{
    NPC::drawImage(screen, topx, topy);
    drawHealthBar(screen, topx, topy);
    if (_name.length() > 0)
        drawName(screen, topx, topy);
}

void Player::fireWeaponCurrent(const bool primary) 
{
    // **** AUTO-SELECT Possible Weapon to Fire ****
    if (!canFireCurrentWeapon() && selectNextWeapon()) {
        appendConsoleMsg("Selected " + _weapons[_active_weapon]->getName() + ".");
    }

    if (_weapons.size() == 0) {
        appendConsoleMsg("You need a weapon first.");
        return;
    }
    
    assert(_active_weapon >= 0);
    assert(_active_weapon < _weapons.size());
    assert(_weapons[_active_weapon] != 0);
    
    if (!_weapons[_active_weapon]->has_weapon) {
        appendConsoleMsg("You have ammo, but no weapon.");
        return;
    }

    // Note that the next line auto-decrements ammo count
    if (_weapons[_active_weapon]->ammo_count-- <= 0) {
        appendConsoleMsg("It has no ammo left.");
        return;
    }
    
    fireWeapon(_weapons[_active_weapon]->type, primary, _keys);
}

bool Player::getKeyState(const unsigned int& keyId) const
{
    assert(keyId > 0);
    assert(keyId < MAX_KEYS);
    
    return _keys[keyId];
}


void Player::setKeyState(const unsigned int& keyId, const bool& isDown)
{
    
//    cout << "Player " << (isDown?"pressed ":"released ") << " key " << keyId
//         << endl;
    assert(keyId < MAX_KEYS);
        
    // Don't try to access memory that's not there =)    
    if(keyId > MAX_KEYS) return;
    
    // Set key state to the new value.
    _keys[keyId] = isDown;

    // Update facedness
    if(_keys[KEY_UP])         faceUp();
    else if(_keys[KEY_DOWN])  faceDown();

    if(_keys[KEY_LEFT])       faceLeft();
    else if(_keys[KEY_RIGHT]) faceRight();
}

bool Player::selectNextWeapon(bool reverse) {
    // Selects the next weapon in the chain; returns true if another
    // weapon is found (and that next weapon is selected). Returns false
    // if no other weapon can be found with ammo in it.
    //
    unsigned int last_active = _active_weapon;
    
    for (int i=_weapons.size() - 2; i >= 0; i--) {
        _active_weapon = (_active_weapon + 1) % _weapons.size();
        if (_weapons[_active_weapon]->has_weapon && 
            _weapons[_active_weapon]->ammo_count > 0)
                return true;
    }

    _active_weapon = last_active;
    
    return false;
}

} // namespace Game
} // namespace Salp
