#include "PrefBucket.h"

#include <fstream>
#include <iostream>
#include <cstdio>
#include <cassert>

using namespace std;

namespace Salp {
namespace Game {

PrefBucket::PrefBucket()
{
    _error_count = 0;
}

PrefBucket::PrefBucket(string fpath)
{
    PrefBucket(); // do standard initialization
    
    _fpath = fpath;
    revertToSaved();
}

void PrefBucket::erase()
{
     // Esrases all pref key/value pairs stored in object;
     // even the defaults!
     
    _prefs.erase(_prefs.begin(), _prefs.end());
}

int PrefBucket::getIntVal(const string key)
{
    if(_prefs.find(key) == _prefs.end())
        return 0;
    else
        return atoi(_prefs[key].c_str());
}

int PrefBucket::getIntVal(const string key, const string default_val)
{
    // If <key> already exists, the corresponding value is returned.
    // if not, the key is created, with a value of <default_val> and
    // then <default_val> is returned.
    
    if(_prefs.find(key) == _prefs.end()) 
        _prefs[key] = default_val;
     
    return atoi(_prefs[key].c_str());
}

string PrefBucket::getVal(const string key, const string default_val) 
{
    // If <key> already exists, the corresponding value is returned.
    // if not, the key is created, with a value of <default_val> and
    // then <default_val> is returned.

    if(_prefs.find(key) == _prefs.end()) 
        _prefs[key] = default_val;
     
    return _prefs[key].c_str();
}

string PrefBucket::getVal(const string key)
{
    if(_prefs.find(key) == _prefs.end())
        return "";
    else
        return _prefs[key];
}


bool PrefBucket::hasErrors()
{
    return _error_count > 0;
}

void PrefBucket::logError(string msg)
{
    _error_count++;
    cerr << msg << endl;
}

void PrefBucket::parseOpts(char **argv, int argc) 
{
    char urlip[200], urlport[10] = "poop";

    // For each option, error or set up key/val pairs
    for(int i=1; i<argc; ++i) {
        // parse an option
        if (sscanf(argv[i], "salp://%190[^:]:%5[0-9]*", urlip, urlport) == 2) {
            // Catch url format of salp://host:port/ as client connect url
            _prefs["host"] = urlip;
            _prefs["port"] = urlport;
            _prefs["mode"] = "client";
        } else {
            // find first half (key goes from index [0, fi)  )
            string opt = argv[i];
            int fi = 0;
            while(opt[fi] != '=' && opt[fi] != 0) ++fi;
        
            if(opt[fi] == 0 || opt[fi] != '=' || opt[fi+1] == 0)
                logError("\"" + opt + "\" is not a validly formatted option.\n");
            else
                // set key = val
                _prefs[opt.substr(0, fi)] = opt.substr(fi + 1);
        }
    }
}

bool PrefBucket::revertToSaved()
{
    // Reverts to saved version of prefbucket
    assert(_fpath.length() != 0);
    
    // Open the file
    ifstream fin;
    fin.open(_fpath.c_str());
    if(!fin.is_open()) {
        cerr << "Warning: could not open preference file " << _fpath << " for reading. Using default values." << endl;
        return false;
        //exit(1);
    }

    char keystr[101], valstr[200], currline[200];
    int  lineno = 0;
    
    while(!fin.eof() && fin.good())
    {
        // get the next line of the file
        ++lineno;
        fin.getline(currline, 200-1);

        // skip comments
        if(sscanf(currline, " # %198s", valstr) ||
           !strlen(currline))
            continue;
            
        // otherwise get token name
        if(sscanf(currline, " %100[^ \t] %198[^\r] ", keystr, valstr) != 2)
        {
            cout << "PrefBucket: Unrecognized Line " << lineno << ": " << currline << endl;
        } else {
            _prefs[keystr] = valstr;
        }
    }
    
    if(!fin.good() && !fin.eof()) {
        cout << "An error occurred when reading the level file, at line "
             << (lineno)
             << "." <<endl;
        exit(1);
    }    

    fin.close();
    
    return true;
}

void PrefBucket::save()
{
    assert(_fpath.length() != 0);

    // Open keyboard map file
    ofstream fout;
    fout.open(_fpath.c_str(), ios::out );
    if(!fout.is_open()) {
        cerr << "Error opening file " << _fpath << " for writing." << endl;
        exit(1);
    }
    
    // Cycle through keyboard map data, and save it!
    map<string, string>::iterator i;
    for(i=_prefs.begin(); i != _prefs.end(); i++) {
        assert(i->first.find("\n") == string::npos);
        assert(i->second.find("\n") == string::npos);
        assert(i->first.find("\r") == string::npos);
        assert(i->second.find("\r") == string::npos);
        fout << i->first;
        for (unsigned int j=i->first.length(); j<20; ++j)
            fout << ' ';
        fout << ' ' << i->second << "\n";
    }
    
    fout.close();
}

void PrefBucket::setDefault(string key, string val)
{    if(_prefs.find(key) == _prefs.end())
        _prefs[key] = val;
}

void PrefBucket::setPath(string fpath) 
{
    // Sets filepath where preference file can be found.
    // This lets you use the save() and revertToSaved() functions.
    _fpath = fpath;
}

void PrefBucket::setVal(string key, string val)
{
    _prefs[key] = val;
}

void PrefBucket::setVal(string key, int val)
{
    char tmp[20];
    sprintf(tmp, "%d", val);
    _prefs[key] = tmp;
}


} // namespace Game
} // namespace Salp
