#include "Scoreboard.h"
#include <cassert>
#include <ctime>
#include "Game/Item.h"
#include "General/general.h"
#include "General/graphics.h"
#include "Game/NPC.h"
#include "Game/Server.h"
#include "Game/State.h"
#include "Game/OSD.h"
#include "General/media.h"

using namespace Salp::General;

namespace Salp {
namespace Game {

char * PlayerRec::writeDataUpdate()
{
    writeDataUpdate(&General::tmp_data[0]);
    return &General::tmp_data[0];
}

void PlayerRec::writeDataUpdate(char *data)
{
    SET_SMSG_TYPE(data, MS_UPDATE_PLAYER_SCORE);
    SET_SMSG_LEN(data, 4 + 6); // dont' forget to change length if you add more
    data += 4; // get to data segment
    DATA_ICOPY_32(data, id);
    DATA_ICOPY_16(data, score);
}

// *****************************************************************

Scoreboard::Scoreboard(Game::State *gs)
{
    assert(gs != 0);
    _gs = gs; // We want to see the state too!
    
    reset();
}

Scoreboard::~Scoreboard()
{
    reset();
}

void Scoreboard::addEvent(DamageRec *event) {
    // Add to event history
    _events.push_back((ScoreRec *) event);
    
    // And interpret them!
    PlayerRec *vrec = getPlayerRec(event->_victim_id),
              *krec = getPlayerRec(event->_killer_id);
    
    
    Uint32 curr_time = SDL_GetTicks();
    
    if (debugScoreboard)
        cout << "Got DamageRec event; Victim: " << event->_victim_id << (vrec?"":" *NOT RECORDING*")
             << ", " << event->_killer_id << (krec?"":" *NOT RECORDING*") << "; " << (event->_fatal?"FATAL":"NOT fatal") << "\n";
         
    // Note basic damage stuff (fatal or nonfatal)
    // ### NOTE: event->_wtype not accounted for yet!!
    if (event->_damage >= 0) {
        if (vrec) vrec->dmg_taken_total += event->_damage;
        if (krec) krec->dmg_given_total += event->_damage;
    } else {
        if (vrec) vrec->healing_taken_total -= event->_damage;
        if (krec) krec->healing_given_total -= event->_damage;
    }
    
    // Did the victim die?
    if (event->_fatal) {
        Game::NPC *killer = (Game::NPC *) _gs->getItemMaybe(event->_killer_id),
                  *victim = (Game::NPC *) _gs->getItemMaybe(event->_victim_id);
        
        assert(!killer || killer->isNPC());
        assert(!victim || victim->isNPC());
        
        if (vrec) {
            vrec->death_count_total++;
            
            // Is this the end of victim's salpfest?
            if (vrec->kill_count_nodeath >= 5 && killer && victim) {
                _gs->appendConsoleMsgAll(killer->getName() + " has ended " + victim->getName() + "'s salpfest.");
            }

            // Reset death-related kill counters
            vrec->kill_count_nodeath = 0;
            vrec->kill_count_rapid   = 0;
            vrec->death_time_last    = curr_time;            
            
            if (_gs->_server)
                _gs->_server->sendAllState(vrec->writeDataUpdate());
        }
        
        if (krec) {
            krec->setScore(krec->score + 1);
            krec->kill_count_total++;
            krec->kill_count_nodeath++;
            
            if ((curr_time - krec->kill_time_last) <= 5000) { // msec
                krec->kill_count_rapid++;

                // insert multikill handlers here
                if (krec->kill_count_rapid > 1) {
                    int multikill_type = krec->kill_count_rapid - 2;
                    if (multikill_type > 3) multikill_type = 3;
                    General::sound_multi_kill[multikill_type]->playOnce();
                    //_gs->appendConsoleMsgAll(killer->getName() + " has multi-salped.");
                }
            } else
                krec->kill_count_rapid=1;
            
            if (!(krec->kill_count_nodeath % 5)  && killer ) {
                _gs->appendConsoleMsgAll(killer->getName() + " is going on a salpfest!");
            } 
                
            krec->kill_time_last = curr_time;
            
            if (_gs->_server)
                _gs->_server->sendAllState(krec->writeDataUpdate());
        }
        
        // Someone died, so maybe the game ended. Let's check it out.
        hasGameEnded();
        
        // Maybe I should send a cue to the clients to display the updated
        // score? Well, we send them the updates anyway; maybe we can let
        // them worry about that. Besides, it gets in the way of gameplay.
        // UT and other games show the scores after death, when the player
        // has time and room to view it.
    }
    
}

void Scoreboard::addPlayer(Uint32 id)
{
    assert(_players.find(id) == _players.end());
    PlayerRec *newrec = new PlayerRec(id);
    assert(newrec != 0);
    _players[id] = newrec;
    
    // Let's update everyone's info on the new player
    assert(_gs != 0);
    if (_gs->_server) {
        char data[512];
        newrec->writeDataUpdate(&data[0]);
        _gs->_server->sendAllState(data);    
    }
}

void Scoreboard::drawScoreSummary(SDL_Surface *screen)
{
    // Draw headers
    map<Uint32, PlayerRec *>::iterator i;
    static int last_player_count=1;
    
    assert(last_player_count >= 0); // or else weirdness?
    
    
    
    Game::NPC *gi;
    int x = 50;
    int y = 60;

    int cell_padding=10;
    
    // draw pretty background
    static SDL_Rect dstrect;
    static Uint32   bkgdcolor = SDL_MapRGB(screen->format, 0, 128, 128);
    dstrect.x=x-cell_padding;
    dstrect.h=5 + (last_player_count-1)*10 + cell_padding*2;
    dstrect.y=y-cell_padding;
    dstrect.w= 35 * 6 + cell_padding*2; // don't care about x-padding?? the 5 is arbitrary
    General::fillRectAlpha(screen, &dstrect, bkgdcolor, 128);
    dstrect.y += 2;
    dstrect.x += 2;
    General::fillRectAlpha(screen, &dstrect, bkgdcolor, 128);

    int active_player_count = 0;
    for (i=_players.begin(); i != _players.end(); i++) {
        if ((gi = (Game::NPC *) _gs->getItemMaybe(i->first)) != 0   &&  gi->isNPC()  && gi->getName().length()>0) {
            OSD::drawText5x5(screen, x, y, (char *) gi->getName().c_str());
            OSD::drawText5x5(screen, x + 6*30, y, (char *) i->second->scorestr);
            y += 10;
            ++active_player_count;
        }
    }
    
    last_player_count = active_player_count;
}

PlayerRec * Scoreboard::getPlayerRec(Uint32 id) 
{
    //assert(_players.find(id) != _players.end());
    return _players.find(id) == _players.end() ? 0 : _players[id];
}

PlayerRec * Scoreboard::getPlayerRecGuaranteed(Uint32 id) 
{
    // returns pointer to player rec, creating a new one if necessary
    return _players.find(id) == _players.end() ? (_players[id] = new PlayerRec(id)) : _players[id];
}

string Scoreboard::getScoreString()
{
    // Builds and returns a string representing
    // the score standings.
    string tmp = "Score:";
    
    map<Uint32, PlayerRec *>::iterator i;
    Game::Item *gi;
    for (i=_players.begin(); i != _players.end(); i++) {
        if ((gi = _gs->getItemMaybe(i->first)) != 0   &&  gi->isNPC()) {
            tmp += " " + (((Game::NPC *) gi)->getName()) + ": " + itoa(i->second->score);
        }
    }
    
    return tmp;
}

Uint32 Scoreboard::getTimeElapsed()
{
    assert(time_start != 0);  // check isTimerStarted first
    return time(0) - time_start;
}

Uint32 Scoreboard::getTimeLeft()
{
    assert(time_max != 0);
    assert(time_start != 0);
    return time_max - time_start;
}

bool Scoreboard::hasGameEnded()
{
    if (_has_game_ended)  return true;
    
    // otherwise, check if the game really did end,
    // and set _has_game_ended to true if it did.
    
    int _winning_individual_score = 2;
    
    PlayerRec * current_player; 
    // Check individual scores
    map<Uint32, PlayerRec *>::iterator pi;
    for (pi=_players.begin(); pi != _players.end(); ++pi) {
        assert(pi->second != 0);
        current_player = pi->second;
        if (_winning_individual_score > 0 &&
            current_player->score >= _winning_individual_score)
            _has_game_ended = true;
    }

    return _has_game_ended;
}

bool Scoreboard::isTimerStarted() {
    return time_start != 0;
}

void Scoreboard::reset(bool delete_players)
{
    time_start = 0; // game hasn't started yet
    time_max   = 0; // no max time defined
    _has_game_ended = false;
    
    // Clear event list
    unsigned int i = _events.size();
    while (i--) {
        delete _events[i];
        _events.pop_back();
    }
    
    // Clear player list
    map<Uint32, PlayerRec *>::iterator pi;
    if (delete_players) {
        while ((pi = _players.begin()) != _players.end()) {
            delete pi->second;
            _players.erase(pi);
        }
    } else {
        for (pi=_players.begin(); pi != _players.end(); ++pi) {
            assert(pi->second != 0);
            pi->second->resetScore();
            _gs->_server->sendAllState(pi->second->writeDataUpdate());
        }
    }
}

void Scoreboard::setTimeLeft(Uint32 sec)
{
    time_max = time_start + sec;
}

void Scoreboard::startTimer(Uint32 max_length_sec)
{
    time_start = time(0);
    time_max   = max_length_sec==0?0:time_start+max_length_sec;
}

void Scoreboard::updatePlayerData(char *data) {
    assert(SMSG_ELEN(data) == 6);
    assert(SMSG_TYPE(data) == MS_UPDATE_PLAYER_SCORE);
    
    data += 4; // get to data segment
    
    // Read ID and retrieve record (if possible)
    Uint32 id;
    DATA_IREAD_32(data, id);
    PlayerRec *prec = getPlayerRecGuaranteed(id);
    
    DATA_IREAD_16(data, prec->score);


    prec->updateScoreStr();
}

void Scoreboard::writeDataUpdateFull(char *dptr, int &len) {
    len=0;
    map<Uint32, PlayerRec *>::iterator pi;
    
    for (pi=_players.begin(); pi != _players.end(); ++pi) {
        pi->second->writeDataUpdate(dptr);
        len += SMSG_LEN(dptr);
        dptr += SMSG_LEN(dptr);
        assert(len <= 1024);
    }

    return;
}

}; 
};
