#ifndef __GAMESCOREBOARD_H
#define __GAMESCOREBOARD_H

#include <map>
#include <vector>
#include <string>
#include <ctime>
#include <cassert>

#include "SALP_SDL.h"
#include "SALPSTUB.h"

using namespace std;

namespace Salp {
namespace Game {
    
    /*
    
       The goal of the scoreboard is to:
       
       Track Individual Scores
           - 5 points for kills of flagholders
           - 3 points for kills of "killing spree"ers
           - 1 point for all other kills
       
       Track Individual Special Instances
           - Track instances of killing sprees (ie, kills w/o dying)
           - Track instances of double/multi-kills
           
       Track Team Scores
           - Aggregate individual scores into team scores.
           - Track flag captures.       
       
       Track Level Goals
           - Game ends when an individual reaches score N
           - Game ends when a team reaches score N
    
    
       NOTE: in Event stuff, consider using Item's ID instead? Unless
             the ID isn't guaranteed to be unique. In which case, ouch.
    */
    enum { SCORERECTYPE_DAMAGE=1 };
    struct PlayerRec {
        PlayerRec(Uint16 pid) {
            // set up naming stuff
            strcpy(scorestr, "    0");
            id                  = pid;
            
            // and clear mr. scorey score
            resetScore();
        };
        
        void resetScore() {
            death_count_total   = 0;
            kill_count_total    = 0;
            kill_count_nodeath  = 0;
            kill_count_rapid    = 0;
            kill_time_last      = 0;
            score               = 0;
        }
        
        void setScore(int newscore)  {
            score = newscore;
            updateScoreStr();
        };
        
        void updateScoreStr() {  
            assert(abs(score) < 1000000);
            sprintf(scorestr, "%5d", score);
        };
        
        char * writeDataUpdate();
        void writeDataUpdate(char *data);
        
        // Properties (starred properties mean "updated to clients")
        
        Uint16  death_count_total;
        Uint32  death_time_last;
        
        Uint32  dmg_taken_total;      //
        Uint32  dmg_given_total;      //

        Uint32  healing_taken_total;      //
        Uint32  healing_given_total;      //
        
        
        Uint16  kill_count_total;     //   Total kills made by player, ever
        Uint8   kill_count_nodeath;   //         kills since last death
        Uint8   kill_count_rapid;     //         kills connected by last 2 sec interval

        Uint32  kill_time_last;       //   Tickcount at time of last kill
        Sint16  score;                // * Aggregate score
        char    scorestr[20];
        
        Uint32  id;

    };
    
    class ScoreRec {
        public:
            Uint8 getType();
        
        protected:
            Uint8 _t;
    };

    class DamageRec : public ScoreRec {
        
        public:        
            DamageRec(Uint32 victim_id, WeaponType wtype, Sint16 damage, bool fatal, Uint32 killer_id=0)
            {
                _t = SCORERECTYPE_DAMAGE;
                _killer_id = killer_id;
                _victim_id = victim_id;
                _fatal = fatal;
                _damage = damage;
                _time = time(0);
                _wtype = wtype;
            };
            
            
        
            Uint32     _killer_id, _victim_id;
            bool       _fatal;
            Sint16     _damage;
            Uint32     _time; // seconds, not msec
            WeaponType _wtype;
    };
        
	class Scoreboard {
	    public:
			Scoreboard(Game::State * gs);
            ~Scoreboard();
            
            void addEvent(DamageRec * event);
            void addPlayer(Uint32 id);
            bool hasGameEnded(); // checks if game has ended, returns true if so.
            void drawScoreSummary(SDL_Surface *screen);
            PlayerRec * getPlayerRec(Uint32 id); // Warning: may return 0 (null pointer)
            PlayerRec * getPlayerRecGuaranteed(Uint32 id); // never is 0, always succeeds (cuz auto-creates)
            string getScoreString();
            Uint32 getTimeElapsed(); // Number of seconds the game has lasted so far
            Uint32 getTimeLeft();
            bool isTimerStarted();
            void reset(bool delete_players=true);
            void setTimeLeft(Uint32 sec);
            void startTimer(Uint32 max_length_sec=0);
            void updatePlayerData(char *msg);
            void writeDataUpdateFull(char *dptr, int &len);
            
        private:
            vector<ScoreRec *>       _events;
            map<Uint32, PlayerRec *> _players;
            bool _has_game_ended;
            
            Uint32 time_start; // measured in seconds (epoch time from time(0))
            Uint32 time_max;   // max time(0) that the game can last through
            Game::State *_gs;
	};

} // namespace Game
} // namespace Salp

#endif
