#include "Sock.h"

namespace Salp {
namespace Game {


PeerData::PeerData()
{
    wants_game_state = false;
    player           = 0;
    peerid           = 0;
    _has_error       = false;
    tcpsock          = 0;
    udpsock          = 0;
    udpchan          = -1;
    udpport          = 0;
}

PeerData::~PeerData()
{
    
    assert(player == 0); // Make sure it's been properly deleted, or we dont have one
    if (tcpsock) 
        SDLNet_TCP_Close(tcpsock);
    if (udpsock)
        SDLNet_UDP_Close(udpsock);
};


// message type, data, message length
int PeerData::sendMsg(Uint16 mtype, const char *data, Uint16 len) {
    assert(mtype != 0);
    assert(tcpsock != 0);
    
    len += 4;
    char *buf = new char[len]; 
    char *writehead = buf;
    
    // package it -- may be unneccessary to buffer once
    // our own sockets are buffered
    DATA_ICOPY_16(writehead, len);
    DATA_ICOPY_16(writehead, mtype);
    DATA_ICOPY_N(writehead, data, len - 4);
    
    // send it
    int chars = SDLNet_TCP_Send(tcpsock, (void *) buf, len);
    if(chars != len)          setError("Error sending message");
    
    delete[] buf;
    
    return 1; // ### THIS SHOULD CHANGE
}

void PeerData::sendMsg(Uint16 mtype, string data) {
   assert(mtype != 0);
   sendMsg(mtype, data.c_str(), data.length());
}

void PeerData::sendMsg(const char * val, Uint16 len) {
    assert(tcpsock != 0);
    int chars = SDLNet_TCP_Send(tcpsock, (void *) val, len);

    if(chars != len)          setError("Error sending message");
}

void PeerData::sendMsgRaw(const char * val, Uint16 len) {
    assert(tcpsock != 0);
    assert(SMSG_TYPE(val) != 0);
// replace with integrity check    assert(SMSG_LEN(val) == len);
    int chars = SDLNet_TCP_Send(tcpsock, (void *) val, len);
    if(chars != len)          setError("Error sending message");
}

void PeerData::setError(string msg)
{
//    cout << "PeerData error: " << msg << endl;
    _has_error = true;
}



//////////////////////////////////////

char * SockBuff::getEvent() {
    // Checks whether or not we have a full usable message yet.
    // If we do, a pointer to the message stream is returned. 
    // Otherwise, returns 0

    if(_bufsize >    sizeof(Uint16) &&
       _bufsize >=   SMSG_LEN(_buf) ) {
//        cout << "getEvent says: TYPE: " << SMSG_TYPE(_buf) << "; LEN: " << SMSG_LEN(_buf) << "; _bufsize: " << _bufsize << endl;
        assert(SMSG_LEN(_buf) == _bufsize); // Hope we didn't overslurp
        // we have a message!
        return _buf;
    }

    return 0;
};

bool SockBuff::readSock(TCPsocket sock)
{
    // reads from sock into stream.
    // returns false if user disconnected.

    assert(!getEvent()); // we shouldn't already have a message

    if(_bufsize < sizeof(Uint16)) {
        // get length (first [2] bytes of message)
        int bytes = SDLNet_TCP_Recv(sock, _buf + _bufsize, sizeof(Uint16) - _bufsize);
        if(bytes <= 0) { return false;} // really -1 on failure

        _bufsize += bytes;

    } else if (_bufsize < SMSG_LEN(_buf)) {
        // Make sure that the client isn't sending message lengths that
        // go beyond our buffer size
        if (SMSG_LEN(_buf) >= (SBUF_SIZE - 4)) {
            cerr << "LEN " << SMSG_LEN(_buf) << " is too long!\n";
            cerr << "MTYPE was " << SMSG_TYPE(_buf) << endl;
            exit(1);
        }

        int bytes = SDLNet_TCP_Recv(sock, _buf + _bufsize, SMSG_LEN(_buf) - _bufsize);

        if(bytes <= 0) { return false; } // really -1 on failure

        _bufsize += bytes;
    }

    return true;// got something
}

}
}



int HTTPGet(void *http_data)
{
    assert(http_data != 0);
    
    HTTPData *data = (HTTPData *) http_data;
    
    string body; // returned by function call    
    
    if (!HTTPGet(data->host, data->port, data->path, body))
    {
        cerr << "Encountered errors in HTTPGet of: \n\t"
             << data->host << ":" << data->port << data->path << endl; 
    }
    
    data->in_transit = false;
    return 1;
}


bool HTTPGet(string host, int port, string path, string &body) {

    // Attempts to get http://host:port/path
    // returns false on error, true on success.
    //
    assert(port >= 0 && port < 70000);
    if (port <0 || port > 70000)
        return false; // no go if bad port

    // resolve the hostname
    IPaddress ip;
    SDLNet_ResolveHost(&ip, (char *) host.c_str(), port);
    if(ip.host == INADDR_NONE) {
        cerr << "Could not resolve host " << host << ".\n";
        return false;
    }

    // try connecting
    TCPsocket sout = NULL;
    sout = SDLNet_TCP_Open(&ip);
    if (sout == NULL) {
        cerr << "Could not connect to " << host << ":" << port
        << "\n";
        return false;
    }

    // request page
    char sport[10];
    sprintf(sport, "%d", port); // overflow ickiness could occur here (but see above assertion)
    string request  = "GET " + path + " HTTP/1.1\x0d\x0a";
    request += "Connection: close\x0d\x0a";
    request += "Host: " + host + ":" + sport + "\x0d\x0a\x0d\x0a";
    SDLNet_TCP_Send(sout, (char *) request.c_str(), request.length());

    // get page
    char rbuf[200];
    int chars = 0;
    do {
        chars = SDLNet_TCP_Recv(sout, rbuf, sizeof(rbuf));
        body.append(rbuf, chars);
//        cout << "APPENDING: " << string(rbuf, chars) << endl;
    } while (chars != 0);

    SDLNet_TCP_Close(sout);

    return true;
}


bool HTTPPostData(string host, int port, string path, string &data, string &body) {
    // Attempts to get http://host:port/path
    // returns false on error, true on success.
    //
    assert(port >= 0 && port < 70000);
    if (port <0 || port > 70000)
        return false; // no go if bad port

    // resolve the hostname
    IPaddress ip;
    SDLNet_ResolveHost(&ip, (char *) host.c_str(), port);
    if(ip.host == INADDR_NONE) {
        cerr << "Could not resolve host " << host << ".\n";
        return false;
    }

    // try connecting
    TCPsocket sout = NULL;
    sout = SDLNet_TCP_Open(&ip);
    if (sout == NULL) {
        cerr << "Could not connect to " << host << ":" << port
        << "\n";
        return false;
    }

    string boundary = "AaBbCcDdEeFfGgxx1234"; // hopefully this isn't in our data ;-)

    // Pre-generate content portion (so we can calculate the length).
    string content = "--" + boundary + "\x0d\x0a";
    content += "Content-Disposition: form-data; name=filedata\x0d\x0a";
    content += "Content-Transfer-Encoding: binary\x0d\x0a\x0d\x0a";
    content += data;
    content += "\x0d\x0a--" + boundary + "--\x0d\x0a";


    // Generate the header of page request and cat it together.
    char sport[10];
    sprintf(sport, "%d", port); // overflow ickiness could occur here (but see above assertion)
    string request  = "POST " + path + " HTTP/1.1\x0d\x0a";
    request += "Host: " + host + ":" + sport + "\x0d\x0a";
    request += "Content-type: multipart/form-data, boundary=" + boundary + "\x0d\x0a";
    sprintf(sport, "%d", int(content.length()));
    request += "Content-length: " + string(sport) + "\x0d\x0a";
    request += "Connection: close\x0d\x0a\x0d\x0a" + content;

    // Send the request out
    SDLNet_TCP_Send(sout, (char *) request.c_str(), request.length());

    // Get the response
    char rbuf[200];
    int chars = 0;
    do {
        chars = SDLNet_TCP_Recv(sout, rbuf, sizeof(rbuf));
        body.append(rbuf, chars);
    } while (chars != 0);
// cout << body; // for debugging

    // Then we out!
    SDLNet_TCP_Close(sout);

    return true;

}

int HTTPPostData(void *http_data)
{
    assert(http_data != 0);
    
    HTTPData *data = (HTTPData *) http_data;
    
    string body; // returned by function call    
    
    HTTPPostData(data->host, data->port, data->path, data->post_data, body);
    
    data->in_transit = false;
    return 1;
}

string getIPString(IPaddress *ip)
{
    assert(ip != 0);
    char tmp[20];
    sprintf(tmp, "%d.%d.%d.%d", ip->host & 0xff, (ip->host >> 8) & 0xff, (ip->host >> 16) & 0xff, (ip->host >> 24) & 0xff);
    return tmp;
}

