#include "Sound.h"
#include "Server.h"
#include <cassert>

namespace Salp {
namespace Game {

Game::State *Sound::_gamestate = 0;
Uint8 Sound::_master_cutoff = 0;
bool Sound::_audio_enabled = true;

Sound::Sound() {
    _chunk = 0;
    _initial_volume = 128;
    _id = -1;
   // _channel = -1;
}

Sound::Sound(string name, Uint8 volume) {
    _chunk = 0;
    _id = -1;
    assert(_id == -1);
    assert(volume > 0 && volume <= 128);

    _name = name; // do this before the short-circuit return below.
                  // we still want to know the audio file name even if
                  // we won't be hearing it ourselves!

    // dont waste memory if we dont need the audio
    if (!_audio_enabled)    return;

    _chunk = Mix_LoadWAV(name.c_str());
    if (_chunk) {
        _initial_volume = _chunk->volume = volume;
    } else {
        cerr << "Warning: Could not open audio file " << name 
             << ": " << Mix_GetError() << ".\n";
    }
}

Sound::~Sound() {
    if(_chunk) Mix_FreeChunk(_chunk); // Free sound if we loaded it
}


Uint8 Sound::decMasterVolume(Uint8 volume)
{
    if (int(volume) + int(_master_cutoff) > 128)
        return setMasterVolume(128);
    else
        return setMasterVolume(_master_cutoff + volume);
}

string Sound::getName() {
    return _name;
}

int Sound::getId() {
    return _id;
}

string Sound::getMasterVolumeString()
{
    // Return master volume in string, measured as percentage.
    char volstr[20];
    sprintf(volstr, "%6.2f", float(128 - _master_cutoff)/1.28);
    return volstr;
}

Uint8 Sound::incMasterVolume(Uint8 volume)
{
    if (volume >= _master_cutoff)
        return setMasterVolume(0);
    else
        return setMasterVolume(_master_cutoff - volume);
}

void Sound::setId(unsigned int id) {
    _id = id;
}

bool Sound::isGood() {
    return !isAudioEnabled() || _chunk != 0;
}

void Sound::playOnce() {
    
    // If we be servin', pass it on.
    if(Game::Server::_server) {
        char data[20];
        int msg_len = (sizeof(unsigned short) << 1) + 1;
        SET_SMSG_LEN(data, msg_len); // init length
        SET_SMSG_TYPE(data, Game::MS_SOUND_PLAY_ONCE);
        assert(getId() <= 255);

        *SMSG_DATA(data) = getId() & 255; // set sound id

///        cout << "Telling to play sound " << int(*SMSG_DATA(data));
        
        Game::Server::_server->sendAllState(data, msg_len, true); // send
    }

    // If we dont want to hear audio, or we can't play it, get out of here
    if (!_audio_enabled || !_chunk)   return; 
        
    // If our vol won't even let us play it, dont try.
    if (_initial_volume <= _master_cutoff)
        return;
         
    // Mod sound to our volume level
    _chunk->volume = _initial_volume - _master_cutoff;

    // Actually play the sound
    Mix_PlayChannel(-1, _chunk, 0); // any channel, _chunk, play once
    
}

void Sound::setGameState(Game::State *gs) {
    _gamestate = gs;
}

Uint8 Sound::setMasterVolume(Uint8 volume)
{
    return _master_cutoff = volume;
}



};
};
