#include "Game/State.h"

#include <cassert>
#include <fstream>
#include <cstdio>

#include "Tile/Loader.h"
#include "Sock.h"
#include "Server.h"
#include "OSD.h"
#include "General/general.h"
#include "Game/Event.h"
#include "Game/Vector.h"

using namespace Salp::General;

namespace Salp {
namespace Game {

State::State()
{
    _title = "Untitled Level";
    _author = "Unknown";
    _levelid = "";
    _osd = 0;
    _prefbucket = 0;
    _client = 0;
    _server = 0;
    _startmap = 0;
    _autoloadtiles = true;
    _scoreboard = new Scoreboard(this);
    _team_start_locs.resize(MAX_TEAMS);
}

State::~State()
{
    delete _client; // DO THESE FIRST
    delete _server; // DO THESE FIRST

    delete _scoreboard;

    // delete our various tile sets/sequences/maps from
    // memory

    unsigned int i;
    for(i=0; i<_tilesets.size(); i++)
        delete _tilesets[i];
    for(i=0; i<_sequences.size(); i++)
        delete _sequences[i];
    for(i=0; i<_maps.size(); i++)
        delete _maps[i];
    for(i=0; i<_sounds.size(); i++)
        delete _sounds[i];
    cleanupEvents();

    // free items
    map<Uint32, Game::Item*>::iterator gi;
    while((gi=_items.begin()) != _items.end())
    {
        delItem(gi->first);
    }
    
    delete _osd;    
    delete _prefbucket;
}

void State::addEvent(Event& e)
{
    _events.push(e); // add event to queue
}

void State::addItem(Game::Item *gi)
{
    assert(gi != 0);
    assert(gi->getId() != 0);
    assert(_items.find(gi->getId()) == _items.end());
    assert(gi->getMap() != 0);

    // set State to ME!
    gi->setGameState(this);

    // Tell everyone about it
    //
    if (_server) {
        assert(sizeof(unsigned short) == 2);
        // Create items that I have.
        //
        unsigned short msg_len = sizeof(unsigned short) << 1;
        SET_SMSG_TYPE(tmp_data, Game::MS_CREATE_ITEM);
        msg_len += gi->writeDataCreate(SMSG_DATA(tmp_data));  
        SET_SMSG_LEN(tmp_data, msg_len);
        _server->sendAllState(tmp_data, msg_len);
    }

    // add piece to master item map
    _items[gi->getId()] = (Game::Item *) gi;
    
    if(gi->isNPC())
        _items_npc[gi->getId()] = (Game::NPC *) gi;
    
    // do this last, if possible. don't want to mix the
    // create/delete messages (create from above, delete
    // from potential collision).
    if(gi->getType() & GI_CAN_COLLIDE_WITH_ITEM) {
        _items_icollidable[gi->getId()] = (Game::Item *) gi;
        if (_server)
            gi->detectCollisionItemAll(); // do this last in function!
    }
}

Game::Sound * State::addSound(Game::Sound *gs)
{    
    // Adds the sound to the state and returns the pointer
    // to the sound. Returns 0 on failure.
    
    // Make sure we're getting a good sound object
    assert(gs != 0);
    assert(gs->getId() == -1);
    assert(gs->isGood());
    if(!gs || !gs->isGood()) {
        cerr << "Invalid Game::Sound object passed to State::addSound.\n";
        exit(1);
    }
    
#ifndef NDEBUG
    // Make sure we didn't already 
    for(unsigned int i=0; i<_sounds.size(); ++i) {
        assert(_sounds[i] != gs);
    }
#endif
    
    // Give the sound a unique id (pointing to its entry in this list)
    // and add to our vector.
    gs->setId(_sounds.size());
    _sounds.push_back(gs);
   
    return gs;
}

void State::addTeamStartLoc(unsigned int team_id, const StartLoc loc)
{
    assert(team_id < MAX_TEAMS);
    
    _team_start_locs[team_id].push_back(loc);
}

void State::addTileMap(Tile::Map *tm)
{
    // Adds a newly allocated TileMap to our array of
    // TileMaps. Once you add it, the State is
    // responsible for deleting it.

    assert(tm != 0);
    assert(getTileMap(tm->getName()) == 0);
    _maps.push_back(tm);
    
    tm->setId(_maps.size() - 1);
    tm->setGameState(this);
    
    if (_server) // Only interpret mapcodes if we're serving.. this may
                 // change later, but messes up how we currently do 
                 // ItemPowerUp stuff in the method.
        tm->interpretMapCodes();
    
    if (tm->needsAnimation())
        _maps_anim.push_back(tm);
}

void State::addTileSequence(Tile::Sequence *ts)
{
    // Adds a newly allocated Tile::Sequence to our array of
    // Tile::Sequences. Once you add it, the State is
    // responsible for deleting it.

    assert(ts != 0);
    assert(getTileSequence(ts->getName()) == 0);
    ts->setId(_sequences.size());
    _sequences.push_back((Tile::Sequence *) ts);
    
}

void State::addTileSet(const Tile::Set *ts)
{
    // Adds a newly allocated TileSet to our array of
    // TileSets. Once you add it, the State is
    // responsible for deleting it.

    assert(ts != 0);
    assert(getTileSet(ts->getName()) == 0);
    _tilesets.push_back((Tile::Set *) ts);
}


void State::animate(unsigned int frames)
{
    assert(_server);
    assert(frames > 0);
    
    // Process events!    x += 3; // add 3 to x ;-)
    processEvents();
    
    // Animate maps
    animateMaps(frames);
    
    // tell all items to animate
    map<Uint32, Game::Item*>::iterator gi;

    // 
    // Animate them -- custom item code. Server-only.
    //
    for(gi=_items.begin(); gi != _items.end(); gi++)
    {
        if((gi->second->getType()  & GI_CAN_ANIMATE) &&
                (gi->second->getType()  & GI_NOT_DELETED) ) {

//            Sint8 oldxvel = gi->second->getXVel(), oldyvel = gi->second->getYVel();
            
            gi->second->animatePrimitive();
            
            for (unsigned int i = frames; i--; )
                gi->second->animate();

//            if(oldxvel != gi->second->getXVel()) gi->second->setDirty(DIRTY_XVEL);
//            if(oldyvel != gi->second->getYVel()) gi->second->setDirty(DIRTY_YVEL);
        }

    }

    flushItemDeletionQueue();

    // Do simple physics stuff to animate their *position*
    for(gi=_items.begin(); gi != _items.end(); gi++)
    {
        if((gi->second->getType() & GI_CAN_MOVE) &&
                (gi->second->getType() & GI_NOT_DELETED)) {
            Sint32 oldxpos = gi->second->getXPos(), oldypos = gi->second->getYPos();
//            Sint8 oldxvel = gi->second->getXVel(), oldyvel = gi->second->getYVel();

            gi->second->animatePos(frames);

            if(oldxpos != gi->second->getXPos()) gi->second->setDirty(DIRTY_XPOS);
            if(oldypos != gi->second->getYPos()) gi->second->setDirty(DIRTY_YPOS);
//            if(oldxvel != gi->second->getXVel()) gi->second->setDirty(DIRTY_XVEL);
//            if(oldyvel != gi->second->getYVel()) gi->second->setDirty(DIRTY_YVEL);
        }
    }

    flushItemDeletionQueue();

    if (_server) {
        // Update clients' knowledge of the gamestate.
        //
        static int frames_waited = 0;
        if (!frames_waited) {

            char *dataptr = &tmp_data[0];
            char *maxptr = dataptr + 1024 - 104; // assumes 100 bytes is max update length per item

            for(gi=_items.begin(); gi != _items.end(); gi++)
            {
                if( gi->second->isDirty()) {
                    // Update items that I have.

                    Uint16 msg_len = 4; // init length
                    msg_len += gi->second->writeDataUpdate(dataptr + 4);  
                    assert(msg_len <= 104);                     // watch for large updates.. we dont want these ;-)
                    SET_SMSG_LEN(dataptr, msg_len); 
                    SET_SMSG_TYPE(dataptr, Game::MS_UPDATE_ITEM);

                    dataptr += msg_len;

                    if (dataptr >= maxptr) { 
                        // send it out
                        _server->sendAllState(&tmp_data[0], dataptr - &tmp_data[0], true);
                        dataptr = &tmp_data[0];
                    }

                    gi->second->clearDirty(); // scrub a dub dub
                }
            }

            if (dataptr != &tmp_data[0]) { 
                // send it out
                _server->sendAllState(&tmp_data[0], dataptr - &tmp_data[0], true);
                dataptr = &tmp_data[0];
            }
        }

        frames_waited = 0; //(frames_waited + 1) % 4;    
    }
}

void State::animateMaps(const unsigned int frames)
{
    for (unsigned int i=0; i<_maps_anim.size(); ++i)
        _maps_anim[i]->animate(frames);
}

void State::animatePosSoft(Game::Item *except)
{
    assert(!_server);
    
    map<Uint32, Game::Item*>::iterator gi;

    // move our x and y closer to the server's
    if (except) {
        for(gi=_items.begin(); gi != _items.end(); gi++)
            if(gi->second != except)
                gi->second->animatePosSoft();
    } else { 
        for(gi=_items.begin(); gi != _items.end(); gi++)
                gi->second->animatePosSoft();
    }
}

void State::appendConsoleMsgAll(string msg)
{
    assert(_osd != 0);
    //assert(_server != 0); // must be server to use this..  else change function to sendAll 
                          // instead of sendAllState.
    
    _osd->appendMsg(msg);
    

    if (_server != 0) {
        unsigned short msg_len = 4 + msg.length();
        char *data = new char[msg_len];
        SET_SMSG_LEN(data, msg_len);
        SET_SMSG_TYPE(data, MS_CONSOLE_MESSAGE);
        strncpy(data + 4, msg.c_str(), msg.length());
        _server->sendAllState(data, msg_len);
        delete[] data;
    }
}

void State::delUnusedObjects()
{
    // delete unused maps
    unsigned int i;
    for(i=0; i<_maps.size(); i++)
        if(!_maps[i]->getUsedInLevel())
        {
            _maps.erase(_maps.begin() + i);
            --i;
        }
}

void State::delItem(const Uint32 id, bool just_remove)
{    
    // Note: If you "just_remove", make sure you properly delete the Item
    //       object eventually. If just_remove is FALSE (default), then the
    //       Item is actually deleted (as in memory delete), not just removed
    //       from the gamestate.


    assert(_items.find(id) != _items.end());

    // Tell everyone else to delete it
    if(_server) {
        char data[200];
        assert(sizeof(unsigned short) == 2);
        // Create items that I have.
        //
        SET_SMSG_TYPE(data, Game::MS_DELETE_ITEMS);
        unsigned int msg_len = sizeof(unsigned short) * 2 + sizeof(Uint32);
        SET_SMSG_LEN(data, msg_len);
        *((Uint32 *) SMSG_DATA(data)) = hton32(id);
        _server->sendAllState(data);
    }

    // if i can icollide, take me out of icollide list
    if(_items[id]->getType() & GI_CAN_COLLIDE_WITH_ITEM) {
        assert(_items_icollidable.find(id) != _items_icollidable.end());
        _items_icollidable.erase(id);
    }

    // if i am npc, take me out of npc list
    if(_items[id]->isNPC()) {
        assert(_items_npc.find(id) != _items_npc.end());
        _items_npc.erase(id);
    }
    
    // delete from memory
    if (!just_remove) {
        delete _items[id];
    }
    
    // delete from map
    _items.erase(id);
}


void State::delItemDelayed(Game::Item *gi)
{
    // If we're not serving, we aren't allowed to delete things this way!
    // Let the server explicitly tell us what exists and dies.
    if (!_server) return;
    assert(_server);
    
    // Allow multiple delete-queues (is this okay? I think so..)
    if(!(gi->getType() & GI_NOT_DELETED))
        return;

    assert(_items.find(gi->getId()) != _items.end());
    assert(gi->getType() & GI_NOT_DELETED);

    gi->setType(gi->getType() & ~GI_NOT_DELETED);
    _item_delete_queue.push_back(gi->getId());
}

void State::display(SDL_Surface *screen, Game::Item *gi, Sint32 &scrollx, Sint32 &scrolly)
{
    // draws game screen from Game::Item's point of view.
//    static Sint32 scrollx=0, scrolly=0;

    Tile::Map *tm = gi->getMap();

    // Draw map (the map is responsible for all background issues)
    tm->display(screen, scrollx, scrolly);

    // Draw all items in map (there's room for optimization here)
    map<Uint32, Game::Item*>::reverse_iterator stateitem;
    for(stateitem=_items.rbegin(); stateitem != _items.rend(); stateitem++)
    {
        if(stateitem->second->getType() & Game::GI_IS_VISIBLE)
            if (stateitem->second->canDrawImage(screen, scrollx, scrolly))
                stateitem->second->drawImage(screen, scrollx, scrolly);
    }
    
    // Draw map foreground on top of this crud!
    tm->displayForeground(screen, scrollx, scrolly);

}

void State::doCmd(const char * cmd)
{
    assert(_prefbucket != 0);
    assert(_osd != 0);
    assert(cmd != 0);
    assert(*cmd != 0);
    
    char args[201];
    if (sscanf(cmd, " say %200[^\n]", args) == 1) {
        // Send a message to ALL players in the game.
        //
        
        if (!_client) {
            _osd->appendMsg("You are not currently connected to a server.");
        } else if (args[0] == 0) {
            _osd->appendMsg("Message too short; not sent.");
        } else {
            _client->_peer->sendMsg(MC_SAY_TEXT, args, strlen(args));
        }
    } else if (sscanf(cmd, " name %30[^\n]", args) == 1) {
        // Change the user's name
        //
        
        if (args[0] == 0) {
            _osd->appendMsg("Name too short; not changed.");
        } else if (args == _prefbucket->getVal("PLAYERNAME")) {
            _osd->appendMsg("That's your name, don't wear it out!");
        } else {
            _prefbucket->setVal("PLAYERNAME", args);
            if (_client) {
                _client->_peer->sendMsg(MC_SET_PLAYERNAME, _prefbucket->getVal("PLAYERNAME"));
            } else {
                _osd->appendMsg("Name changed.");
            }
        }
    } else {
        if (_client)
            _client->_peer->sendMsg(MC_DO_COMMAND, cmd, strlen(cmd));
        else
            _osd->appendMsg("Invalid command.");
    }
}

void State::flushItemDeletionQueue() {
    // delete items in the deletion queue, and clear it

    if(_item_delete_queue.size()) {
        for(unsigned int i=0; i<_item_delete_queue.size(); i++) {
            delItem(_item_delete_queue[i]);
        }

        _item_delete_queue.erase(_item_delete_queue.begin(), _item_delete_queue.end());
    }

}

string State::getAuthor() const
{
    return _author;
}

Game::Item * State::getItem(const Uint32 id) {
    assert(id != 0);
    assert(_items.find(id) != _items.end());

    return _items[id];
}

Game::Item * State::getItemMaybe(const Uint32 id) {
    return (_items.find(id) != _items.end()) ? _items[id] : 0;
}


unsigned int State::getMaxTeams() {
    // Returns max teams supported by the maps that have been loaded.
    // Note that this should be between 1 and MAX_TEAMS. If it's any less,
    // you've got problems (or haven't fully loaded the level).
    
    unsigned int max_teams = 0;
    while (_team_start_locs[max_teams].size() > 0 && max_teams < MAX_TEAMS)
        ++max_teams;
    return max_teams;
}

bool State::hasItem(const Uint32 id) {
    return _items.find(id) != _items.end();
}

map<Uint32, Game::Item*> * State::getItemsICollidable()
{
    // Return pointer to map which
    // reflects items that can collide with other
    // items.

    return(&_items_icollidable);
}

string State::getLevelId() {
    return _levelid;
}

string State::getSlotServerDesc() {
    return   string("ver=")          + itoa(SALP_WARS_VERSION)
           + "&min_ver="             + itoa(SALP_WARS_MIN_VERSION)
           + "&level="               + URLEncode(getLevelId()) 
           + "&server_name="         + URLEncode(getPrefBucket()->getVal("SERVERNAME")) 
           + "&level_name="          + URLEncode(getTitle());
}

Game::Sound *State::getSound(const string name, bool auto_load)
{
    for(unsigned int i=0; i<_sounds.size(); i++)
        if(_sounds[i]->getName() == name)
            return _sounds[i];

    if(auto_load) {
        // If auto-load, load the sound and re-search
        addSound(new Game::Sound(name));
        return getSound(name, false);
    }

    return 0;

}

Tile::Map * State::getStartMap()
{
    assert(_startmap != 0);
    return _startmap;
}

unsigned int State::getTeamWeakest()
{
    // gets team needing players the most; favors team 0 if there
    // are no teams or all teams are equal.
    
    unsigned int weakest_team = 0, weakest_points = (unsigned int) -1;
    unsigned int max_teams = getMaxTeams();
    assert(max_teams > 0);
    for (unsigned int i=0; i<max_teams; ++i) {
        unsigned int tmp_points = 0;
        // Build points
        
        // o  # of players in team boosts pointage
        tmp_points += _team_start_locs[i].size();
        
        // Compare to last
        if (tmp_points < weakest_points) {
            weakest_team = i;
            weakest_points = tmp_points;
        }
    }
    
    return weakest_team;
    
}

Tile::Map * State::getTileMap(const unsigned int id) const
{
    // Returns a pointer to a TileMap object, at index <id>
    // into the list of loaded TileMaps.
    assert(id < _maps.size());
    assert(_maps[id] != 0);

    return(_maps[id]);
}


Tile::Map * State::getTileMap(const string name, bool auto_load)
{
    // Returns a pointer to a TileMap object if
    // one matches the name passed. Else returns 0.
    //
    // This is an expensive function to call since it
    // uses a linear search, and should not be used
    // for most anything besides map loading.
    //
    assert(auto_load && _autoloadtiles || !auto_load);
        
    for(unsigned int i=0; i<_maps.size(); i++)
        if(_maps[i]->getName() == name)
            return _maps[i];

    if(auto_load && getTileSet(name.substr(0, name.find_last_of("/")), true) ) {
        // If auto-load and we have corresponding tileset, re-search!
        return getTileMap(name, false);
    }

    return 0;
}

unsigned int State::getTileMapCount() const
{
    // Returns the number of TileMap objects loaded.

    return(_maps.size());
}

Tile::Sequence * State::getTileSequence(const unsigned int id) const
{
    // Returns a pointer to a TileSequence object, at index <id>
    // into the list of loaded TileSequences.

    assert(id < _sequences.size());
    assert(_sequences[id] != 0);

    return(_sequences[id]);
}

Tile::Sequence * State::getTileSequence(const string name, bool auto_load)
{
    // Returns a pointer to a TileSequence object if
    // one matches the name passed. Else returns 0.
    //
    // This is an expensive function to call since it
    // uses a linear search, and should not be used
    // for most anything besides map loading.
    //

    assert(auto_load && _autoloadtiles || !auto_load);

    for(unsigned int i=0; i<_sequences.size(); ++i)
        if(_sequences[i]->getName() == name)
            return _sequences[i];

    if(auto_load && getTileSet(name.substr(0, name.find_last_of("/")), true) ) {
        // If auto-load and we have corresponding tileset, re-search!
        return getTileSequence(name, false);
    }

    return 0;
}


Tile::Set * State::getTileSet(const unsigned int id) const
{
    // Returns a pointer to a TileSet object, at index <id>
    // into the list of loaded TileSets.

    assert(id < _tilesets.size());
    assert(_tilesets[id] != 0);

    return(_tilesets[id]);
}

Tile::Set * State::getTileSet(const string name, bool auto_load)
{
    // Returns a pointer to a TileSet object if
    // one matches the name passed. Else returns 0.
    //
    // This is an expensive function to call since it
    // uses a linear search, and should not be used
    // for most anything besides map loading.
    //

    assert(auto_load && _autoloadtiles || !auto_load);

    for(unsigned int i=0; i<_tilesets.size(); i++)
        if(_tilesets[i]->getName() == name)
            return _tilesets[i];


    if(auto_load) {
        // If auto-load and not found, load it! Then re-search.
        Tile::loadTileSet(name, this, _screen);
        return getTileSet(name, false);
    }

    return 0;
}

string State::getTitle() const
{
    return _title;
}

unsigned int State::getUniqueTeamId()
{
    static unsigned int lastTID = 0;
    assert(lastTID != ((unsigned int) -1));
    return ++lastTID;
}

unsigned int State::getVersion() const
{
    return _version;
}

bool State::isGood() {
    if (getMaxTeams() <= 0) {
        cerr << "ERROR: The map(s) need to have starting locations for at least TEAM A. Either none such maps were loaded by the level, the level hasn't been loaded, or the map author(s) forgot to set starting locations on the map.\n";
        return false;
    }
    return true;
}

void State::loadLevel(const string levpath)
{
    assert(_levelid.length() == 0); // don't load another lev on top of this
    // NO, clmode could be "yes" here.  assert(_screen != 0);

    // Load the level
    Tile::loadLevel(levpath, this, _screen);
    
    _levelid = levpath;

    //	gGameState->delUnusedObjects(); // kill level data that isn't used
 
    assert(getTileMapCount() != 0);

    // Set window title to level title.
    if (_screen) 
        SDL_WM_SetCaption((char *) getTitle().c_str(), 0);
}

void State::lockAutoLoad()
{
    _autoloadtiles = false;
}

void State::playSound(unsigned int id) {
    // Play the sound of id <id> from our list.
    //
    if (id < _sounds.size())
        _sounds[id]->playOnce();
}

void State::cleanupEvents() {
    // Performs cleanup on all our events (called when ~State() is called)
    //
    
    while (!_events.empty()) {
        Event *e = (Event *) &_events.top();
        
        switch (e->type) {
            case EVENT_ADD_ITEM:
                delete e->data.add_item.item;
                break;
        
            default:
                cerr << "Error: encountered unknown event type "
                     << int(e->type) << endl;
        };
        
        _events.pop();
    }
}

void State::processEvents() {
    Uint32 now = time(0);
    while (!_events.empty() && _events.top().start_time <= now) {
        Event *e = (Event *) &_events.top();
        
        switch (e->type) {
            case EVENT_ADD_ITEM:
                addItem( e->data.add_item.item );
                break;
        
            default:
                cerr << "Error: encountered unknown event type "
                     << int(e->type) << endl;
        };
        
        _events.pop();

    }
}

void State::sendInitGameState(Game::PeerData *pdata)
{
    char * data = &tmp_data[0];
    unsigned int i;

    // Load some tilesets (before loading items)
    //
    SET_SMSG_TYPE(data, Game::MS_LOAD_TILE_SET);
    for (i=0; i<_tilesets.size(); i++) {
        SET_SMSG_DATA_STR(data, _tilesets[i]->getName());
        SET_SMSG_LEN(data, 4 + _tilesets[i]->getName().length());
        pdata->sendMsgRaw(data, SMSG_LEN(data));
    }

    // Load some sounds
    //
    SET_SMSG_TYPE(data, Game::MS_LOAD_SOUND);
    for (i=0; i<_sounds.size(); i++) {
        SET_SMSG_DATA_STR(data, _sounds[i]->getName());
        SET_SMSG_LEN(data, 4 + _sounds[i]->getName().length());
        pdata->sendMsgRaw(data, SMSG_LEN(data));
    }

    // Initialize scoreboard
    int len;
    _scoreboard->writeDataUpdateFull(data, len);
    pdata->sendMsgRaw(data, len);

    // Load the base files for this level
    //
    pdata->sendMsg(Game::MS_LOAD_LEVEL, (char *) _levelid.c_str(), _levelid.length() + 1); // don't forget the 0-char ;)

    map<Uint32, Game::Item*>::iterator gi;
    
    // Create items that I have.
    //
    static unsigned int msg_size;    

    SET_SMSG_TYPE(data, Game::MS_CREATE_ITEM);
    for(gi=_items.begin(); gi != _items.end(); gi++)
    {
        msg_size = 4;
        msg_size += gi->second->writeDataCreate(data + 4);  
        SET_SMSG_LEN(data, msg_size);
        pdata->sendMsgRaw(data, msg_size);
    }

    // Update items that I have.
    //
    SET_SMSG_TYPE(data, Game::MS_UPDATE_ITEM);
    for(gi=_items.begin(); gi != _items.end(); gi++)
    {
        msg_size = 4;
        msg_size += gi->second->writeDataUpdate(data + 4, true); // True means, send EVERYTHING, not just the dirty stuff :-)  
        SET_SMSG_LEN(data, msg_size);
        pdata->sendMsgRaw(data, msg_size);
    }

}

void State::setAuthor(const string author)
{
    _author = author;
}

void State::setScreen(SDL_Surface *screen)  {
    _screen = screen;
}

void State::setStartMap(Tile::Map *tm)
{
    _startmap = tm;
}

void State::setTitle(const string title)
{
    _title = title;
}

void State::setVersion(const unsigned int version)
{
    _version = version;
}

void State::updateCameraPosition(SDL_Surface *screen, Game::Item *gi, Sint32 &scrollx, Sint32 &scrolly, float scroll_minthresh, float scroll_maxthresh)
{
    Tile::Map *tm = gi->getMap();

    //
    // Set camera position.
    //

    int maxscrollx   = (int) tm->getPixWidth()  - screen->w;
    int maxscrolly   = (int) tm->getPixHeight() - screen->h;
    int leftxthresh  = scrollx + int(screen->w  *scroll_minthresh);
    int rightxthresh = scrollx + int(screen->w  *scroll_maxthresh);
    int leftythresh  = scrolly + int(screen->h  *scroll_minthresh);
    int rightythresh = scrolly + int(screen->h  *scroll_maxthresh);

    // If the user doesn't have a buddy or extra hand to use 
    // the mouse to scroll around, we use our auto window code.
        if(gi->getXPos() < leftxthresh) {
            scrollx = gi->getXPos() - int(screen->w*scroll_minthresh);
            if (scrollx < 0)               scrollx = 0;
            else if (scrollx > maxscrollx) scrollx = maxscrollx;
        } else if(gi->getXPos()+gi->getWidth() > rightxthresh) {
            scrollx = gi->getXPos()+gi->getWidth() - int(screen->w * scroll_maxthresh);
            if (scrollx < 0)               scrollx = 0;
            else if(scrollx > maxscrollx)  scrollx = maxscrollx;
        }

        if (gi->getYPos() < leftythresh) {
            scrolly = gi->getYPos() - int(screen->h*scroll_minthresh);
            if (scrolly < 0)               scrolly = 0;
            else if (scrolly > maxscrolly) scrolly = maxscrolly;
        } else if (gi->getYPos()+gi->getHeight() > rightythresh) {
            scrolly = gi->getYPos() + gi->getHeight() - int(screen->h*scroll_maxthresh);
            if (scrolly < 0)               scrolly = 0;
            else if (scrolly > maxscrolly) scrolly = maxscrolly;
        }

    // Bounds check
    if(scrolly < 0)                scrolly = 0;
    else if(scrolly > maxscrolly)  scrolly = maxscrolly;
    
    if(scrollx < 0)                scrollx = 0;
    else if(scrollx > maxscrollx)  scrollx = maxscrollx;
}


} // namespace Game
} // namespace Salp
