#ifndef __STATE_H
#define __STATE_H

#include "SALPSTUB.h"


#include "Tile/Sequence.h"
#include "Tile/Map.h"
#include "Game/Item.h"
#include "Game/Sound.h"
#include "Game/PrefBucket.h"
#include "Game/Scoreboard.h"
#include "Game/Vector.h"

#include "Sock.h"

#include <vector>
#include <map>
#include <string>
#include <deque>
#include <queue>
#include "SALP_SDL.h"

namespace Salp {
namespace Game {

struct StartLoc {
    explicit StartLoc (Tile::Map *start_map, const Vector start_coord) { 
       coord = start_coord;
       map = start_map;
    };
    
    Game::Vector coord;
    Tile::Map   *map;
};

class State
{
public:
    State();
    ~State();

    void addEvent(Event& e);
    void addItem(Game::Item *gi);
    Game::Sound * addSound(Game::Sound *gs);
    void addTeamStartLoc(unsigned int team_id, const StartLoc loc);
    void addTileMap(Tile::Map * tm);
    void addTileSequence(Tile::Sequence *ts);
    void addTileSet(const Tile::Set *ts);
    void animate(unsigned int frames=1);
    void animateMaps(const unsigned int frames);
    void animatePosSoft(Game::Item *except=0);
    void appendConsoleMsgAll(string msg);
    void cleanupEvents(); // clears all events, performing cleanup as necessary
    void delUnusedObjects();
    void delItem(const Uint32 id, bool just_remove=false);
    void delItemDelayed(Item *gi);
    void display(SDL_Surface *screen, Game::Item *gi, Sint32& scrollx, Sint32& scrolly);
    void doCmd(const char * cmd);
    string          getAuthor() const;
    Game::Client   *getClient() { return _client; };
    Game::Item     *getItem(const Uint32 id);
    Game::Item     *getItemMaybe(const Uint32 id); // can return 0 if non-exist
    string          getLevelId(); // returns identifier for loc of level data
    unsigned int    getMaxTeams(); // max teams supported by game state (1 .. MAX_TEAMS)
    map<Uint32, Game::Item*> * getItemsICollidable();
    string          getSlotServerDesc(); // returns query-string desc of me for meta server
    Game::OSD      *getOSD()    { assert(_osd != 0);    return _osd; }  // we "guarantee" an OSD
    Game::PrefBucket * getPrefBucket() { return _prefbucket; }
    Game::Scoreboard * getScoreboard() { assert(_scoreboard != 0); return _scoreboard; }
    SDL_Surface    *getScreen() { assert(_screen != 0); return _screen; };
    Game::Server   *getServer() { return _server; };
    Game::Sound    *getSound(const string name, bool auto_load=false);
    Tile::Map      *getStartMap();
    unsigned int    getTeamWeakest(); // gets team needing players the most
    Tile::Map      *getTileMap(const unsigned int id) const;
    Tile::Map      *getTileMap(const string name, bool auto_load=false);
    unsigned int    getTileMapCount() const;
    Tile::Sequence *getTileSequence(unsigned int id) const;
    Tile::Sequence *getTileSequence(const string name, bool auto_load=false);
    Tile::Set      *getTileSet(const unsigned int id) const;
    Tile::Set      *getTileSet(const string name, bool auto_load=false);
    string          getTitle() const;
    static unsigned int    getUniqueTeamId();
    unsigned int    getVersion() const;
    bool            hasItem(const Uint32 id);
    bool            isGood();
    bool            isLevLocked() { return(!_autoloadtiles); };
    void loadLevel(const string levpath);
    void lockAutoLoad(); 
     
    void playSound(unsigned int id);
    void sendInitGameState(Game::PeerData *pdata);
    
    void setAuthor(const string author);
    void setClient(Game::Client * client) { _client = client; }
    void setOSD(Game::OSD * osd) { _osd = osd; }
    void setPrefBucket(Game::PrefBucket * prefbucket) { assert(prefbucket != 0); _prefbucket = prefbucket; }
    void setServer(Game::Server * server) { _server = server; }
    void setScreen(SDL_Surface *screen);
    void setStartMap(Tile::Map *tm);
    void setTitle(const string title);
    void setVersion(const unsigned int version);
    void updateCameraPosition(SDL_Surface *screen, Game::Item *gi, Sint32 &scrollx, Sint32 &scrolly, float scroll_minthresh, float scroll_maxthresh);

    friend class Salp::Game::Scoreboard;
    friend class Salp::Game::OSD;
    
private:
    void flushItemDeletionQueue();
    void processEvents();

    Game::Scoreboard    *_scoreboard;
    Game::OSD           *_osd;
    Game::Client        *_client;
    Game::Server        *_server;
    Game::PrefBucket    *_prefbucket;
    Tile::Map           *_startmap;
    
    vector<vector<StartLoc> >      _team_start_locs;
    vector<Tile::Set *>         _tilesets;
    vector<Tile::Sequence *>    _sequences;
    vector<Tile::Map *>         _maps;
    vector<Tile::Map *>         _maps_anim;
    vector<Game::Sound *>       _sounds;
    priority_queue<Game::Event> _events;
    SDL_Surface               * _screen;
    map<Uint32, Game::Item *>  _items;             // all items
    map<Uint32, Game::NPC *>   _items_npc;             // all npcs
    map<Uint32, Game::Item *>  _items_icollidable; // items that can collide with
    // other items
    vector<Uint32> _item_delete_queue;
    string _title;
    string _author;
    string _levelid; // level identifier
    unsigned int _version;
    bool _autoloadtiles;

};



} // namespace Game
} // namespace Salp


#endif
