#include "graphics.h"
#include <cassert>
#include <fstream>
#include <cstring>
#include "SDL_thread.h"

using namespace std;

namespace Salp {
namespace General {

SDL_Surface *tempsurface = 0;

SDL_Surface *copyToTempSurface(SDL_Surface *screen)
{
    assert(screen != 0);
    assert(tempsurface != 0);
    assert(screen->w == tempsurface->w);
    assert(screen->h == tempsurface->h);
    
    SDL_SetAlpha(tempsurface, SDL_SRCALPHA, SDL_ALPHA_OPAQUE);
    
    SDL_BlitSurface(screen, 0, tempsurface, 0); 
    
    return tempsurface;
}


void initTempSurface(SDL_Surface *screen)
{
    assert(screen != 0);
    assert(tempsurface == 0);
    tempsurface = SDL_CreateRGBSurface(SDL_SWSURFACE, screen->w, screen->h,
                               screen->format->BitsPerPixel, screen->format->Rmask,
                               screen->format->Gmask, screen->format->Bmask, screen->format->Amask);
}

SDL_Surface * cloneBitmap(SDL_Surface *src, Uint32 oldcolor, Uint32 newcolor)
{
    SDL_Surface * dst = SDL_CreateRGBSurface(SDL_HWSURFACE, src->w, src->h,
                               src->format->BitsPerPixel, src->format->Rmask,
                               src->format->Gmask, src->format->Bmask, src->format->Amask);

    Uint32 pixcolor;
    for (int x=dst->w-1; x>=0; --x)
        for (int y=dst->h-1; y>=0; --y) {
            pixcolor = getPixel(src, x, y);
            if (pixcolor == oldcolor) {
                putPixel(dst, x, y, newcolor);
            } else {
                putPixel(dst, x, y, pixcolor);
            }
        
        }
   
   return dst;

}

// alpha value of 128 is optimized
void fillRectAlpha(SDL_Surface *dst, SDL_Rect *dstrect, Uint32 color, Uint8 alpha)
{
    assert(tempsurface != 0);
    assert(dst->w == tempsurface->w);
    assert(dst->h == tempsurface->h);

    if (SDL_FillRect(tempsurface, dstrect, color) < 0)
        fprintf(stderr, "fillRectAlpha FillRect error: %s\n", SDL_GetError());

    SDL_SetAlpha(tempsurface, SDL_SRCALPHA, alpha); 

    if (SDL_BlitSurface(tempsurface, dstrect, dst, dstrect) < 0)
        fprintf(stderr, "fillRectAlpha BlitSurface error: %s\n", SDL_GetError());
}
    
Uint32 getPixel(SDL_Surface *image, unsigned short x, unsigned short y) {
    int bpp = image->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)image->pixels + y * image->pitch + x * bpp;

    switch(bpp) {
    case 1:
        return *p;

    case 2:
        return *(Uint16 *)p;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;

    case 4:
        return *(Uint32 *)p;

    default:
        return 0;       /* shouldn't happen, but avoids warnings */
    }
}


/*
 * Set the pixel at (x, y) to the given value
 * NOTE: The surface must be locked before calling this!
 */
void putPixel(SDL_Surface *surface, int x, int y, Uint32 pixel)
{
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        *p = pixel;
        break;

    case 2:
        *(Uint16 *)p = pixel;
        break;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
            p[0] = (pixel >> 16) & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = pixel & 0xff;
        } else {
            p[0] = pixel & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = (pixel >> 16) & 0xff;
        }
        break;

    case 4:
        *(Uint32 *)p = pixel;
        break;
    }
}

SDL_Surface *conv_surf_gl(SDL_Surface *s, int want_alpha)
{
    Uint32 rmask, gmask, bmask, amask;
    SDL_Surface *conv;
    int bpp = want_alpha ? 32 : 24;

    /* SDL interprets each pixel as a 24 or 32-bit number, so our
       masks must depend on the endianness (byte order) of the
       machine. */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff000000 >> (32 - bpp);
    gmask = 0x00ff0000 >> (32 - bpp);
    bmask = 0x0000ff00 >> (32 - bpp);
    amask = 0x000000ff >> (32 - bpp);
#else
    rmask = 0x000000ff;
    gmask = 0x0000ff00;
    bmask = 0x00ff0000;
    amask = want_alpha ? 0xff000000 : 0;
#endif

    /* check if the surface happens to be in the right format */
    if(s->format->BitsPerPixel == bpp
       && s->format->Rmask == rmask
       && s->format->Gmask == gmask
       && s->format->Bmask == bmask
       && s->format->Amask == amask
       && !(s->flags & SDL_SRCCOLORKEY)) {
	/* no conversion needed */
	return s;
    }

    /* wrong format, conversion necessary */

    /* SDL surfaces are created with lines padded to start at 32-bit boundaries
       which suits OpenGL well (as long as GL_UNPACK_ALIGNMENT remains
       unchanged from its initial value of 4) */
    conv = SDL_CreateRGBSurface(SDL_SWSURFACE, 16, 32, /* s->w, s->h, */ bpp,
				rmask, gmask, bmask, amask);
    if(!conv) {
	    SDL_FreeSurface(conv);
	    return NULL;
    }

    if(want_alpha) {
	    /* SDL sets the SDL_SRCALPHA flag on all surfaces with an
	       alpha channel. We need to clear that flag for the copy,
	       since SDL would attempt to alpha-blend our image otherwise */
	    SDL_SetAlpha(s, 0, 255);
    }

    /*
     * Do the conversion. If the source surface has a colourkey, then it
     * will be used in the blit. We use the fact that newly created software
     * surfaces are zero-filled, so the pixels not blitted will remain
     * transparent.
     */
    if(SDL_BlitSurface(s, NULL, conv, NULL) < 0) {
	    /* blit error */
	    SDL_FreeSurface(conv);
	    conv = NULL;
    }
    SDL_FreeSurface(s);

    return conv;
}

struct FileSaveRec {
    ~FileSaveRec() {
        delete data;
        delete fpath;
    };
    
    char * data;
    int len;
    char * fpath;
};

int saveFile(void * fileSaveRec) {
    // fsrec MUST BE OF TYPE FileSaveRec, or you're going to
    // run into problems! :)
    FileSaveRec *fsrec = (FileSaveRec *) fileSaveRec;

    ofstream fout(fsrec->fpath, ios::binary);
    
    if (fout.is_open()) {
        fout.write(fsrec->data, fsrec->len);
        fout.close();
    }

    delete fsrec;
    
    return 0;    
}

bool savePPM(SDL_Surface *s, const char *file)
{
    // Attempts to save surface data as a PPM file; returns true
    // on success, false on error

    SDL_LockSurface(s);
    
    FileSaveRec *fsrec = new FileSaveRec;
    
    char ppmhead[30];
    sprintf(&ppmhead[0], "P6 %d %d 255\r", int(s->w), int(s->h));
    fsrec->data  = new char[fsrec->len = (s->w * s->h * 3 + strlen(ppmhead))];
    fsrec->fpath = new char[strlen(file)+1];
    
    strcpy(fsrec->fpath, file);
    strcpy(fsrec->data, ppmhead);
    
    Uint8 *dptr = (Uint8 *) fsrec->data + strlen(ppmhead);
             
    for (int y=0; y<s->h; y++)
        for (int x=0; x<s->w; x++) {
            SDL_GetRGB(getPixel(s, x, y), s->format, &dptr[0], &dptr[1], &dptr[2]);
            dptr += 3;
        }

    SDL_CreateThread(saveFile, fsrec);
    SDL_UnlockSurface(s);
    return true;
}

};
};
