#include "init.h"
#include "input.h"

#include "General/general.h"
#include "Tile/Map.h"
#include "Tile/Loader.h"
#include "Game/Sound.h"
#include "Game/Server.h"
#include "Game/PlayerStomper.h"
#include "Game/ItemProjectile.h"
#include "Game/NPCBushBomber.h"
#include "Tile/Loader.h"
#include "General/media.h"

using namespace General;

void initClient()
{
    gClient = new Game::Client(gGameState, &gPlayer);
    if (!gClient) {
        cerr << "Not enough memory available to create game client.\n";
        exit(1);
    }

    int start_port = gCmdLine.getIntVal("port")-1;
    int tries = 0;
    bool started_client = false;
    while (++tries <= gPrefBucket->getIntVal("CLIENT_MAX_PORT_DRIFT") && !started_client) {
        ++start_port;
        started_client = gClient->connectTo(gCmdLine.getVal("host"), start_port);
    }
    gCmdLine.setVal("port", start_port);

    if (!started_client) {
        cerr << "Could not connect to server at " << gCmdLine.getVal("host") << ":" << gCmdLine.getIntVal("port")-1 << "\n";
        exit(1);
    }

    cout << "Connected to " << gCmdLine.getVal("host") << ':' << gCmdLine.getVal("port") << endl;

    // Tell server we're remote, if we are.
    if (!gServer) {
        gClient->_peer->sendMsg(Game::MC_STATE_INIT_REMOTE, "", 0);
        // Tell server we like UDP (if we do)
        // *** Make sure this comes after INIT_REMOTE; else it will be ignored!
        if (gCmdLine.getVal("udp") == "yes") {
            Uint16 my_udp_port = hton16(gClient->_peer->udpport);
            gClient->_peer->sendMsg(Game::MC_USE_UDP, (char *) &my_udp_port, sizeof(my_udp_port));
        }
    } else
        gClient->_peer->sendMsg(Game::MC_STATE_INIT, "", 0);

    // Tell the server our name
    gClient->_peer->sendMsg(Game::MC_SET_PLAYERNAME, (string) gPrefBucket->getVal("PLAYERNAME"));


    // Tell the server our gender
    if (gPlayer) gPlayer->getGender().setGender(gPrefBucket->getVal("PLAYERGENDER").c_str());
    gClient->_peer->sendMsg(Game::MC_SET_PLAYERGENDER, (string) gPrefBucket->getVal("PLAYERGENDER").substr(0,1));
    gGameState->setClient(gClient);
}

void initCommandLine(int argc, char *argv[])
{
    // ********************************************
    //  Process command-line arguments.
    // ********************************************

    // To run as cli & server: mode=server port=12394
    // To run as cli: mode=client host=localhost port=12394
    //  * screen=window|full

    // Defaults
    gCmdLine.setDefault(       "audio", "yes");    // yes, no
    gCmdLine.setDefault(      "clmode", "no");     // shortcut for command-line mode options (no audio, no joysticks, no display, ..)
    gCmdLine.setDefault(      "drawbg", "yes");    // yes, no
    gCmdLine.setDefault(      "drawfg", "yes");    // yes, no
    gCmdLine.setDefault(      "height", "240");
    gCmdLine.setDefault(        "host", "localhost");
    gCmdLine.setDefault(       "level", "levels/runway.lev");
    gCmdLine.setDefault(        "mode", "server"); // server, client, serveronly
    gCmdLine.setDefault(        "port", "12394");
    gCmdLine.setDefault(      "screen", "window"); // window, full, no
    gCmdLine.setDefault( "tcp_nodelay", "no");     // should I set my sockets to use TCP_NODELAY?
    gCmdLine.setDefault(       "width", "320");
    gCmdLine.setDefault(         "udp", "no");     // should I request that the server send me UDP game update packets?
    gCmdLine.setDefault(   "updateenv", "no");     // yes, no
    gCmdLine.setDefault( "usejoystick", "yes");    // yes, no
   
    gCmdLine.setDefault("frame_delay",  "30"); // 1000/30 = 33.3333, so 30ms.
    gCmdLine.setDefault( "safemode",  "yes");     // yes, no
    gCmdLine.setDefault( "predrawmaps",  "no");     // yes, no

    gCmdLine.setDefault(  "autoupdate", "no");    // yes, no

    // Let user overwrite defaults (though in theory this
    // could come before defaults are set).
    gCmdLine.parseOpts(argv, argc);

    if (gCmdLine.hasErrors())
        exit(2);

    if (gCmdLine.getVal("clmode") == "yes") {
        gCmdLine.setVal("screen",       "no");
        gCmdLine.setVal("usejoystick",  "no");
        gCmdLine.setVal("audio",        "no");
        gCmdLine.setVal("mode",         "serveronly");
    }

    if (gCmdLine.getVal("updateenv") == "yes") {
        if (General::associateSalpProtocolWithApp())
            cout << "Updated support for salp:// shortcut.\n";
        else 
            cout << "Sorry, could not setup support for salp:// on your system.\n";
    }

    // Warn of common errors with command-line parms
    if (gCmdLine.getVal("host") != "localhost" &&
        gCmdLine.getVal("host") != "127.0.0.1" &&
        gCmdLine.getVal("mode") == "server")
    {
        cerr << "### WARNING\n"
        << "###     You appear to be hosting a game, yet connecting\n"
        << "###     to a different one. Are you sure that's what you\n"
        << "###     want? You probably meant to add \"mode=client\" to\n"
        << "###     your command-line arguments.\n";
    }

    if (gCmdLine.getIntVal("width")  < 320) {
        cerr << "Error: minimum screen width is 320.\n";
        exit(1);
    }

    if (gCmdLine.getIntVal("height") < 240) {
        cerr << "Error: minimum screen height is 320.\n";
        exit(1);
    }

    // Set full screen mode if user requested it
    if(gCmdLine.getVal("screen") == "full")
        gFullScreen = SDL_FULLSCREEN;

    // Set background (scenery) map drawing depending on user's prefs
    Tile::Map::setDrawBackgrounds(gCmdLine.getVal("drawbg") == "yes");
    
    // Set foreground (scenery) map drawing depending on user's prefs
    Tile::Map::setDrawForegrounds(gCmdLine.getVal("drawfg") == "yes");

    // Should I turn on TCP_NODELAY?
    General::useTCPNoDelay = gCmdLine.getVal("tcp_nodelay") == "yes";
    
    
}

void initDisplay()
{
    // ********************************************
    //  Initialize the display
    // ********************************************

    assert(gCmdLine.getVal("screen") != "no");

    const SDL_VideoInfo *info = SDL_GetVideoInfo();
    if(!info) {
        fprintf( stderr, "Video query failed: %s\n",
                 SDL_GetError( ) );
        exit(1);
    }

    int bpp = SDL_GetVideoInfo()->vfmt->BitsPerPixel <= 8 ? 8 : 16;

    // Load icon info - must be called BEFORE the first call to SDL_SetVideoMode
    SDL_Surface *winicon_img = IMG_Load("tiles/winicon.bmp");
    if (!winicon_img) {
        fprintf(stderr, "Couldn't load window icon %s: %s\n", "tiles/winicon.bmp", SDL_GetError());
        exit(1);
    }
    SDL_SetColorKey(winicon_img, SDL_SRCCOLORKEY, SDL_MapRGB(winicon_img->format, 255,0,255));
    SDL_WM_SetIcon(winicon_img, 0);

    screen = SDL_SetVideoMode(gCmdLine.getIntVal("width"),
                              gCmdLine.getIntVal("height"),
                              bpp,
                              
#ifdef USING_OPENGL
                              SDL_GLSDL |
#else
                              SDL_DOUBLEBUF |
#endif
                              SDL_ANYFORMAT |
                            //  SDL_HWPALETTE |
                              SDL_HWSURFACE |
                              gFullScreen); // SDL_RESIZABLE SDL_FULLSCREEN SDL_SWSURFACE

    if ( screen == NULL ) {
        fprintf(stderr, "Couldn't set video mode: %s\n", SDL_GetError());
        exit(1);
    }

    // Give'em something to look at.
    screenBlit_OpenSplash(screen);
}

void initJoysticks()
{
    //
    // Load joystick support, if we have any.
    //
    if (gCmdLine.getVal("usejoystick") == "yes") {
        if (SDL_NumJoysticks() > 0) {
            cout << "Loading joystick support\n";
            int joyid = 0;
            gJoystick = SDL_JoystickOpen(joyid);
            cout << ".. you should use '" << SDL_JoystickName(joyid) << "'\n";
            SDL_JoystickEventState(SDL_ENABLE);

#ifdef WIN32
            string joyininame = "prefs/winjoystick.ini";
#else
            string joyininame = "prefs/linjoystick.ini";
#endif

            if (  !loadKeyboardMap(gJoystickMap, joyininame ) &&
                    configJoystickInteractively(gJoystickMap, SDL_JoystickName(joyid))
               ) {
                cout << "*** SAVED JOYSTICK PREFERENCES ***\n";
                saveKeyboardMap(gJoystickMap, joyininame );
                // loadJoystickMapDefaults(gJoystickMap, SDL_JoystickName(joyid));
            }
        } else
            cout << "#### NO JOYSTICKS FOUND ####\n";
    }
}

void initMusic()
{
    // Load some music for them! (who cares if it sucks)
    //
/*    if(music == NULL) {
          music = Mix_LoadMUS("Burst.ogg");  // load my tune
          if (!music) {
              cerr << "Could not load music file; exiting.\n";
              exit(1);
          }
          Mix_PlayMusic(music, 0);
    //      Mix_FadeInMusic(music, 0, 3000); // play my tune; loop 0 times (-1 for infinite)
          Mix_HookMusicFinished(musicDone);  // and when it's done, call musicDone
    }  
  */  
}

void initOSD() {
    Game::PrefBucket p("prefs/initmedia.ini");

    gOSD = new Game::OSD( gGameState->getTileSet("tiles/score_font", true),
                          gGameState->getTileSet("tiles/health_bar", true),
                          gGameState->getTileSet("tiles/alpha5x5", true),
                          gGameState,
                          screen
                        );
    if (gCmdLine.getVal("screen") == "no")
        gOSD->setEchoSTDOUT(true);

    Tile::Set *fexf = gGameState->getTileSet(p.getVal("OSD__ALPHA5x5", "tiles/alpha5x5"), true);
    fexf->changeColor(SDL_MapRGB(fexf->getSurface()->format, 0, 0, 0), SDL_MapRGB(fexf->getSurface()->format, 51*5, 51*5, 51*5));

    p.save();
}

void initPrefs()
{
    gPrefBucket = new Game::PrefBucket("prefs/general.ini");
    assert(gPrefBucket != 0);
    gPrefBucket->setDefault("PLAYERGENDER", "male"); // female, neuter
    
    gPrefBucket->setDefault("PLAYERNAME", General::getUserName());   
    gPrefBucket->setDefault("SERVERNAME", gPrefBucket->getVal("PLAYERNAME") + "'s Salp Haven");   
    gPrefBucket->setDefault("SS_HOST", "www.salpwars.com"); 
    gPrefBucket->setDefault("SS_PATH_REGISTER", "/cgi-bin/register_game"); 
    gPrefBucket->setDefault("SS_PATH_QUERY", "/cgi-bin/nph-sw_query");
//    gPrefBucket->setDefault("SS_PATH_QUERY", "/cgi-bin/nph-sw_query");
    gPrefBucket->setDefault("SS_UPDATEINTERVAL", "4"); // every N minutes    
    gPrefBucket->setDefault("SS_UPDATETYPE", "text"); // color, grayscale, text
    gPrefBucket->setDefault("CLIENT_MAX_PORT_DRIFT", "10"); // color, grayscale
    gPrefBucket->setDefault("SERVER_MAX_PORT_DRIFT", "10"); // color, grayscale
    assert(gGameState != 0);
    gPrefBucket->save();
    gGameState->setPrefBucket(gPrefBucket);
    
    int i;
    if( gPrefBucket->getVal("mode").substr(0,6) == "server" && (i = gPrefBucket->getIntVal("SS_UPDATEINTERVAL"))) { 
        cout <<   "# WARNING: Your game will be posted to the " << gPrefBucket->getVal("SS_HOST")
             << "\n#          metaserver every " << gPrefBucket->getIntVal("SS_UPDATEINTERVAL") << " minutes. If you do not want this,"
             << "\n#          set it to 0 in prefs/general.ini and restart the game.\n";
    }
}

void initSDL() 
{
    // ********************************************
    //  Initialize the SDL libraries
    // ********************************************

    cout << "Initializing SDL library.\n";

    int init_opts = 0;
    
    if (gCmdLine.getVal("screen") != "no")         init_opts |= SDL_INIT_VIDEO;
    if (gCmdLine.getVal("audio")  != "no")         init_opts |= SDL_INIT_AUDIO;
    if (gCmdLine.getVal("usejoystick")  != "no")   init_opts |= SDL_INIT_JOYSTICK;

    if( SDL_Init(init_opts) < 0 ) {
        fprintf(stdout, "Couldn't initialize SDL: %s\n", SDL_GetError());
        exit(1);
    }

    // Following code taken from SDL doc project (you can tell
    // by the use of "narf" ;-).
    if (gCmdLine.getVal("debug") == "yes"  && (init_opts & SDL_INIT_VIDEO))
    {

        SDL_VideoInfo * narf= (SDL_VideoInfo *) SDL_GetVideoInfo();
        printf("Is it possible to  create hardware surfaces?\n\t");
        printf("hw_available=%d\n",narf->hw_available);
        printf("Is there a window manager available?\n\t");
        printf("wm_available=%d\n",narf->wm_available);
        printf("Are hardware to hardware blits accelerated?\n\t");
        printf("blit_hw=%d\n",narf->blit_hw);
        printf("Are hardware to hardware colorkey blits accelerated?\n\t");
        printf("blit_hw_CC=%d\n",narf->blit_hw_CC);
        printf("Are hardware to hardware alpha bits accelerated?\n\t");
        printf("blit_hw_A=%d\n",narf->blit_hw_A);
        printf("Are software to hardware blits accelerated?\n\t");
        printf("blit_sw=%d\n",narf->blit_sw);
        printf("Are software to hardware colorkey blits accelerated?\n\t");
        printf("blit_sw_CC=%d\n",narf->blit_sw_CC);
        printf("Are software to hardware alpha blits accelerated?\n\t");
        printf("blit_sw_A=%d\n",narf->blit_sw_A);
        printf("Are color fills accelerated?\n\t");
        printf("blit_fill=%d\n",narf->blit_fill);
        printf("\n\n\t\tTotal amount of video_mem:%d\n",narf->video_mem);
    } 

    atexit(SDL_Quit);     // cleanup on exit

    // Initialize the network
    if ( SDLNet_Init() < 0 ) {
        fprintf(stderr, "Couldn't initialize SDL_net library: %s\n",
                SDLNet_GetError());
        exit(1);
    }

    atexit(SDLNet_Quit);  // cleanup on exit

    if (gCmdLine.getVal("audio") != "no") {
        // Initialize audio
        int audio_rate =  44100;           // was 22050, and did not get stereo
        Uint16 audio_format = AUDIO_S16;  // 16-bit stereo
        int audio_channels = 2;           // 2 audio channels (left/right, presumably)
        int audio_buffers = 2048;         // was 4096

        if(Mix_OpenAudio(audio_rate, audio_format, audio_channels, audio_buffers)) {
            printf("Unable to open audio!\n");
            exit(1);
        }

        /* cout << "INITIALIZED AUDIO:\n\tRate:\t" << audio_rate
             << "\n\tFormat:\t" << audio_format << " (16-bit stereo is " << AUDIO_S16 << ")"
             << "\n\tChannels:\t" << audio_channels
             << "\n\tBuffers:\t" << audio_buffers
             << endl;
        */

        // If we actually care about what we got, we can ask here.
        Mix_QuerySpec(&audio_rate, &audio_format, &audio_channels);

        atexit(Mix_CloseAudio);

    } 

}

void initServer()
{
    ::initStaticMedia();
    General::initStaticMedia(); // new way of doing it; migrate here instead of old 

    gServer = new Game::Server(gGameState);
    if (!gServer) {
        cerr << "Not enough memory available to create game server.\n";
        exit(1);
    }

    int start_port = gCmdLine.getIntVal("port")-1;
    int tries = 0;
    bool started_server = false;
    while (++tries <= gPrefBucket->getIntVal("SERVER_MAX_PORT_DRIFT") && !started_server) {
        ++start_port;
        started_server = gServer->listenOn(start_port);
    }

    if (!started_server) {
        cerr << "Could not listen to connections on ports "
        << gCmdLine.getIntVal("port") << "-" << start_port
        << ".\n";
        delete gServer;
        gServer = 0;
        exit(1);
    }

    gCmdLine.setVal("port", start_port);
    cout << "Serving game on port " << gCmdLine.getVal("port") << endl;

    gGameState->setServer(gServer); // put me before loadLevel
    gGameState->loadLevel(gCmdLine.getVal("level"));
    
    assert(gGameState->isGood());
}


void initStaticMedia()
{
    //
    // Load required tiles / sequences.
    //
    Game::PrefBucket p("prefs/initmedia.ini");
    p.erase();
    
    Tile::loadTileSet("tiles/Starshot", gGameState); // REQUIRED

    Game::PlayerStomper::__seq_idle     = gGameState->getTileSequence(p.getVal("PLAYERSTOMPER__SEQ_IDLE", "tiles/Taer/1"), true);
    Game::PlayerStomper::__seq_jump     = gGameState->getTileSequence(p.getVal("PLAYERSTOMPER__SEQ_JUMP", "tiles/Taer/3"), true);
    Game::PlayerStomper::__seq_walk     = gGameState->getTileSequence(p.getVal("PLAYERSTOMPER__SEQ_WALK", "tiles/Taer/4"), true);
    Game::PlayerStomper::__seq_run      = gGameState->getTileSequence(p.getVal("PLAYERSTOMPER__SEQ_RUN", "tiles/Taer/5"), true);
    Game::PlayerStomper::__seq_poofin   = gGameState->getTileSequence(p.getVal("PLAYERSTOMPER__SEQ_POOFIN", "tiles/poofcloud/2"), true);
    Game::PlayerStomper::__seq_poofout  = gGameState->getTileSequence(p.getVal("PLAYERSTOMPER__SEQ_POOFOUT", "tiles/poofcloud/1"), true);
    Game::NPC::__seq_proj_tracer     = gGameState->getTileSequence(p.getVal("PLAYERSTOMPER__SEQ_PROJ_TRACER", "tiles/Starshot/3"), true);
    Game::NPC::__seq_proj_homing     = gGameState->getTileSequence(p.getVal("PLAYERSTOMPER__SEQ_PROJ_HOMING", "tiles/Starshot/1"), true);

    Game::NPCBushBomber::__seq_idle     = gGameState->getTileSequence(p.getVal("NPCBUSHBOMBER__SEQ_IDLE", "tiles/bushbomber/1"), true);
    Game::NPCBushBomber::__seq_walk     = gGameState->getTileSequence(p.getVal("NPCBUSHBOMBER__SEQ_WALK", "tiles/bushbomber/3"), true);
    Game::NPCBushBomber::__seq_peek     = gGameState->getTileSequence(p.getVal("NPCBUSHBOMBER__SEQ_PEEK", "tiles/bushbomber/4"), true);


    gGameState->getTileSequence("tiles/Starshot/1", true);
    gGameState->getTileSequence("tiles/RocketThin/1", true);
    gGameState->getTileSequence("tiles/RocketThin/2", true);
    gGameState->getTileSequence("tiles/RocketThin/3", true);
    gGameState->getTileSequence("tiles/bubble76x48/1", true);

    Game::NPC::__snd_fire   = gGameState->addSound(new Game::Sound(p.getVal("PLAYERSTOMPER__SND_FIRE", "sounds/hurl.wav")));
    Game::PlayerStomper::__snd_jump   = gGameState->addSound(new Game::Sound(p.getVal("PLAYERSTOMPER__SND_JUMP", "sounds/jump.wav")));
    Game::PlayerStomper::__snd_land   = gGameState->addSound(new Game::Sound(p.getVal("PLAYERSTOMPER__SND_LAND", "sounds/land.wav")));
    p.save();
}

void screenBlit_OpenSplash(SDL_Surface *screen) {
    //
    // Loads, blits and frees our splash image to <screen>
    //

    string bmpfname = "graphics/splash_loading.png";

    SDL_Surface *image = IMG_Load(bmpfname.c_str());
    if (!image) {
        fprintf(stderr, "Warning: Couldn't load screenshot %s: %s\n", bmpfname.c_str(), SDL_GetError());
        return; // soft-exit
    }

#ifndef USING_OPENGL
    if(SDL_BlitSurface(image, 0, screen, 0) < 0)
        fprintf(stderr, "BlitSurface error: %s\n", SDL_GetError());

    SDL_Flip(screen);

    SDL_FreeSurface(image);
#endif

}
