#include <iostream>
#include "Tile/Loader.h"
#include "SALPSTUB.h"
#include "Game/ItemPlatform.h"
#include "General/general.h"
#include "Main/init.h"

using namespace Salp::General;
using namespace std;

namespace Salp {
namespace Tile {

void loadLevel(string levname, Game::State *gs, SDL_Surface *screen) 
{
    // NO, because it could run with clmode=yes;  assert(screen != 0);

    // Open our .lev (level) information file.
    ifstream fin;
    fin.open(levname.c_str());
    if(!fin.is_open()) {
        cerr << "Error opening file " << levname << endl;
        exit(1);
    }
        
    Tile::Map *mainMap = 0; // refers to player-level map
    Tile::Map *lastMap = 0; // refers to last map: FOREGROUND, BACKGROUND or SUPER-FOREGROUND
    Game::ItemPlatform *currPlatform = 0;
 
    unsigned int lineno = 0;
    char currline[400];
    char tokenstr[20], tempstr[400];
    unsigned int tempuint;
    int tempint;
    while(!fin.eof() && fin.good())
    {
        // get the next line of the file
        ++lineno;
        fin.getline(currline, 400-1, '\n');

        trim(currline);

        // skip comments
        if(!strlen(currline) ||
           sscanf(currline, " # %s", tempstr) >0 ||
           !strcmp(currline, "#") 
           )
            continue;

        // cout << lineno << " " << currline << endl;

        // otherwise get token name
        if(sscanf(currline, " %25[^ ] %399[^\r]", tokenstr, tempstr) < 1)
        {
            cout << "loadLevel: Unrecognized Line " << lineno << ": " << currline << endl;
        } else if (tokenstr == string("VERSION")) {
        
            unsigned int version;
            if(!sscanf(tempstr, " %d ", &version))
                cout << "(" << lineno << ") Could not parse version.\n";
            else 
                gs->setVersion(version);
        } else if (tokenstr == string("TITLE")) {
            gs->setTitle(tempstr);
        } else if (tokenstr == string("AUTHOR")) {
            gs->setAuthor(tempstr);
        } else if (tokenstr == string("MAP") ||
                   tokenstr == string("BACKGROUND") ||
                   tokenstr == string("FOREGROUND")) {
            char mapname[150];
            char tsfile[150];
            if(!sscanf(tempstr, " %30s %90s ", tsfile, mapname)) {
                cout << "(" << lineno << ") Could not parse filename/mapname.\n";
                exit(1);
            } else {

                // load map (auto-load tileset if necessary)
                lastMap = gs->getTileMap(tsfile + string("/") + mapname, true);
                if(!lastMap) {
                    cout << "(" << lineno << ") Map '" << mapname << "' not found in tileset '" << tsfile << "'\n";
                    exit(1);
                }
                
                if (gCmdLine.getVal("predrawmaps") == "yes"  &&  screen) {
                    // Takes more memory, but uses less CPU in computing the map drawage.
                    lastMap->predrawMap();
                }
                
                lastMap->setUsedInLevel();

                if(tokenstr == string("MAP"))  {
                    mainMap = lastMap;
                } else {
                    if(!mainMap) {
                        cout << "(" << lineno << ") BACKGROUND/FOREGROUND may not be specified before a MAP is specified.\n";
                        exit(1);
                    }
                    
                    lastMap->setRepeatX(true);
                    lastMap->setRepeatY(true);
                    
                    if(tokenstr == string("BACKGROUND")) { mainMap->addBackgroundMap(lastMap); }
                    else { mainMap->addForegroundMap(lastMap); }
                }
           }

        } else if (tokenstr == string("STARTMAP")) {
            assert(mainMap != 0);
            gs->setStartMap(mainMap);

        } else if (tokenstr == string("MAP_ALPHA")) {
            assert(lastMap != 0);
            
            if(!sscanf(tempstr, " %d ", &tempint)) {
                cout << "(" << lineno << ") Could not parse alpha value for " << lastMap->getName() << ".\n";
                exit(1);
            } else if (tempint < 1 || tempint > 254) {
                cout << "(" << lineno << ") Alpha value for " << lastMap->getName() << " should be between 1 (almost transparent) and 254 (almost opaque).\n";
                exit(1);
            } else {
                lastMap->setAlpha(tempint);
            }

        } else if (tokenstr == string("MAP_BGALPHA")) {
            assert(lastMap != 0);
            
            if(!sscanf(tempstr, " %d ", &tempint)) {
                cout << "(" << lineno << ") Could not parse background alpha value for " << lastMap->getName() << ".\n";
                exit(1);
            } else if (tempint < 1 || tempint > 254) {
                cout << "(" << lineno << ") Background alpha value for " << lastMap->getName() << " should be between 1 (almost transparent) and 254 (almost opaque).\n";
                exit(1);
            } else {
                lastMap->setBackgroundAlpha(tempint);
            }
            
        } else if (tokenstr == string("MAP_BGCOLOR")) {
            assert(lastMap != 0);
            
            int r, g, b;
            if(!sscanf(tempstr, " %d %d %d ", &r, &g, &b)) {
                cout << "(" << lineno << ") Could not parse background colors for " << lastMap->getName() << ".\n";
                exit(1);
            } else if (r < 0 || r > 255 || g < 0 || g > 255 || b < 0 || b > 255) {
                cout << "(" << lineno << ") Background color values (r, g, b) should be between 0 and 255.\n";
                exit(1);
            } else if (screen != 0) {
                assert(screen != 0);
                lastMap->setBackgroundColor(SDL_MapRGB(screen->format, r, g, b));
            }

        } else if (!gs->getServer()) {
             
             // ####################################################################
             // # All following level commands are to be used by the server only   #
             // ####################################################################
             
        } else if (tokenstr == string("PLATFORM")) {
            // If we already have a platform in the works, log it
            if (currPlatform) {
                gs->addItem(currPlatform);
                currPlatform = 0;
            }

            assert(currPlatform == 0);

            if(!sscanf(tempstr, " %s ", tempstr)) {
                cout << "(" << lineno << ") Could not parse platform creation data.\n";
                exit(1);
            } else {
                if(gs->getTileSequence(tempstr, true) == 0) {
                    cerr << "Could not load tile sequence '" << tempstr << "'\n";
                    exit(1);
                }
                currPlatform = new Game::ItemPlatform(gs->getTileSequence(tempstr));
                currPlatform->setPath(new Game::PathCP());
                assert(mainMap != 0);
                currPlatform->setMap(mainMap);
            }

        } else if (tokenstr == string("PLATFORM_CONTROLPOINT")) {
            

            assert(currPlatform != 0);

            int x_offset=0, y_offset=0;
            unsigned int mapcode = 0;

            if(sscanf(tempstr, " %x %d %d ", &mapcode, &x_offset, &y_offset) < 1) {
                cout << "(" << lineno << ") Could not parse platform waypoint data.\n";
            } else {
                Sint32 map_x = 0, map_y = 0;
                assert(mainMap != 0);
                if(mainMap->getMapCellCoordsByMapCode(mapcode, map_x, map_y)) {
                    
                    // Add control point to path
                    Game::PathCP  *currPath = (Game::PathCP *) currPlatform->getPath();
                    currPath->addControlPoint(map_x + x_offset, map_y + y_offset);

                    // If this is the first point, also set it to my position.
                    if(currPath->getControlPointCount() == 1)
                        currPlatform->setPos(map_x + x_offset, map_y + y_offset);
                } else {
                    cout << "Could not find mapcode " << mapcode << " in map " << mainMap->getName() << endl;
                    exit(1);
                }
            }

        } else if (tokenstr == string("PLATFORM_BOUNCE_SEQ")) {
            
            assert(currPlatform != 0);

            int bounce_dir=Game::DIR_UP;
            char bounce_seq[1024];

            if(sscanf(tempstr, " %1000s %d ", &bounce_seq[0], &bounce_dir) < 1) {
                cout << "(" << lineno << ") Could not parse platform bounce data.\n";
            } else {
                Tile::Sequence *tseq;
                if ((tseq = gs->getTileSequence(&bounce_seq[0], true)) == 0) {
                    cerr << "Could not load tile sequence '" << tempstr << "'\n";
                    exit(1);
                }
                
                assert(bounce_dir & Game::DIR_ANYWHERE);
                
                currPlatform->setBounceEffect(tseq, bounce_dir);
            }

        } else if (tokenstr == string("PLATFORM_SPEED")) {
            

            assert(currPlatform != 0);

            if(!sscanf(tempstr, " %u ", &tempuint)) {
                cout << "(" << lineno << ") Could not parse platform speed.\n";
                exit(1);
            } else {
                currPlatform->setSpeed(tempuint);
            }


        } else if (tokenstr == string("MAPCODETAB")) {
            
            if(!sscanf(tempstr, " %s ", tempstr))
                cout << "(" << lineno << ") Could not parse filename.\n";
            else 
                cout << "## I noticed your MAPCODE tab. Well, I don't use those yet. Maybe not ever. Don't make me cry.\n";
            // *** INSERT MAPCODE LOADING HERE ****
        } else {
            cout << "(" << lineno << ") Unrecognized LEVEL Token: "
                 << tokenstr << endl;
        }
    }
    
    if(!fin.good() && !fin.eof()) {
        cout << "An error occurred when reading the level file, at line "
             << (lineno)
             << "." <<endl;
        exit(1);
    }    


    // Clean up items-in-the-making
    if (currPlatform)
        gs->addItem(currPlatform);


}

float strToScrollFactor(string str)
{
    if(str.size() == 1)
    {
        return -1;
    } else {
        char op;
        float val;
        if(sscanf((char *) str.c_str() + 1, " %c %f ", &op, &val)) {
            if(op == '*') 
                return(val);
            else if(op == '/')
                return(1/val);
        }

        // If we got this far, there was a parse error.
        // Try just using 1 instead.
        cout << "Could not parse scroll factor of '" << str
             << "'; using '1' instead.\n";
        return -1;
    }
}

Tile::Map * loadTileMap(Tile::Set *tset, ifstream& fin)
{
    
    // read sequence data from file
    string label, name, scrollxstr, scrollystr;
    unsigned short width=0, height=0, tileid;
    unsigned short bounds, mapcode;
    
    fin >> label >> name >> width >> height >> scrollxstr >> scrollystr;
    Tile::Map *tmap = new Tile::Map(tset, width, height);
    
    // Get/set scroll multipliers/info
    //
    float xmult = strToScrollFactor(scrollxstr);
    float ymult = strToScrollFactor(scrollystr);
    Sint32 auto_scrollx=0, auto_scrolly=0;
    
    if (xmult < 0) {
        auto_scrollx = atoi(scrollxstr.c_str());
        xmult = 1;
    }
    
    if (ymult < 0) {
        auto_scrolly = atoi(scrollystr.c_str());
        ymult = 1;
    }
    
    tmap->setScrollMultiplier(xmult, ymult);
    tmap->setAutoScroll(auto_scrollx, auto_scrolly);
    
    
    tmap->setName(name);
    //cout << "Map name '" << name << "', dims "
    //     << width << " X " << height << endl;
        
    for(unsigned short y=0; y<height; y++) 
    {
        for(unsigned short x=0; x<width; x++)
        {
            fin >> tileid >> bounds >> mapcode;
            tmap->setCellData(x, y, tileid, bounds, mapcode);
        }
    }

    return tmap;
}

Tile::Sequence * loadTileSequence(Tile::Set *tset, ifstream& fin)
{
    // read sequence data from file
    string label;
    unsigned short seqid, seqlen, tileid, framedelay;
    
    fin >> label >> seqid >> seqlen;
    Tile::Sequence *tseq = new Tile::Sequence(tset, seqlen);

    // set its name to its id
    char seqname[10];
    sprintf(seqname, "%d", seqid);
    tseq->setName(seqname);
    
    //cout << "Sequence " << seqid << ": " << endl;
    for(unsigned short frameid=0; frameid<seqlen; frameid++) 
    {
        fin >> tileid >> framedelay;
        tseq->addFrame(tileid-1, framedelay*10);
        //cout << "Frame " << frameid << ": Tile " << tileid << endl;
    }
    
    return tseq;
}


void loadTileSet(string tsetname, Game::State *gs, SDL_Surface *screen)
{   
    // If we already have this Tile::Set in the gamestate,
    // then complain.
    assert(gs->getTileSet(tsetname) == 0);
  
    //
    // Otherwise, load it!
    //
    string tsetfname = tsetname + ".tset";
    
//    cout << "Loading TileSet '" << tsetfname << "'\n";

    // Open our .tset (Tileset) information file.
    ifstream fin;
    fin.open(tsetfname.c_str());
    if(!fin.is_open()) {
        cerr << "Error opening file '" << tsetfname << "'\n";
        return;
    }
    
    string tlabel;
    
    Tile::Set *ts = new Tile::Set;
    if(!ts)
    {
        cout << "Could not allocate memory for tileset.\n";
        exit(1);
    }
    
    // Set name (filepath minus .tset)
    ts->setName(tsetname); 

    // Load the bitmap associated with this tileset.
    ts->loadImage(tsetname + ".bmp", screen);
    
    // Read file labels, and parse them accordingly.
    while(!fin.eof() && fin.good()) {
    
        fin >> tlabel;
        //cout << "Label: " << tlabel << endl;

        if(fin.eof())
            break;

        if(tlabel == "TILESET") {
            // IGNORED
            string tsname;
            fin >> tsname;
            ////////ts->setName(tsname);
        } else if (tlabel == "TRANSPARENTCOLOR") {
            unsigned short r, g, b;
            fin >> r >> g >> b;
            ts->setTransparentColor(r, g, b);
        } else if (tlabel == "TILECOUNT") {
            Uint32 tilecount;
            fin >> tilecount;
            ts->setCount(tilecount);
        } else if (tlabel == "TILEDIM") {
            unsigned short tilew, tileh;
            fin >> tilew >> tileh;
            ts->setWidth(tilew);
            ts->setHeight(tileh);
        } else if (tlabel == "SEQUENCES") {
            unsigned short seqcount;
            fin >> seqcount;
            for(unsigned int seq = 0; seq < seqcount; seq++) {
                Tile::Sequence *tileseq = loadTileSequence(ts, fin);
                tileseq->setName(tsetname + "/" + tileseq->getName());
                gs->addTileSequence(tileseq);
            }
        } else if (tlabel == "MAPS") {
            unsigned short mapcount;
            fin >> mapcount;
            for(unsigned int map=0; map < mapcount; map++) {
                Tile::Map *tilemap = loadTileMap(ts, fin);
                tilemap->setName(tsetname + "/" + tilemap->getName());
                gs->addTileMap(tilemap);
            }
        } else {
            cerr << "Encountered unknown label when parsing "
                 << tsetfname << ": " << tlabel << endl;
            exit(1);
        }
        
        
        if(fin.bad()) { cout << "ERROR when reading fin\n"; break; }
    }

    // By now we're done loading, so it's safe to verify
    // the integrity of what we've just created.
    ts->verifyIntegrity();

    fin.close();
    
    gs->addTileSet(ts);
}

} // namespace Tile
} // namespace Salp
