#include "Tile/Map.h"

#include <iostream>
#include <assert.h>
#include "General/graphics.h"
#include "General/media.h"
#include "SALPSTUB.h"
#include "Game/ItemPowerUp.h"
#include "Game/WeaponStatus.h"
#include "Game/State.h"

namespace Salp {
namespace Tile {

bool   Map::_draw_backgrounds = true;
bool   Map::_draw_foregrounds = true;

Map::Map(const Tile::Set *tileset,
         const unsigned int width,
         const unsigned int height)
{
    _tileset = (Tile::Set *) tileset;
    _height = height;
    _width = width;
    _xscrollmult = 1;
    _yscrollmult = 1;
    _pixwidth = tileset->getWidth() * width;
    _pixheight = tileset->getHeight() * height;
    assert(_pixheight != 0);
    assert(_pixwidth != 0);
    _repeatx = _repeaty = false;
    _fullmap = 0;
    _usedinlevel = false;
    _alpha = SDL_ALPHA_TRANSPARENT; // same as opaque, except no messing with alpha commands 
    _bgcolor = 0;
    _bgalpha = SDL_ALPHA_TRANSPARENT;
    _maptype = 0;
    _auto_scroll = false;
    _auto_scroll_xrate = 0;
    _auto_scroll_yrate = 0;
    _xpos = 0;
    _ypos = 0;
    _gs = 0;
    // allocate map cells
    _mapcells =  new MapCell*[width];
    assert(_mapcells != 0);
    for(unsigned int i=0; i<width; ++i) {
        _mapcells[i] = new MapCell[height];
        assert(_mapcells[i] != 0);
    }


    if(!_mapcells) {
        cerr << "Could not load mapcells.\n";
        exit(1);
    }
}

Map::~Map() {
    unsigned int i;

    // free map cells
    //
    assert(_mapcells != 0);
    for(i=0; i<_width; i++) {
        assert(_mapcells[i] != 0);
        delete[] _mapcells[i];
    }

    delete[] _mapcells;
    
}

void Map::addBackgroundMap(const Map *tm)
{
    _bmaps.push_back((Map *) tm);
}

void Map::addForegroundMap(const Map *tm)
{
    _fmaps.push_back((Map *) tm);
}

void Map::addItemAtMapCell(Game::Item *gi, unsigned int x, unsigned int y)
{
    assert(gi != 0);
    assert(_gs != 0);
    assert(_tileset != 0);
    
    gi->setMap(this);
    gi->setPos(x * _tileset->getWidth(), y * _tileset->getHeight());
    
    _gs->addItem(gi);
}


void Map::display(SDL_Surface *screen,
                  const Sint32 leftx,
                  const Sint32 topy)
{

    unsigned int i;
    
    // draw all background stuffs
    if (_draw_backgrounds) {
        for(i=_bmaps.size(); i--; ) {
            _bmaps[i]->display(screen, leftx, topy);
        }
    } else {
        for(i=_bmaps.size(); i--; ) {
            _bmaps[i]->drawBackgroundColor(screen);
        }
    }
    
    // draw me!
    drawBackgroundColor(screen);
    if (_auto_scroll)
        drawMapRepeated(screen, leftx+_xpos, topy+_ypos, _repeatx, _repeaty);
    else
        drawMapRepeated(screen, Sint32(leftx*_xscrollmult), Sint32(topy*_yscrollmult), _repeatx, _repeaty);
    
}

void Map::displayForeground(SDL_Surface *screen,
                  const Sint32 leftx,
                  const Sint32 topy)
{    
    // draw all foreground stuffs
    if (_draw_foregrounds) {
        for(unsigned int i=_fmaps.size(); i--; )
            _fmaps[i]->display(screen, leftx, topy);
    }
}

void Map::drawBackgroundColor(SDL_Surface *screen)
{
    // draw initial background color
    if (_bgalpha != SDL_ALPHA_TRANSPARENT)
        fillRectAlpha(screen, 0, _bgcolor, _bgalpha); 
    
    return;
}

unsigned char Map::getId() const
{
    return _id;
}

MapCell * Map::getMapCell(const int x, const int y) const
{
    // given a map cell index, get the map cell
    assert(x >= 0);
    assert(x < (int) _width);
    assert(y >= 0);
    assert(y < (int) _height);
    
    return(&_mapcells[x][y]);

}

MapCell * Map::getMapCellByCoords(const Sint32 x,
                                  const Sint32 y) const
{
    // given a map cell index, get the map cell
    return(getMapCell(getMapCellIndexX(x), getMapCellIndexY(y)));

}

void Map::getMapCellCoords(const unsigned int ix,
                           const unsigned int iy,
                           Sint32& x,
                           Sint32& y) const {
     
     //
     // Sets x and y to the upper left map coords of a mapcell's tile
     // given the mapcell's index value in _mapcells[][].
     //
     
     assert(ix > 0);
     assert(ix < _width);
     assert(iy > 0);
     assert(iy < _height);
     
     x = ix * _tileset->getWidth();
     y = iy * _tileset->getHeight();
}
                       
bool Map::getMapCellCoordsByMapCode(unsigned int mapcode, Sint32& x, Sint32& y, unsigned int startx, unsigned int starty) const
{
    for(unsigned int scany=starty; scany<_height; scany++)
        for(unsigned int scanx=startx; scanx<_width; scanx++) 
            if(_mapcells[scanx][scany].mapcode == mapcode)
            {
                 x = scanx * _tileset->getWidth();
                 y = scany * _tileset->getHeight();

                 return true;
            };

    return false;
}

void Map::getMapCellIndex(const Sint32 x,
                          const Sint32 y,
                          unsigned int &ix,
                          unsigned int &iy) const
{
    // given a tile's position (x, y) on the map, set ix and iy
    // to the map cell index. if an x or y value is off the map, the
    // closest valid index is chosen.
    
    ix = getMapCellIndexX(x);
    iy = getMapCellIndexY(y);
        
    return;
}

unsigned int Map::getMapCellIndexX(const Sint32 x) const
{
    if(x < 0)
        return(0);
    else {
        unsigned int ix = x/_tileset->getWidth();                  
        if(ix >= _width) ix = _width - 1;
        assert(ix < _width);
        return(ix);
    }    
}

unsigned int Map::getMapCellIndexY(const Sint32 y) const
{
    if(y < 0)
        return(0);
    else {
        unsigned int iy = y/_tileset->getHeight();                  
        if(iy >= _height) iy = _height - 1;
        assert(iy < _height);
        return(iy);
    }    
}

unsigned int Map::getMapHeight() const {
    return _height;
}

unsigned int Map::getMapWidth() const {
    return _width;
}

string Map::getName() const {
    return _name;
}

unsigned int Map::getPixHeight() const {
    return _pixheight;
}

unsigned int Map::getPixWidth() const {
    return _pixwidth;
}


float Map::getScrollMultiplierX() const {
    return _xscrollmult;
}

float Map::getScrollMultiplierY() const {
    return _yscrollmult;
}

Tile::Set * Map::getTileSet() const {
    return _tileset;
}

bool Map::getUsedInLevel() const
{
    return _usedinlevel;
}

void Map::interpretMapCodes()
{
    using namespace Salp::Game;
    
    assert(_gs != 0);
    if (isBackground() || isForeground())
        return;
    
    
    for(unsigned int scany=0; scany<_height; scany++)
        for(unsigned int scanx=0; scanx<_width; scanx++) {
            Game::ItemPowerUp *gipu = 0;
         
            switch(_mapcells[scanx][scany].mapcode) {
                case MC_WEAPON_SHOTGUN:
                    gipu = new Game::ItemPowerUp(General::seq_powerup_shotgun);
                    gipu->setWeaponStatus(new Game::WeaponStatus(Game::WEAPON_SHOTGUN, true, 10));
                    break;
                case MC_WEAPON_SPREAD:
                    gipu = new Game::ItemPowerUp(General::seq_powerup_spreadshot);
                    gipu->setWeaponStatus(new Game::WeaponStatus(Game::WEAPON_SPREAD, true, 10));
                    break;
                case MC_WEAPON_SALP_FROGS:
                    gipu = new Game::ItemPowerUp(General::seq_powerup_salp_frog);
                    gipu->setWeaponStatus(new Game::WeaponStatus(Game::WEAPON_SALP, true, 10));
                    break;
                case MC_WEAPON_SALP_GRAPES:
                    gipu = new Game::ItemPowerUp(General::seq_powerup_salp_grape);
                    gipu->setWeaponStatus(new Game::WeaponStatus(Game::WEAPON_SALP, true, 10));
                    break;
                case MC_WEAPON_HOMING:
                    gipu = new Game::ItemPowerUp(General::seq_powerup_homing);
                    gipu->setWeaponStatus(new Game::WeaponStatus(Game::WEAPON_HOMING, true, 10));
                    break;
                case MC_HEALTH_20:
                    gipu = new Game::ItemPowerUp(General::seq_powerup_health);
                    gipu->setHP(20);
                    break;
                case MC_START_TEAM_A:
                    _gs->addTeamStartLoc(0, StartLoc(this, Vector(scanx, scany)));
                case MC_START_TEAM_B:
                    _gs->addTeamStartLoc(1, StartLoc(this, Vector(scanx, scany)));
                case MC_START_TEAM_C:
                    _gs->addTeamStartLoc(2, StartLoc(this, Vector(scanx, scany)));
                case MC_START_TEAM_D:
                    _gs->addTeamStartLoc(3, StartLoc(this, Vector(scanx, scany)));
                case 0x00: // no mapcode; default.
                    break;
//                default:
//                    cerr << "Warning: uninterpreted mapcode: " << _mapcells[scanx][scany].mapcode << endl;
            }
        
            if (gipu != 0)   addItemAtMapCell(gipu, scanx, scany);
        }
}

bool Map::isDefinedMapCode(unsigned int mapcode) const
{
    for(unsigned int scany=0; scany<_height; scany++)
        for(unsigned int scanx=0; scanx<_width; scanx++) 
            if(_mapcells[scanx][scany].mapcode == mapcode)
                 return true;
    return false;
}

bool Map::needsAnimation()
{
    return _auto_scroll;
}

void Map::predrawMap()
{
    assert(!_fullmap);
    
    // if we already have one, don't do it again
    if(_fullmap) return;
 
    // otherwise create the surface (with the full dims of the map)
    SDL_Surface *fullmap = _tileset->getSimilarSurface(
                         _width * _tileset->getWidth(),
                         _height * _tileset->getHeight()
                      );
    if(!fullmap) return;
    
    // Fill it with transparency
    SDL_FillRect(fullmap,0,_tileset->getTransparentColor());
    
    // Okay, so we have the surface allocated. Now just draw
    // the map to this screen, and we're game.
    drawMap(fullmap, 0, 0);
    
    SDL_SetColorKey(fullmap, SDL_SRCCOLORKEY | SDL_RLEACCEL,
                _tileset->getTransparentColor());

    // Then slide'er over
    _fullmap = SDL_DisplayFormat(fullmap);
    
    assert(_fullmap != 0);
    SDL_FreeSurface(fullmap);
    
    return;
}

void Map::setAutoScroll(Sint32 xrate, Sint32 yrate)
{
    _auto_scroll_xrate = xrate;
    _auto_scroll_yrate = yrate;
    _auto_scroll = xrate || yrate ? true : false;
}


void Map::setCellData(unsigned short x, unsigned short y, 
             unsigned short tileid, unsigned char bounds,
             unsigned char mapcode) {
    /*
        _mapcells should be in the form of
        
        (0, 0) ... ... ... ...
        ...
        ...
        ...
        ...
                              (w-1, h-1)
    
    */
    
    MapCell *cell = &(_mapcells[x][y]);
    
    cell->tileid = tileid;
    cell->bounds = bounds;
    cell->mapcode = mapcode;

}

void Map::setName(string name)
{
    _name = name;
}

void Map::setRepeatX(bool repeatit)
{
    _repeatx = repeatit;
}

void Map::setRepeatY(bool repeatit)
{
    _repeaty = repeatit;
}

void Map::setScrollMultiplier(float xmult, float ymult)
{
     _xscrollmult = xmult;
     _yscrollmult = ymult;
}

void Map::setUsedInLevel()
{
    _usedinlevel = true;
}

bool Map::verifyIntegrity()
{
    assert(isDefinedMapCode(MC_START_TEAM_A));
//    assert(isDefinedMapCode(MC_START_TEAM_B));
//    assert(isDefinedMapCode(MC_START_TEAM_C));
//    assert(isDefinedMapCode(MC_START_TEAM_D));
    return true;
}


} // namespace Tile
} // namespace Salp
