#ifndef __Map_H
#define __Map_H

#include <string>
#include <vector>
#include <cassert>
#include <cmath>
#include "SALP_SDL.h"
#include "SALPSTUB.h"
#include "Tile/Set.h"
#include "Game/Item.h"

#define MC_TOP_BOUND    1
#define MC_LEFT_BOUND   2
#define MC_BOTTOM_BOUND 4
#define MC_RIGHT_BOUND  8


namespace Salp {

namespace Tile {

// User-specified mapcodes found in the map files
// ----------------------------------------------
enum {
   MC_START_TEAM_A        = 0xf0,  // Starting location(s) for Team A
   MC_START_TEAM_B        = 0xf1,  // Starting location(s) for Team B
   MC_START_TEAM_C        = 0xf2,  // Starting location(s) for Team C
   MC_START_TEAM_D        = 0xf3,  // Starting location(s) for Team D
   MC_WEAPON_SHOTGUN      = 0xe0,  // Spawnpoint for shotgun weapon 
   MC_WEAPON_SPREAD       = 0xe1,  // Spawnpoint for spreadshot weapon
   MC_WEAPON_SALP_FROGS   = 0xe2,  // Spawnpoint for frog-salp ammo/weapon
   MC_WEAPON_SALP_GRAPES  = 0xe3,  // Spawnpoint for grape-salp ammo/weapon
   MC_WEAPON_HOMING       = 0xe4,  // Spawnpoint for homing missle weapon
   MC_HEALTH_20           = 0xd0   // Spawnpoint for +20 health box
};

enum {
    MAPTYPE_BACKGROUND=1,
    MAPTYPE_FOREGROUND=2
};


class MapCell {
    public:
        unsigned int tileid;
        unsigned int bounds;
        unsigned int mapcode;
    private:
};


class Map {
    public:
        explicit Map(const Tile::Set *tileset,
                     const unsigned int width,
                     const unsigned int height);
        ~Map();
        
        void addBackgroundMap(const Map *tm);
        void addForegroundMap(const Map *tm);
        void addItemAtMapCell(Game::Item *gi, unsigned int x, unsigned int y);
        void animate(unsigned int cycles) {
            assert(_auto_scroll);
            assert(cycles != 0);
            //if (_auto_scroll)
                _xpos += _auto_scroll_xrate * cycles;
                _ypos += _auto_scroll_yrate * cycles;
            //}
        }
        
        void display(SDL_Surface *screen, const Sint32 leftx, const Sint32 topy);
        void displayForeground(SDL_Surface *screen, const Sint32 leftx, const Sint32 topy);
        
        void drawBackgroundColor(SDL_Surface *screen);
        
        void drawMap(SDL_Surface *screen,
                     const Sint32 leftx,
                     const Sint32 topy) const;
        void drawMapRepeated(SDL_Surface *screen,
                             const Sint32 leftx,
                             const Sint32 topy,
                             const bool repeatx=true,
                             const bool repeaty=true) const;
        unsigned char getId() const;
        MapCell *getMapCell(const int x, const int y) const;
        MapCell *getMapCellByCoords(const Sint32 x, const Sint32 y) const;
        void getMapCellCoords( const unsigned int ix,
                               const unsigned int iy,
                               Sint32& x,
                               Sint32& y) const;
        bool getMapCellCoordsByMapCode(unsigned int mapcode, Sint32& x, Sint32& y, unsigned int startx=0, unsigned int starty=0) const;

        
        void getMapCellIndex(
                          const Sint32 x,
                          const Sint32 y,
                          unsigned int &ix,
                          unsigned int &iy) const;
        unsigned int getMapCellIndexX(const Sint32 x) const;
        unsigned int getMapCellIndexY(const Sint32 y) const;
        unsigned int getMapHeight() const;
        unsigned int getMapWidth() const;
        string getName() const;
        unsigned int getPixHeight() const; // cell rows * tile::set height
        unsigned int getPixWidth() const;  // cell cols * tile::set width
        float getScrollMultiplierX() const;
        float getScrollMultiplierY() const;
        Tile::Set * getTileSet() const;
        bool getUsedInLevel() const;
        void interpretMapCodes();
        bool isBackground() const { return (_maptype & MAPTYPE_BACKGROUND) != 0; }
        bool isDefinedMapCode(unsigned int mapcode) const;
        bool isForeground() const { return (_maptype & MAPTYPE_FOREGROUND) != 0; }
        bool needsAnimation();
        void predrawMap();
        
        void setCellData(const unsigned short x,
                         const unsigned short y, 
                         const unsigned short tileid,
                         const unsigned char bounds,
                         const unsigned char mapcode);
        static void setDrawForegrounds(bool should_draw) { _draw_foregrounds = should_draw; }
        static void setDrawBackgrounds(bool should_draw) { _draw_backgrounds = should_draw; }
        void setAlpha(Uint8 alpha) { _alpha = alpha; }
        void setAutoScroll(Sint32 xrate, Sint32 yrate);  
        void setBackgroundAlpha(Uint8 bgalpha) { _bgalpha = bgalpha; }
        void setBackgroundColor(Uint32 bgcolor) { _bgcolor = bgcolor; }
        void setGameState(Game::State *gs) { _gs = gs; }
        void setId (const unsigned char id) { _id = id; }
        void setName(const string name);
        void setRepeatX(const bool repeatit);
        void setRepeatY(const bool repeatit);
        void setScrollMultiplier(const float xmult,
                                 const float ymult);
        
		void setUsedInLevel();
        bool verifyIntegrity();
        friend class Salp::Game::Item;
        
    private:
        Tile::Set    *_tileset; // parent's tile set ( dont delete at destruct )
        SDL_Surface  *_fullmap; // full drawring of the map (delete at destruct)
        MapCell     **_mapcells; // mapcell array (delete at destruct)
        unsigned int  _width,        _height; // number of tiles in _mapcells array
        unsigned int  _pixwidth,     _pixheight;
        float         _yscrollmult,  _xscrollmult;
        bool          _repeatx,      _repeaty; // repeat x/y patterns as appropriate?
		bool          _usedinlevel; // is this map used in the current level?
        static bool   _draw_backgrounds;
        static bool   _draw_foregrounds;
        string _name;
        vector<Map *> _bmaps; // background maps (back to front) ( dont delete)
        vector<Map *> _fmaps; // maps to be drawn in front of this map (back to front)
        unsigned char _id;    // the map id
        Uint8 _alpha;
        Uint8 _bgalpha;
        Uint32 _bgcolor;
        int _maptype;
        Game::State *_gs;
        // auto-scroll data
        bool _auto_scroll; // Should the map auto-scroll?
        int _auto_scroll_xrate, _auto_scroll_yrate; // If _auto_scroll is true, this is the xrate and yrate of the scroll 
        Sint32 _xpos, _ypos; // Position of the map to draw at (sort of - it's only really used for auto-scrolling so far.. the other types have x/y passed in the function call. be careful ;-)).
};

//
// Include inline functions here for proper linkage.
//

inline void Map::drawMap(SDL_Surface *screen,
                         const Sint32 leftx,
                         const Sint32 topy) const
{
    // We want to draw the map so that the corner of
    // (leftx, topx) is at the top of the screen.
    
    // If we have a full map in memory, copy it over and
    // be done with the map draw.
    if(_fullmap) {
            SDL_Rect rectfrom, rectto;
    
            rectfrom.x = leftx;
            rectfrom.y = topy;
            rectfrom.w = screen->w;
            rectfrom.h = screen->h;
            
            memcpy(&rectto, &rectfrom, sizeof(rectto));    
            rectto.x = 0;
            rectto.y = 0;
                             
            assert(screen->format->BitsPerPixel ==
                   _fullmap->format->BitsPerPixel);
                   
            if(SDL_BlitSurface(_fullmap, &rectfrom, screen, &rectto) < 0)
                fprintf(stderr, "BlitSurface error: %s\n", SDL_GetError());
                
            return;
    } 

    //
    // Otherwise, copy the map to the screen tile-by-tile.
    //
    
    int x, y;
    int xmax, ymax, xno, yno, theight, twidth;
    theight = _tileset->getHeight();                  
    twidth  = _tileset->getWidth();                  
    
    // init the for-loop
    xno=0;
    x = leftx / twidth;
    xmax = x + screen->w/twidth + ((x * twidth < leftx) ?1:0);

    if(x<0)
    {
        xno = -x;
        x = 0;
    }
    
    
    for(;        
        (x <= xmax)         &&
        (x < (int) _width);
        
        x++, xno++)
    {        
        
        // init the y for loop
        yno  = 0,
        y    = topy/theight,
        ymax = y + screen->h/theight + ((y*theight < topy)?1:0);

        if(y<0) { 
            yno = -y;
            y = 0;
        }
        
        MapCell *ystrip = _mapcells[x] + y; // increment this pointer instead of referring to
                                            // _mapcells[x][y] all the time (less mult'ing)

        for(;
            
            y<=ymax && y < (int) _height;
            
            y++, yno++, ystrip++)
        {
           
            // map val is 0 if no cells are here
            if(ystrip->tileid == 0)
                continue;
            
            _tileset->drawImage(screen, ystrip->tileid-1,
                    xno * twidth - leftx%twidth, yno * theight - topy%theight, _alpha);
        
        }
        
    }
    
}

inline void Map::drawMapRepeated(SDL_Surface *screen,
                                 const Sint32 leftx,
                                 const Sint32 topy,
                                 const bool repeatx,
                                 const bool repeaty) const
{
    assert(leftx >= 0);
    assert(topy >= 0);
    
    // Warning: (int) may be bad in the future if maps are huge
    //          .. though the parallax maps shouldnt be this large.
    
    Sint32 y, x;
    // +----+----+----+----+----+----+----+----+
    // +---------+---------+---------+
    // 0         100       200       300
    if(!repeatx && !repeaty) {
        drawMap(screen, leftx, topy);
    } else if(!repeatx) {    
        for(y=topy % _pixheight; 
            y > -screen->h;
            y -= _pixheight) 
                drawMap(screen, leftx, y);
              
    } else if(!repeaty) {   
        // ***************************************** 
        for(x=leftx % _pixwidth; 
            x > -screen->w;
            x -= _pixwidth) {
                drawMap(screen, x, topy);
            }
    } else {    
        for(x=leftx % _pixwidth; 
            x > -screen->w;
            x -= _pixwidth)
        for(y=topy % _pixheight; 
            y > -screen->h;
            y -= _pixheight) 
               drawMap(screen, x, y);
    }

}

} // namespace Tile
} // namespace Salp

#endif
