#include "SALP_SDL.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <iostream>

#include "Tile/Set.h"
#include "General/graphics.h"
#include <fstream>

using namespace std;
using namespace Salp::General;

namespace Salp {
namespace Tile {


Set::Set()
{
    _image = 0;
    _tilew = _tileh = _tilecount = _transparentcolor = 0;
    _hflipped = _hastransparentcolor = false;

    return;
}

Set::~Set()
{
    /* Free the allocated BMP surface */
    if(_image != 0)
        SDL_FreeSurface(_image);
}

void Set::changeColor(Uint32 cfrom, Uint32 cto)
{
    // Changes all instances of color <cfrom> to the color <cto>, 
    // linearly.
    //

    if(SDL_LockSurface(_image) != 0)
        cerr << "Couldn't lock surface for Tile::Set.\n";
    
    Uint32 pixcolor;
    for (int x=_image->w-1; x>=0; --x)
        for (int y=_image->h-1; y>=0; --y) {
            pixcolor = getPixel(x, y);
            if (pixcolor == cfrom) {
                putPixel(_image, x, y, cto);
            }
        }

    SDL_UnlockSurface(_image);
}

void Set::drawImage(SDL_Surface *screen, unsigned short tileid, int x, int y, int alpha)
{

    // assert that we should even be drawing this thing
    // at all
    assert(x < screen->w + _tilew);
    assert(x > -_tilew);
    assert(y < screen->h + _tileh);
    assert(y > -_tileh);

    if(alpha) { 
        SDL_SetAlpha(_image, SDL_SRCALPHA, alpha);
    }

    SDL_Rect rectfrom, rectto;

    rectfrom.x = 0;
    rectfrom.y = tileid * _tileh;
    rectfrom.w = _tilew;
    rectfrom.h = _tileh;

    if(x < 0) {
        // Clip left side of screen
        //

        // advance the eaten portion to the
        // part we can actually draw.
        // (if x < 0, then we are going to miss
        //  -x pixels)
        rectfrom.x += -x;
        rectfrom.w -= -x;

        // start drawing where allowed
        x = 0;
    } else if (x+rectfrom.w > screen->w) {
        // Clip right side of screen
        //

        rectfrom.w = screen->w - x;
    }

    if(y < 0) {
        // Clip top of screen
        //

        // advance the eaten portion to the
        // part we can actually draw
        // (if y < 0, then we are going to miss
        //  -y pixels)
        rectfrom.y += -y;
        rectfrom.h -= -y;

        // start drawing where allowed
        y = 0;
    } else if (y + rectfrom.h > screen->h) {
        // Clip bottom of screen
        //
        rectfrom.h = screen->h - y;
    }

    memcpy(&rectto, &rectfrom, sizeof(rectto));
    rectto.x = x;
    rectto.y = y;

    // Blit onto the screen surface
    if(SDL_BlitSurface(_image, &rectfrom, screen, &rectto) < 0)
        fprintf(stderr, "BlitSurface error in Set::drawImage (%s): %s\n", getName().c_str(), SDL_GetError());

    // Make sure this portion of the screen is updated
    // -- unnecessary since we're using SDL_Flip
    // SDL_UpdateRect(screen, x, y, _tilew, _tileh);

    if(alpha) {
        SDL_SetAlpha(_image, SDL_SRCALPHA, SDL_ALPHA_OPAQUE);
    }

}

void Set::drawImageHFlipped(SDL_Surface *screen, unsigned short tileid, int x, int y, int alpha)
{

    // assert that we should even be drawing this thing
    // at all
    assert(x < screen->w + _tilew);
    assert(x > -_tilew);
    assert(y < screen->h + _tileh);
    assert(y > -_tileh);


    if(alpha) SDL_SetAlpha(_image, SDL_SRCALPHA, alpha);

    SDL_Rect rectfrom, rectto;

    rectfrom.y = tileid * _tileh;
    rectfrom.h = _tileh;

    rectfrom.w = 1;
    rectfrom.x = _tilew-1;
    int pic_firstx = 0; // last col on MASTER to even consider scanning

    if(x < 0) {
        rectfrom.x -= -x;
        x = 0;
    } else if (x+rectfrom.w > screen->w) {
        rectfrom.x = screen->w - x  -1;
        pic_firstx -= x;
    }

    if(y < 0) {
        rectfrom.y += -y;
        rectfrom.h -= -y;
        y = 0;
    } else if (y + rectfrom.h > screen->h) {
        rectfrom.h = screen->h - y;
    }


    memcpy(&rectto, &rectfrom, sizeof(rectto));
    rectto.x = x;
    rectto.y = y;

    // Blit image to screen.
    // Scans MASTER columns from RIGHT to LEFT.
    // Scans SURFACE columns from LEFT to RIGHT
    while(rectfrom.x >= pic_firstx) {
        // blit
        if(SDL_BlitSurface(_image, &rectfrom, screen, &rectto) < 0)
            fprintf(stderr, "BlitSurface error: %s\n", SDL_GetError());

        // next col
        rectto.x++;
        rectfrom.x--;
    }

    if(alpha) SDL_SetAlpha(_image, SDL_SRCALPHA, SDL_ALPHA_OPAQUE);

}


string Set::getName() const
{
    return _name;
}

Uint32 Set::getPixel(unsigned short x, unsigned short y)
{    int bpp = _image->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)_image->pixels + y * _image->pitch + x * bpp;

    switch(bpp) {
    case 1:
        return *p;

    case 2:
        return *(Uint16 *)p;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;

    case 4:
        return *(Uint32 *)p;

    default:
        return 0;       /* shouldn't happen, but avoids warnings */
    }
}

SDL_Surface *Set::getSimilarSurface(Uint32 width, Uint32 height)
{
    // Returns a pointer to a new SDL_Surface, whose properties
    // are similar to mine (i.e. masks, bpp, etc)
    SDL_Surface *newsurface = SDL_CreateRGBSurface(
                                  SDL_SWSURFACE,
                                  //_image->pixels,//SDL_SRCCOLORKEY, /* SDL_SRCALPHA */
                                  width, height,
                                  _image->format->BitsPerPixel,
                                  _image->format->Rmask, _image->format->Gmask,
                                  _image->format->Bmask, _image->format->Amask);

    // Didn't get it? Stop now!
    if(!newsurface) return 0;

    // Else, set transparent color too!
    if(_hastransparentcolor)
        SDL_SetColorKey(newsurface, SDL_SRCCOLORKEY | SDL_RLEACCEL,
                        _transparentcolor);

    // NOTE: We don't DisplayFormat it yet, since we want to take
    //       advantage of any acceleration we can get after images are
    //       copied over.

    //// finally format it to the new surface.
    //newsurface = SDL_DisplayFormat(newsurface);


    return newsurface;
}

int Set::getCount() const
{
    return(_tilecount);
}

Uint32 Set::getTransparentColor() const
{
    return(_transparentcolor);
}

void Set::loadImage(string bmpfname, SDL_Surface *screen)
{
    // Loads image at filepath bmpfname and attempts to accelerate the image.
    //

    // if we already have an image, free it first
    if(_image)
        SDL_FreeSurface(_image);

    // load the BMP file into a surface or die
    _image = SDL_LoadBMP(bmpfname.c_str());
    if (!_image) {
        fprintf(stderr, "Couldn't load %s: %s\n", bmpfname.c_str(), SDL_GetError());
        exit(1);
    }

/*
    ** We'll try without this for now -- we were getting errors with this
    ** related to SDL_SetAlpha on HWSURFACEs (under directx, maybe others).
    ** If it helps to reinstate this, give it a shot, but make sure it's
    ** not a hwsurface.. or it works with SDL_SetAlpha. Sigh!

    According to kokido from irc.us.openprojects.net:
    [15:43:05] <kokido> calls to SDL_SetAlpha, SDL_SetColorkey need to be
                        done before calls to SDL_Displayformat[Alpha] for
                        HWSURFACES.

    So basically, if I'm going to vary the alpha values a lot (or maybe
    at all) with a given surface, I should just stick to a SWSurface.
    In the future it might be worthwhile to actually store Tile::Set info
    on whether the tile will *ever* be able to be alpha-valued, so we can
    take advantage of the HWSURFACE stuff. At this point, we don't use alpha
    stuff that much. Thanks, kokido!

    // format a new version of the image, optimized for our screen
    SDL_Surface *accelimage = SDL_DisplayFormatAlpha(_image);

    // didn't get the reformatted version? Bail!
    if(!accelimage) {
        cout << "Could not allocate accelerated image for " << bmpfname << ".\n";
        exit(1);
    }

    // free old image
    SDL_FreeSurface(_image); 

    // move reformatted version in
    _image = accelimage;
*/

    if (screen && !(screen->flags & SDL_HWSURFACE)) {
        // optimize me!
        // (note: really we'd do better as a HWSURFACE but this can
        //  get us into trouble when we do the alpha blitting..)
        SDL_Surface *accelimage = SDL_ConvertSurface(_image, screen->format, SDL_SWSURFACE | SDL_SRCCOLORKEY);
        if (accelimage) {
            SDL_FreeSurface(_image);
            _image = accelimage;
            assert(!(_image->flags & SDL_HWSURFACE));
        } else {
            cerr << "SDL_ConvertSurface optimization failed.\n";
            exit(1); // probably unnecessary
        }
    }
    
    // we're not hflipped yet..
    _hflipped = false;
  //  cout << "WE OUTTA HERE\n";
}

void Set::matchPalette(SDL_Surface *screen)
{
    // Matches the _image's palette with the screen's
    // if applicable.

    /*
     * Palettized screen modes will have a default palette (a standard
     * 8*8*4 colour cube), but if the image is palettized as well we can
     * use that palette for a nicer colour matching
     */

    if (_image->format->palette && screen->format->palette) {
        SDL_SetColors(screen, _image->format->palette->colors, 0,
                      _image->format->palette->ncolors);
    }
}

void Set::setName(string name)
{
    _name = name;
}

void Set::setCount(int tilecount)
{
    assert(tilecount > 0);
    _tilecount = tilecount;
}

void Set::setHeight(int tileh)
{
    assert(tileh > 0);
    _tileh = tileh;
}

void Set::setWidth(int tilew)
{
    assert(tilew > 0);
    _tilew = tilew;
}

void Set::setTransparentColor(unsigned char r, unsigned char g, unsigned char b) {
    _transparentcolor = SDL_MapRGB(_image->format, r, g, b);
    _hastransparentcolor = true;

    // The 0,0 pixel color will always be my transparent color
    // so mark it as such.
    SDL_SetColorKey(_image, SDL_SRCCOLORKEY | SDL_RLEACCEL,
                    _transparentcolor);

}

void Set::verifyIntegrity()
{
    assert(_image != 0);
    assert(_tilew != 0);
    assert(_tileh != 0);
    assert(_tilecount > 0);
    assert(_name.size() > 0);
    assert(_transparentcolor != 0);
    // image width/height should be divisible by tile width/height
    assert(_image->w % _tilew == 0);
    assert(_image->h % _tileh == 0);
    assert(_image->w != 0);
    assert(_image->h != 0);

    assert(_image->w == _tilew * (_hflipped?2:1));
}


} // namespace Tile
} // namespace Salp
